/* lw-support/src/lib/Enums/IOTimeoutMode.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {

/*! \brief I/O timeout modes.

There are a few different ways for I/O to time out, depending on
how you want the program to behave. There can either be a "hard
timeout", which specifies the absolute maximum waiting time, or a "soft
timeout", which can be overruled by chunks of incoming data. Similarly,
it might be acceptable to return with only some data transferred, or it
might not.

*/
enum IOTimeoutMode {
    /*! \brief Hard timeout, partial transfer OK.

    This mode has a hard timeout, but partial data transfers are OK. In
    this mode, the read is guaranteed to return within the timeout with
    whatever data is available, but it doesn't guarantee to transfer the
    full amount of data. No data will be lost. An IOTimeout exception
    will only be thrown if no data is transferred at all during the
    specified timeout.

    */
    IOTimeoutHardPartial,

    /*! \brief Hard timeout, full data only.

    This mode has a hard timeout during which all data must be
    transferred. If only partial data transfer occurs, an IOTimeout
    exception is thrown at the end of the timeout period.

    */
    IOTimeoutHardFull,

    /*! \brief Soft timeout, partial transfer OK.

    This mode has a soft timeout, which is to say that the timeout
    period is restarted every time a partial transfer of data takes
    place. An IOTimeout exception will only be thrown if no data is
    transferred at all during the specified timeout.

    */
    IOTimeoutSoftPartial,

    /*! \brief Soft timeout, full data only.

    This mode has a soft timeout, which is to say that the timeout
    period is restarted every time a partial transfer of data takes
    place. However, upon reaching the end of the timeout period, an
    exception will be thrown if all the data has not been transferred.

    */
    IOTimeoutSoftFull,

    /*! \brief One-shot data transfer.

    This mode is used when you want to perform all available I/O and
    then return immediately (for instance, you might be listening for
    commands on a network connection, but not know the size of each
    command in advance).

    It will always return after it has completed as much data transfer
    as possible. It will only throw an IOTimeout exception if no data is
    transferred within the specified time limit.

    */
    IOTimeoutOneShot
};

}
