/* lw-support/src/lib/Events/EventManager.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



const uint32_t IOEventRead = EPOLLIN;
const uint32_t IOEventWrite = EPOLLOUT;
const uint32_t IOEventHUP = EPOLLHUP;
const uint32_t IOEventUrgent = EPOLLPRI;
const uint32_t IOEventError = EPOLLERR;



struct EventManager::pimpl {
    pimpl()
        : epollFd(-1), epollEvents(0), numEvents(0), shutdownFlag(false), processing(false),
        last_ns(now_ns())
    {
        pipeFd[0] = -1;
        pipeFd[1] = -1;
    }

    ~pimpl()
    {
        delete [] epollEvents;
        close(epollFd);
        close(pipeFd[0]);
        close(pipeFd[1]);
    }

    // epoll-related structures
    int epollFd;
    struct epoll_event* epollEvents;
    int numEvents;

    // pipe for signals/shutdown
    int pipeFd[2];

    // operational status
    bool shutdownFlag, processing;
    class ModifiedWhileProcessing : public lw::ProgramException { };
    void processingCheck(const wchar_t* chain)
    {
        if(processing) throw ModifiedWhileProcessing().chain(chain);
    }

    // timer events
    uint64_t last_ns;
    static uint64_t now_ns()
    {
        struct timeval tv;
        gettimeofday(&tv, 0);
        uint64_t ret = tv.tv_sec;
        ret *= tenExp9;
        ret += tv.tv_usec * tenExp3;
        return ret;
    }

    struct TimerEntry {
        EventCallbackTimer* timerCallback;
        ElapsedTime remainingPeriod, recurringPeriod;
        Key key;

        TimerEntry(EventCallbackTimer* timerCallback, ElapsedTime initialPeriod,
                   ElapsedTime recurringPeriod)
            : timerCallback(timerCallback), remainingPeriod(initialPeriod),
            recurringPeriod(recurringPeriod), key(Key::get())
        { }
    };

    struct TimerCmp {
        bool operator()(const struct TimerEntry& a, const struct TimerEntry& b)
        {
            return a.remainingPeriod < b.remainingPeriod;
        }
    };

    std::vector<TimerEntry> timers; // note -- this is a sorted list!

    enum InternalEvent {
        InternalEventShutdown
    };

    void dieHorribly()
    {
        processing = false;
        shutdownFlag = true;
        throw lw::ProgramException()
            .chain(L"EventManager::pimpl::dieHorribly()")
            .chain(L"EventManager::processEvents()");
    }

    void processInternalEvents()
    {
        InternalEvent ev;

        while(true) {
            ssize_t ret = read(pipeFd[pipeRead], (char*)&ev, sizeof(ev));
            if(ret == -1 && errno == EAGAIN) break;
            if(ret != sizeof(ev)) dieHorribly();

            switch(ev) {
            case InternalEventShutdown:
                shutdownFlag = true;
                break;

            default:
                dieHorribly();
            }
        }
    }

    // objects registered for updates (sorted list)
    std::vector<EventCallbackUpdate*> updateList;
};



EventManager::EventManager(int numDevicesHint, int numEvents)
    : p(new pimpl)
{
    try {
        if(numDevicesHint < 0) throw BadArgument(L"numDevicesHint", L">= 0", numDevicesHint);
        if(numEvents < 1) throw BadArgument(L"numEvents", L">= 1", numEvents);

        // create epoll structure
        p->numEvents = numEvents;
        p->epollEvents = new struct epoll_event[numEvents];
        p->epollFd = epoll_create(numDevicesHint + 1); // for the pipe
        if(p->epollFd < 0) throw SystemError().chain(L"epoll_create()");

        // set up pipes (non-blocking mode required)
        if(pipe(p->pipeFd) == -1) throw SystemError().chain(L"pipe()");
        int pipeFdFlags;
        if( (pipeFdFlags = fcntl(p->pipeFd[0], F_GETFL, 0)) == -1 ||
            fcntl(p->pipeFd[0], F_SETFL, pipeFdFlags | O_NONBLOCK) ||
            (pipeFdFlags = fcntl(p->pipeFd[1], F_GETFL, 0)) == -1 ||
            fcntl(p->pipeFd[1], F_SETFL, pipeFdFlags | O_NONBLOCK))
        {
            throw SystemError().chain(L"fcntl()");
        }

        struct epoll_event ev;
        memset(&ev, 0, sizeof(ev));
        ev.events = EPOLLIN | EPOLLERR | EPOLLET;
        if(epoll_ctl(p->epollFd, EPOLL_CTL_ADD, p->pipeFd[pipeRead], &ev))
            throw SystemError().chain(L"epoll_ctl(EPOLL_CTL_ADD)");
    }
    catch(lw::Exception& e) {
        delete p;
        e.chain(L"EventManager::EventManager()");
        throw;
    }
}



EventManager::~EventManager()
{
    delete p;
}



void EventManager::processEvents(int userTimeout)
{
    if(p->shutdownFlag) throw Shutdown();

    // find the shortest required timeout
    int timerTimeout = p->timers.empty() ? -1 : (int)(p->timers[0].remainingPeriod.to_ms());
    int timeout = std::min(userTimeout, timerTimeout);
    if(timeout < 0) {
        if(userTimeout >= 0) timeout = userTimeout;
        else if(timerTimeout >= 0) timeout = timerTimeout;
        else timeout = -1;
    }

    // run the event loop
    int ret = epoll_wait(p->epollFd, p->epollEvents, p->numEvents, timeout);

    if(ret == -1) {
        if(errno == EINTR) return;
        throw SystemError().chain(L"epoll_wait()").chain(L"EventManager::processEvents()");
    }
    if(!ret && timerTimeout == -1) throw Timeout();

    // process events
    for(int i = 0; i < ret; ++i) {
        EventCallbackIO* cb = (EventCallbackIO*)(p->epollEvents[i].data.ptr);
        if(cb) cb->ioReady(p->epollEvents[i].events);
        else p->processInternalEvents();
    }

    // evaluate timers
    uint64_t now_ns = p->now_ns(), ns_elapsed = now_ns - p->last_ns;
    p->last_ns = now_ns;
    ElapsedTime elapsedTime = ElapsedTime::from_ns(ns_elapsed);

    bool timerFired = false;
    int timerErased = 0;
    for(uint i = 0, end = p->timers.size(); i < end; ++i) {
        if(p->timers[i].remainingPeriod <= elapsedTime) {
            // fire timer
            timerFired = true;

            if(!p->timers[i].recurringPeriod) {
                p->timers[i].timerCallback->timer(p->timers[i].key);
                p->timers[i].remainingPeriod.set_ns(~(uint64_t)0);
                ++timerErased;
                continue;
            }

            uint64_t remaining = ns_elapsed,
                oldPeriod = p->timers[i].remainingPeriod.to_ns(),
                recur = p->timers[i].recurringPeriod.to_ns();
            while(true) {
                p->timers[i].timerCallback->timer(p->timers[i].key);
                oldPeriod += recur;
                if(remaining < oldPeriod) break;
                remaining -= recur;
            }

            oldPeriod -= remaining;
            p->timers[i].remainingPeriod.set_ns(oldPeriod);
        } else {
            p->timers[i].remainingPeriod -= elapsedTime;
        }
    }

    // resort list if appropriate
    if(timerFired) {
        std::stable_sort(p->timers.begin(), p->timers.end(), pimpl::TimerCmp());
        if(timerErased) p->timers.erase(p->timers.end() - timerErased, p->timers.end());
    }

    // deal with all the objects that need to be updated
    int x;
    while((x = p->updateList.size())) {
        --x;
        p->updateList[x]->update();
        ignoreUpdate(p->updateList[x]);
    }
}



void EventManager::registerDevice(IODevice* ioDevice, EventCallbackIO* eventCallback,
                                  uint32_t eventFlags)
{
    p->processingCheck(L"EventManager::registerDevice()");
    NullPointer::check(ioDevice, L"ioDevice", L"EventManager::registerDevice()");
    NullPointer::check(eventCallback, L"eventCallback", L"EventManager::registerDevice()");

    struct epoll_event ev;
    memset(&ev, 0, sizeof(ev));
    ev.data.ptr = eventCallback;
    ev.events = (eventFlags & (EPOLLIN | EPOLLOUT | EPOLLPRI | EPOLLERR | EPOLLHUP)) | EPOLLET;
    if(epoll_ctl(p->epollFd, EPOLL_CTL_ADD, ioDevice->getListenFd(), &ev)) throw SystemError()
        .chain(L"epoll_ctl(EPOLL_CTL_ADD)").chain(L"EventManager::registerDevice()");
}



void EventManager::modifyDevice(IODevice* ioDevice, EventCallbackIO* eventCallback,
                                uint32_t eventFlags)
{
    NullPointer::check(ioDevice, L"ioDevice", L"EventManager::modifyDevice()");
    NullPointer::check(eventCallback, L"eventCallback", L"EventManager::modifyDevice()");

    struct epoll_event ev;
    memset(&ev, 0, sizeof(ev));
    ev.data.ptr = eventCallback;
    ev.events = (eventFlags & (EPOLLIN | EPOLLOUT | EPOLLPRI | EPOLLERR | EPOLLHUP)) | EPOLLET;
    if(epoll_ctl(p->epollFd, EPOLL_CTL_MOD, ioDevice->getListenFd(), &ev)) throw SystemError()
        .chain(L"epoll_ctl(EPOLL_CTL_ADD)").chain(L"EventManager::modifyDevice()");
}



void EventManager::ignoreDevice(IODevice* ioDevice)
{
    p->processingCheck(L"EventManager::ignoreDevice()");
    NullPointer::check(ioDevice, L"ioDevice", L"EventManager::modifyDevice()");

    if(epoll_ctl(p->epollFd, EPOLL_CTL_DEL, ioDevice->getListenFd(), 0)) throw SystemError()
        .chain(L"epoll_ctl(EPOLL_CTL_DEL)").chain(L"EventManager::ignoreDevice()");
}



Key EventManager::registerTimer(EventCallbackTimer* timerCallback, ElapsedTime initialPeriod,
                                ElapsedTime recurringPeriod)
{
    p->processingCheck(L"EventManager::registerTimer()");
    NullPointer::check(timerCallback, L"timerCallback", L"EventManager::registerTimer()");
    if(!recurringPeriod) throw BadArgument(L"recurringPeriod", L"non-zero", L"zero");

    pimpl::TimerEntry t(timerCallback, initialPeriod, recurringPeriod);
    p->timers.insert(std::upper_bound(p->timers.begin(), p->timers.end(), t, pimpl::TimerCmp()), t);
    return t.key;
}



Key EventManager::registerTimer(EventCallbackTimer* timerCallback, ElapsedTime period)
{
    p->processingCheck(L"EventManager::registerTimer()");
    NullPointer::check(timerCallback, L"timerCallback", L"EventManager::registerTimer()");

    pimpl::TimerEntry t(timerCallback, period, lw::ElapsedTime());
    p->timers.insert(std::upper_bound(p->timers.begin(), p->timers.end(), t, pimpl::TimerCmp()), t);
    return t.key;
}



void EventManager::ignoreTimer(Key key)
{
    p->processingCheck(L"EventManager::ignoreTimer()");
    for(uint i = 0, end = p->timers.size(); i < end; ++i) {
        if(p->timers[i].key == key) {
            p->timers.erase(p->timers.begin() + i);
            return;
        }
    }
}



void EventManager::registerUpdate(lw::EventCallbackUpdate* updateCallback)
{
    // search for it in the list, in case it's already registered
    std::vector<EventCallbackUpdate*>::iterator iter =
        std::lower_bound(p->updateList.begin(), p->updateList.end(), updateCallback);
    if(*iter == updateCallback) return;

    // do a sorted insert
    p->updateList.insert(iter, updateCallback);

    // register ourselves with it, in case it's deleted
    updateCallback->managers.push_back(this);
}



void EventManager::ignoreUpdate(lw::EventCallbackUpdate* updateCallback)
{
    // search for it in the list, return if not found
    std::vector<EventCallbackUpdate*>::iterator iter =
        std::lower_bound(p->updateList.begin(), p->updateList.end(), updateCallback);
    if(*iter != updateCallback) return;

    // remove it, unregister with it
    std::vector<EventManager*>& evlist = (*iter)->managers;
    p->updateList.erase(iter);

    for(uint i = 0, end = evlist.size(); i < end; ++i) {
        if(evlist[i] == this) {
            evlist.erase(evlist.begin() + i);
            return;
        }
    }
}



void EventManager::shutdown()
{
    pimpl::InternalEvent ev = pimpl::InternalEventShutdown;
    write(p->pipeFd[pipeWrite], (char*)&ev, sizeof(ev));
    // TODO: yes, we ignore error handling here -- but what can we do?
}



}

/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
*/
