/* lw-support/src/lib/Events/EventManager.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/// Flag to denote I/O device has input available.
extern const uint32_t IOEventRead;

/// Flag to denote I/O device is ready for output.
extern const uint32_t IOEventWrite;

/// Flag to denote I/O device has urgent input available.
extern const uint32_t IOEventUrgent;

/// Flag to denote I/O device has developed an error condition.
extern const uint32_t IOEventError;

/// Flag to denote I/O device has signalled HUP.
extern const uint32_t IOEventHUP;



/*! \brief Event manager.

This object is responsible for driving an event-driven application. It collects system events (I/O,
timers and signals) and dispatches them appropriately. This class uses \c epoll() (which is a
\c select() replacement) to implement fast and efficient I/O event notification. To use it,
interested objects must register themselves through the various registration functions.

To poll for events, call the processEvents() function. This will process any events in the queue or,
if the queue is empty, will wait (possibly only for a limited time) for more. Object processing is
split into two phases within this function: first, all of the I/O, timer and signal callbacks are
processed, and then the update callbacks.

During the first phase of callbacks, it is not legal for objects to delete themselves, delete other
objects registered with the event manager, or to call any of the event manager functions (other than
modifyDevice() and registerUpdate()). During the second phase (the update callbacks), these
restrictions no longer apply. It is not possible to throw exceptions from callback functions.

*/
class EventManager : public Uncopyable {
public:
    /*! \brief Constructor. Sets up event manager.

    \param numDevicesHint The maximum likely number of devices that will be polled.
    \param numEvents The number of events to process in a single loop.
    \throws SystemError if an error occurs.

    The constructor creates and sets up the initial event manager. It expects two parameters: a
    hint as to how many devices you might be interested in (the kernel uses this to pre-allocate
    space, but the buffer size is dynamic anyway), and the number of events to process at once. This
    parameter is actually the size of the buffer the kernel uses to pass events to userspace; a
    larger buffer means more events to process per system call, but at the expense of memory usage.
    A reasonable value is to match \a numDevicesHint.

    */
    EventManager(int numDevicesHint, int numEvents);



    /// Destructor. Cleans up.
    ~EventManager();



    /*! \brief Register interest in a device's I/O readiness.

    \param ioDevice The I/O device in which you are interested.
    \param eventCallback The event sink object.
    \param eventFlags A bitfield of flags signifying which events you are interested in receiving.
        You may pass \c ~0 if you are interested in all events.
    \throws SystemError on error.
    \throws NullPointer if \a ioDevice or \a eventCallback are null.

    This function will register interest in receiving readiness notification for an I/O device. You
    specify the event sink object, which receives a callback for each readiness change, and the
    eventFlags object, which is a bitfield comprising the events you are interested in, and may
    include:
     - lw::IOEventRead
     - lw::IOEventWrite
     - lw::IOEventUrgent
     - lw::IOEventError
     - lw::IOEventHUP

    Once registered, you may cancel the callback by closing the device (once the kernel detects
    the file descriptor is closed, it will automatically remove it from the list of devices we are
    interested in), or explicitly by calling ignoreDevice() with the same parameters. If you do
    close the device's file descriptor, you should not call ignoreDevice().

    Note that it is your responsibility to free the EventCallbackIO object after calling
    ignoreDevice() or closing the file descriptor.

    */
    void registerDevice(lw::IODevice* ioDevice, lw::EventCallbackIO* eventCallback,
                        uint32_t eventFlags);



    /*! \brief Modify device.

    \param ioDevice The I/O device in which you are interested.
    \param eventCallback The event sink object.
    \param eventFlags A bitfield of flags signifying which events you are interested in receiving.
        You may pass \c ~0 if you are interested in all events.
    \throws SystemError on error.
    \throws NullPointer if \a ioDevice or \a eventCallback are null.

    This is like registerDevice(), only it modifies the information for a device you have already
    added. It can be used to change the callback object or to change the events in which you are
    interested. It can be called at any time, even from within a callback.

    */
    void modifyDevice(lw::IODevice* ioDevice, lw::EventCallbackIO* eventCallback,
                      uint32_t eventFlags);



    /*! \brief Stop listening to a device.

    \param ioDevice The I/O device in which you are no longer interested.
    \throws SystemError on error.
    \throws NullPointer if \a ioDevice is null.

    This stops reporting events for a particular device. It does not delete the event callback
    object or close the device. If you have already closed the device, you do not need to call this
    function, since devices are automatically removed from the epoll set when their file descriptors
    are closed.

    */
    void ignoreDevice(lw::IODevice* ioDevice);



    /*! \brief Register a timer callback.

    \param timerCallback The callback object.
    \param initialPeriod The initial time to wait before calling the callback.
    \param recurringPeriod After the initial callback, this specifies the period at which the event
        will be repeated.
    \throws NullPointer if \a timerCallback is null.
    \throws BadArgument if \a initialPeriod or \a recurringPeriod are zero.
    \returns A unique key.

    This function allows you to register a timer object for regular callback events. It returns a
    unique key that can be used with \a ignoreTimer (and which is also passed to the
    EventCallbackTimer::timer() callback function).

    This system is limited by the timing resolution of the operating system, which is typically on
    the order of 1-10 milliseconds. However, the average frequency over time will match what you
    request; the events will simply be bunched up.

    */
    lw::Key registerTimer(lw::EventCallbackTimer* timerCallback, lw::ElapsedTime initialPeriod,
                          lw::ElapsedTime recurringPeriod);



    /*! \brief Register a one-shot timer callback.

    \param timerCallback The callback object.
    \param period The time to wait before calling the callback.
    \throws NullPointer if \a timerCallback is null.
    \throws BadArgument if \a period is zero.
    \returns A unique key.

    This function allows you to register a timer object for a single callback event. It returns a
    unique key that can be used with \a ignoreTimer (and which is also passed to the
    EventCallbackTimer::timer() callback function).

    This system is limited by the timing resolution of the operating system, which is typically on
    the order of 1-10 milliseconds.

    */
    lw::Key registerTimer(lw::EventCallbackTimer* timerCallback, lw::ElapsedTime period);



    /*! \brief Stop a timer event.

    \param key The unique key returned from registerTimer().

    Calling this function will stop a timer event from occuring. If the value of \a key you
    specified does not exist, it will be ignored.

    */
    void ignoreTimer(lw::Key key);



    /*! \brief Register an object for updating.

    */
    void registerUpdate(lw::EventCallbackUpdate* updateCallback);



    /*! \brief Remove an object from the update list.

    */
    void ignoreUpdate(lw::EventCallbackUpdate* updateCallback);



    /*! \brief Process events.

    \param timeout The timeout, in milliseconds. Passing 0 means the
        function will time out or return immediately; passing a negative
        value means it will never time out.
    \throws SystemError if an error occurred. This could either be an
        error from the epoll_wait() function used internally, or an
        asynchronous error from the backing thread. In either case, it
        is only safe to delete the object and propagate the error.
    \throws EventManager::Shutdown if the object has been shut down.
    \throws Timeout if no events are processed during the specified
        timeout period.

    This function will wait until there are some events to process. It
    will then call the relevant callbacks and return. See the
    registerDevice() function for more information about callbacks.

    This function is not reentrant. It should only ever be called by one
    thread at once.

    */
    void processEvents(int timeout);

    /// Thrown when an EventManager times out.
    class Timeout : public ProgramException { };



    /*! \brief Shut down the object.

    \throws SystemError on error.

    This function marks the object as shut down. Any thread currently
    waiting in processEvents(), or that calls processEvents() from now
    on, will be

    This function is thread safe and reentrant.

    */
    void shutdown();

    /// Thrown when an EventManager is shut down.
    class Shutdown : public ProgramException { };

private:
    // use pimpl here, since this class will only be created infrequently
    struct pimpl;
    pimpl* p;
};



}

/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
*/
