/* lw-support/src/lib/Exceptions/BadArgument.h
 *
 *  (c)2006, Laurence Withers. Released under the GNU GPL. See file COPYING for more information
 *  and terms of license, or read http://www.gnu.org/copyleft/gpl.html .
*/

namespace lw {



/*! \brief Exception class: bad argument to function.

This exception class is used when an invalid argument is passed to a function (for instance, a
negative size in bytes, etc.). It stores some useful details for tracking down the bad argument.

*/
class BadArgument : public lw::Exception {
public:
    /*! \brief Constructor.

    \param argName Name of the invalid argument.
    \param acceptableRange Human-understandable range of valid arguments.
    \param value Actual value passed to function.

    */
    BadArgument(const std::wstring& argName, const std::wstring& acceptableRange,
                const std::wstring& value);



    /*! \brief Constructor.

    \param argName Name of the invalid argument.
    \param acceptableRange Human-understandable range of valid arguments.
    \param value Actual value passed to function.

    */
    template<class T> BadArgument(const std::wstring& argName, const std::wstring& acceptableRange,
                                  const T& value)
        : Exception(format(argName, acceptableRange, mkValue(value))), argName(argName),
        acceptableRange(acceptableRange), value(mkValue(value))
    { }



    /// Destructor (does nothing).
    virtual ~BadArgument() throw() { }



    /// Name of the invalid argument.
    const std::wstring argName;

    /// Human-understandable range of valid arguments.
    const std::wstring acceptableRange;

    /// Actual value passed to function.
    const std::wstring value;

private:
    template<class T> std::wstring mkValue(const T& value)
    {
        std::wostringstream ostr;
        ostr << value;
        return ostr.str();
    }

    std::wstring format(const std::wstring& argName, const std::wstring& acceptableRange,
                        const std::wstring& value);
};



}

/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
*/
