/* lw-support/src/lib/Exceptions/BadString.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Base exception class for string-related errors.

This simply serves as the base class for string- and character-related
errors.

*/
class BadString : public Exception {
public:
    /*! \brief Constructor.

    \param desc A description of the error.

    */
    BadString(const std::wstring& desc)
        : Exception(desc)
    { }

    /// Destructor.
    virtual ~BadString() throw() { }
};



/*! \brief Cannot represent source character.

This exception is thrown when a source character in a given encoding
cannot be represented in the destination encoding (e.g. ASCII cannot
represent arbitrary Unicode characters).

*/
class CannotConvertChar : public BadString {
private:
    wchar_t srcChar;
    int seqOffset;
    std::wstring srcEnc, destEnc;

    static std::wstring buildDesc(wchar_t srcChar, int seqOffset,
        const std::wstring& srcEnc, const std::wstring& destEnc);

public:
    /*! \brief Constructor.

    \param srcChar The (Unicode) code point that cannot be represented.
    \param seqOffset How far in the source string the conversion process
        reached before encountering the code point.
    \param srcEnc The name of the source encoding, e.g. "UCS-4".
    \param destEnc The name of the destination encoding, e.g. "ASCII".

    */
    CannotConvertChar(wchar_t srcChar, int seqOffset,
        const std::wstring& srcEnc, const std::wstring& destEnc);

    /// Destructor.
    virtual ~CannotConvertChar() throw()
    { }



    /// Find the source character (Unicode).
    wchar_t getSourceChar() const
    { return srcChar; }

    /// Find the offset in the source sequence.
    int getSeqOffset() const
    { return seqOffset; }

    /// Find the source encoding.
    const std::wstring& getSourceEncoding()
    { return srcEnc; }

    /// Find the destination encoding.
    const std::wstring& getDestEncoding()
    { return destEnc; }
};



/*! \brief Bad source character.

This exception is thrown when converting a string from one encoding to
another, and a byte sequence in the source string is not a valid
character. This can happen with UTF-8 for example.

The exception provides some context about where in the conversion
process the error occurred, potentially allowing a continuation of the
process, but certainly providing a detailed human-readable description
of the error.

*/
class BadSourceChar : public BadString {
private:
    std::vector<uint8_t> byteSeq; // the invalid byte sequence
    int seqOffset;
    std::wstring encoding; // the source encoding name
    std::wstring reason; // a specific reason (if there is one)

    static std::wstring buildDesc(const char* seqStart, const char* seqEnd,
        int seqOffset, const std::wstring& encodingName,
        const std::wstring& reason);

public:
    /*! \brief Construct from a byte sequence.

    \param seqStart Pointer to the start of the invalid byte sequence.
    \param seqEnd Pointer to one-past-the-end byte of the invalid byte
        sequence.
    \param seqOffset Offset (in bytes) of the invalid sequence from the
        start of the string. This could potentially be used to restart a
        conversion.
    \param encodingName Name of the encoding (in UCS-4).
    \param reason A specific reason for the character being bad, if one
        can be ascertained.

    */
    BadSourceChar(const char* seqStart, const char* seqEnd,
        int seqOffset, const std::wstring& encodingName,
        const std::wstring& reason = std::wstring());

    /// Destructor.
    virtual ~BadSourceChar() throw() { }



    /// Query the invalid byte sequence.
    const std::vector<uint8_t>& getByteSeq() const
    { return byteSeq; }

    /// Query the error offset.
    int getSeqOffset() const
    { return seqOffset; }

    /// Query the encoding.
    const std::wstring& getEncoding() const
    { return encoding; }

    /// Query the reason.
    const std::wstring& getReason() const
    { return reason; }
};



}
