/* lw-support/src/lib/Exceptions/Exception.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief LW Support exception base class.

This is the base class for all LW Support exceptions. It is derived
from std::exception, allowing you to treat it as a standard library
exception, if that is desired. It provides an implementation of the
what() function that returns a UTF-8 representation of the error; the
error can also be accessed through the toString() and toUtf8()
methods.

Exceptions chains may be built; these simply track the progress of the
exception through the library. An example of such a chain is:

<pre>void a()
{
    throw Exception("errorMessage");
}

void b()
{
    try {
        a();
    }
    catch(Exception& e) {
        e.chain("b()");
    }
}

int main()
{
    try {
        b();
    }
    catch(Exception& e) {
        cout << e.what() << endl;
    }
    return 0;
}</pre>

This would give the following output:

<pre>errorMessage
 From: b()</pre>

It is guaranteed that no exceptions will be thrown from this class,
unless the system is totally out of memory and cannot allocate std::string
objects.

*/
class Exception : public std::exception {
private:
    // description of the error itself
    std::wstring errorDesc;

    // buffer for what()
    mutable char* whatBuffer;

public:
    /*! \brief Construct from error message.

    \param errorDesc A UCS-4 description of the error.

    */
    Exception(const std::wstring& errorDesc);

    /// Destructor.
    virtual ~Exception() throw();



    /*! \brief Add a link in the error chain.

    \param chainDesc A UCS-4 description of the chain.
    \returns a reference to the exception, so you can chain again.

    This allows you to add another link into the exception chain, after
    the current exception. This is intended to be used when you want to
    propagate a lower-level error to a higher level, but want to mark
    where the exception has been on its way through.

    \note It is good practice to give a function name here.

    */
    Exception& chain(const std::wstring& chainDesc);



    /// Return UTF-8 error report.
    virtual const char* what() const throw();

    /// Return UTF-8 error report.
    std::string toUtf8() const;

    /// Return UCS-4 error report.
    const std::wstring& toString() const
    { return errorDesc; }
};



/*! \brief LW Support program exception base class.

Program exceptions are exceptions that are expected to occur, and do
not necessarily indicate system failure. They do signal that an
exceptional condition has arisen, but this is something that the program
should foresee and cope with. This sort of exception probably won't lead
directly to user intervention, so no description is required.

*/
class ProgramException : public Exception {
public:
    ProgramException()
        : Exception(L"Program exception. Should not be seen by users!")
    { }
};



/// Numeric overflow class.
class NumericOverflow : public Exception {
public:
    NumericOverflow()
        : Exception(L"Numeric overflow.")
    { }
};



/// Stream insertion operator.
std::wostream& operator<<(std::wostream&, const lw::Exception&);



}
