/* lw-support/src/lib/Exceptions/IO.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/// Base class for all I/O device exceptions.
class IOException : public Exception {
private:
    IOInterface* ioInterface;

public:
    /// Constructor.
    IOException(const std::wstring& desc, IOInterface* ioInterface)
        : Exception(desc),
          ioInterface(ioInterface)
    { }

    /// Destructor. Does nothing.
    virtual ~IOException() throw()
    { }

    /// Get the underlying device.
    IOInterface* getIOInterface() const
    { return ioInterface; }
};



/// Thrown when trying to read past end-of-file.
class EndOfFile : public IOException {
public:
    /// Constructor.
    EndOfFile(IOInterface* ioInterface)
        : IOException(L"End of file reached.", ioInterface)
    { }

    /// Destructor. Does nothing.
    virtual ~EndOfFile() throw()
    { }
};



/*! \brief Operation not permitted on device in current mode.

This exception is thrown whenever you try to carry out an operation on a
device, but the device isn't opened in the correct mode (e.g. if you try
to read from a write-only device, or if you try to write to an unopened
device).

*/
class IOModeError : public IOException {
private:
    const std::wstring op;
    const IOMode currentMode, requiredMode;

    static std::wstring getModeDesc(IOMode mode);
    static std::wstring getDesc(IOMode currentMode, IOMode requiredMode,
        const std::wstring& op);

public:
    /*! \brief Constructor.

    \param ioInterface The interface which was the source of the error.
    \param requiredMode What the device needed to be opened in for the
        operation to be legal.
    \param op A quick description of the operation (probably the
        function name).

    */
    IOModeError(IOInterface* ioInterface, const std::wstring& op,
        IOMode requiredMode);

    /// Destructor. Does nothing.
    virtual ~IOModeError() throw()
    { }

    /// Returns the name of the operation that was attempted.
    const std::wstring& getOp() const
    { return op; }

    /// Returns the device's current mode.
    IOMode getCurrentMode() const
    { return currentMode; }

    /// Returns the mode required for the operation.
    IOMode getRequiredMode() const
    { return requiredMode; }
};



/*! \brief I/O timed out.

This exception is thrown whenever an I/O operation times out.

*/
class IOTimeout : public IOException {
private:
    const std::wstring op;
    const int timeout;
    const IOTimeoutMode timeoutMode;
    const size_t amountTransferred;

    static std::wstring getDesc(const std::wstring& op, int timeout,
        IOTimeoutMode timeoutMode, size_t amountTransferred);


public:
    /*! \brief Constructor.

    \param ioInterface The device that timed out.
    \param op Name of the operation that was being performed (probably
        the function name).
    \param timeout The timeout value, in milliseconds.
    \param timeoutMode The timeout mode.
    \param amountTransferred The amount actually transferred before
        timing out.

    */
    IOTimeout(IOInterface* ioInterface, const std::wstring& op, int timeout,
        IOTimeoutMode timeoutMode, size_t amountTransferred);

    /// Destructor. Does nothing.
    virtual ~IOTimeout() throw()
    { }

    /// Returns the name of the operation that was attempted.
    const std::wstring& getOp() const
    { return op; }

    /// Returns the timeout value.
    int getTimeout() const
    { return timeout; }

    /// Returns the timeout mode.
    IOTimeoutMode getTimeoutMode() const
    { return timeoutMode; }

    /// Returns the amount of data actually transferred before timeout.
    size_t getAmountTransferred() const
    { return amountTransferred; }
};



/// File not found exception.
class FileNotFound : public Exception {
private:
    static std::wstring getDesc(const char* fileName);

public:
    /// Constructor.
    FileNotFound(const char* fileName);

    /// Destructor.
    ~FileNotFound() throw();

    /// The filename.
    const std::string fileName;
};



}
