/* lw-support/src/lib/IO/Filter.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief I/O filter base class.

The I/O filter class sits between the program and an underlying device
(or another filter, since it is possible to create arbitrary chains of
such filters). It can perform various transformations on the data. One
example is a gzip filter, which compresses data before writing it and
decompresses it when reading.

\warning Closing and reopening the underlying device may leave data in
         the filter's buffers. The solution is to call the reset()
         method (or delete the filter and create a new one).

*/
class IOFilter : public IOInterface {
protected:
    /// The source device.
    IOInterface* ioDev;

public:
    /// Constructor. Takes pointer to underlying device.
    IOFilter(IOInterface* ioDev);

    /// Destructor. Flushes write cache.
    virtual ~IOFilter()
    { }



    /*! \brief Empty the filter's write buffers.

    \throws SystemError if a system error occurs.
    \throws IOModeError if the file isn't open for writing.

    This causes the filter to flush its write buffers to the underlying
    device. If the underlying device has its own write buffer, this
    function does not cause the underlying buffer to be flushed; that
    has to be done explicitly.

    \sa sync()

    */
    virtual void flush() = 0;



    /*! \brief Reset the filter.

    This should only be called if you want to clear all buffer data and
    state information in the filter (e.g. if you have closed and
    reopened the underlying device). It's important to note that calling
    this will simply discard any data waiting to be read/written, so you
    should call flush() first.

    */
    virtual void reset() = 0;



    // implemented virtuals
    virtual IOMode getIOMode() const
    { return ioDev->getIOMode(); }



    /*! \brief Ensure data is written to disk.

    \throws SystemError if a system error occurs.
    \throws IOModeError if the file isn't open for writing.

    This function will flush any output buffers to the underlying
    device, and will then propagate down the chain, until it reaches
    the eventual sink device (e.g. file or network connection), at which
    point it will call fdatasync() to ensure the data is physically
    written to disk/network/whatever.

    \sa flush()

    */
    virtual void sync();
};



}
