/* lw-support/src/lib/IO/Interface.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief I/O interface.

This class is the interface that any I/O device or filter will
implement. It has both blocking and non-blocking functions.

*/
class IOInterface {
protected:
    /// The read/write mode of the device.
    IOMode ioMode;

public:
    /// Constructor. Initialises ioMode.
    IOInterface()
        : ioMode(IOClosed)
    { }



    /// Destructor -- does nothing.
    virtual ~IOInterface() { }



    /// Gets the mode the device was opened in.
    virtual IOMode getIOMode() const = 0;



    /*! \brief Read some data (non-blocking).

    \param buf Buffer to read data into.
    \param amt Maximum amount of data to read (i.e. size of buffer).
    \throws SystemError if a system error occurs.
    \throws EndOfFile if end-of-file has been reached.
    \throws IOModeError if the file isn't open for reading.
    \returns Number of bytes actually read (which will be zero if no
        data is currently available).

    Reads some data in a non-blocking manner. This call will return
    immediately and report the number of bytes read (which may well be
    less than the number of bytes requested). It will return zero if no
    data is available.

    */
    virtual size_t read(char* buf, size_t amt) = 0;



    /*! \brief Write some data (non-blocking).

    \param buf Data to write.
    \param amt Maximum amount of data to write.
    \throws SystemError if a system error occurs.
    \throws IOModeError if the file isn't open for writing.
    \returns Number of bytes actually written (will be zero if the
        device's output buffer is currently full).

    Writes some data in a non-blocking manner. This call will return
    immediately and report the number of bytes written (which may well
    be less than the number of bytes requested). It will return zero if
    the device wasn't ready for writing.

    */
    virtual size_t write(const char* buf, size_t amt) = 0;



    /*! \brief Wait for I/O.

    \param read If \a true, returns when the device has input available.
    \param write If \a true, returns when you can output to the device.
    \param timeout Timeout period, in milliseconds. 0 means immediate;
        negative means forever.
    \throws IOTimeout if a timeout occurs.
    \throws SystemError if a system error occurs.
    \retval true if an exceptional condition occurred.
    \retval false if no exception condition occurred.

    This function waits for the underlying device to become ready for
    the specified I/O operations, and returns when it is. It will
    return immediately if there is some exceptional condition, you can
    check for this by looking at the return value, which will be
    \a true if there is such a condition. Exceptional conditions
    include things such as out-of-band data and asynchronous errors.

    */
    virtual bool wait(bool read, bool write, int timeout) = 0;



    /*! \brief Read some data (blocking) (full buffer).

    \param buf Buffer to read data into.
    \param amt Maximum amount of data to read (i.e. size of buffer).
    \param timeout Timeout period, in milliseconds. 0 means immediate;
        negative means forever.
    \param timeoutMode The timeout mode.
    \throws SystemError if a system error occurs.
    \throws IOModeError if the file isn't open for reading.
    \throws EndOfFile if end-of-file has been reached.
    \throws IOTimeout if a timeout occurs.
    \returns Number of bytes actually read (always more than zero).

    Reads some data, blocking if none is available. Various timeout
    options are available. By default, the entire buffer must be filled
    in the specified time period (but this is set to "forever" by
    default). Should the end of file be reached before the buffer is
    completely filled, an EndOfFile exception will be thrown; instead,
    you could use wait() and read() individually to get around this.

    */
    virtual size_t readBlock(char* buf, size_t amt, int timeout = -1,
        IOTimeoutMode timeoutMode = IOTimeoutHardFull) = 0;



    /*! \brief Write some data (blocking).

    \param buf Data to write.
    \param amt Amount of data to write.
    \param timeout Timeout period, in milliseconds. 0 means immediate;
        negative means forever.
    \param timeoutMode The timeout mode.
    \throws SystemError if a system error occurs.
    \throws IOModeError if the file isn't open for writing.
    \throws IOTimeout if a timeout occurs.
    \returns Number of bytes actually written (always more than zero).

    Writes some data, blocking until it is sent. Various timeout
    options are available. By default, the entire buffer must be sent
    in the specified time period (but this is set to "forever" by
    default).

    */
    virtual size_t writeBlock(const char* buf, size_t amt, int timeout
        = -1, IOTimeoutMode timeoutMode = IOTimeoutHardFull) = 0;



    /*! \brief Ensure data is written to disk.

    \throws SystemError if a system error occurs.
    \throws IOModeError if the file isn't open for writing.

    This function will flush any output buffers, then call fdatasync()
    on the underlying device, ensuring the data is written to disk (or
    network, etc.).

    */
    virtual void sync() = 0;



    /*! \brief Check for asynchronous errors.

    \throws SystemError if there are any asynchronous errors.

    This function should be called to check for asynchronous errors.
    Examples of where you would use it are after a non-blocking
    connection has completed (socket will be marked as ready for
    reading/writing, so you need to check what's going on), and if
    you receive an error signal while waiting for events on the
    device, etc. Implementations of this function may do nothing if
    asynchronous errors can't occur (e.g. on standard disk files).

    */
    virtual void checkErrors() = 0;
};



}
