/* lw-support/src/lib/IO/PosixDevice.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



namespace {

// This class is a helper class used to keep track of timeouts and
// implement timeout behaviour for an I/O device.
class IOCountdown {
private:
    StopWatch started;
    IOInterface* ioInterface;
    std::wstring op;
    int timeout;
    IOTimeoutMode timeoutMode;
    size_t totalTransferred;

public:
    IOCountdown(IOInterface* ioInterface, const std::wstring& op, int timeout,
        IOTimeoutMode timeoutMode)
        : ioInterface(ioInterface), op(op), timeout(timeout),
          timeoutMode(timeoutMode), totalTransferred(0)
    { }

    // Call this when some data is transferred.
    void transferred(size_t amt)
    {
        // record amount transferred
        totalTransferred += amt;

        // reset countdown timer
        if(amt && (timeoutMode == IOTimeoutSoftFull ||
            timeoutMode == IOTimeoutSoftPartial))
        {
            started.start();
        }
    }

    // Call this when poll() returns a timeout. Returns no. of bytes
    // transferred, but throws an exception if specified by behaviour.
    size_t timedOut()
    {
        switch(timeoutMode) {
            case IOTimeoutHardPartial:
            case IOTimeoutSoftPartial:
            case IOTimeoutOneShot:
                if(totalTransferred) break;
                // fall through
            case IOTimeoutHardFull:
            case IOTimeoutSoftFull:
                throw IOTimeout(ioInterface, op, timeout,
                    timeoutMode, totalTransferred);
        }
        return totalTransferred;
    }

    // Call this when you need to update the timeout variable. It will
    // return false if the timeout has happened.
    bool update(int& remaining)
    {
        /* Special case for one-shot mode: after initial transfer has
           completed, we want to continue retrying until the I/O is
           drained; therefore, we set the timeout to zero.

           FIXME: a stray signal could cause poll() to fail with EINTR,
           at which point update() will be called event though no I/O
           has yet occurred, possibly leading to a premature timeout.
        */
        if(timeoutMode == IOTimeoutOneShot) return 0;

        if(timeout < 0) {
            remaining = timeout;
            return true;
        }

        remaining = timeout - started.elapsed().to_ms();
        return (remaining >= 0);
    }
};

}



IOPosixDevice::~IOPosixDevice()
{
    if(fd >= 0) TEMP_FAILURE_RETRY(::close(fd));
}



void IOPosixDevice::close()
{
    if(fd != -1) {
        int ret = TEMP_FAILURE_RETRY(::close(fd));
        fd = -1;
        ioMode = IOClosed;

        if(ret) {
            throw SystemError()
                .chain(L"close()")
                .chain(L"IOPosixDevice::close()");
        }
    }
}



size_t IOPosixDevice::read(char* buf, size_t amt)
{
    try {
        checkRead();

        ssize_t ret = TEMP_FAILURE_RETRY(::read(fd, buf, amt));
        if(!ret) throw EndOfFile(this);
        if(ret == -1) {
            if(errno == EAGAIN) return 0;
            throw SystemError().chain(L"read()");
        }
        return ret;
    }
    catch(Exception& e) {
        e.chain(L"IOPosixDevice::read()");
        throw;
    }
}



size_t IOPosixDevice::write(const char* buf, size_t amt)
{
    try {
        checkWrite();

        ssize_t ret = TEMP_FAILURE_RETRY(::write(fd, buf, amt));
        if(ret == -1) {
            if(errno == EAGAIN) return 0;
            throw SystemError().chain(L"write()");
        }
        return ret;
    }
    catch(Exception& e) {
        e.chain(L"IOPosixDevice::write()");
        throw;
    }
}



bool IOPosixDevice::wait(bool read, bool write, int timeout)
{
    struct pollfd ufd = { fd, POLLPRI | (read ? POLLIN : 0) | (write ? POLLOUT : 0), 0 };
    switch(poll(&ufd, 1, timeout)) {
        case -1:
            if(errno == EINTR) return false;
            throw SystemError().chain(L"IOPosixDevice::wait()").chain(L"poll()");

        case 0:
            // timeout
            throw IOTimeout(this, L"wait()", timeout, IOTimeoutHardFull, 0);

        default:
            break;
    }
    return (ufd.revents & (POLLERR | POLLPRI | POLLHUP | POLLNVAL));
}



size_t IOPosixDevice::readBlock(char* buf, size_t amt, int timeout,
    IOTimeoutMode timeoutMode)
{
    try {
        checkRead();

        size_t remaining = amt;
        int timeout_remaining = timeout;
        struct pollfd ufd = { fd, POLLIN, 0 };
        IOCountdown countdown(this, L"readBlock()", timeout, timeoutMode);

        while(true) {
            switch(poll(&ufd, 1, timeout_remaining)) {
                case -1:
                    if(errno == EINTR) break;
                    throw SystemError().chain(L"poll()");

                case 0:
                    // timeout
                    return countdown.timedOut();

                default:
                    // perform read
                    size_t res = read(buf, remaining);
                    if( !(remaining -= res) ) return amt;
                    buf += res;
                    countdown.transferred(res);
                    break;
            }

            // update timeout
            if(!countdown.update(timeout_remaining))
                return countdown.timedOut();
        }
    }
    catch(Exception& e) {
        e.chain(L"IOPosixDevice::readBlock()");
        throw;
    }
}



size_t IOPosixDevice::writeBlock(const char* buf, size_t amt, int timeout,
    IOTimeoutMode timeoutMode)
{
    try {
        checkWrite();

        size_t remaining = amt;
        int timeout_remaining = timeout;
        struct pollfd ufd = { fd, POLLOUT, 0 };
        IOCountdown countdown(this, L"writeBlock()", timeout, timeoutMode);

        while(true) {
            switch(poll(&ufd, 1, timeout_remaining)) {
                case -1:
                    if(errno == EINTR) break;
                    throw SystemError().chain(L"poll()");

                case 0:
                    // timeout
                    return countdown.timedOut();

                default:
                    // perform read
                    size_t res = write(buf, remaining);
                    if( !(remaining -= res) ) return amt;
                    buf += res;
                    countdown.transferred(res);
                    break;
            }

            // update timeout
            if(!countdown.update(timeout_remaining))
                return countdown.timedOut();
        }
    }
    catch(Exception& e) {
        e.chain(L"IOPosixDevice::writeBlock()");
        throw;
    }
}



void IOPosixDevice::sync()
{
    try {
        checkWrite();
    }
    catch(Exception& e) {
        e.chain(L"IOPosixDevice::sync()");
        throw;
    }

    if(TEMP_FAILURE_RETRY(::fdatasync(fd))) {
        throw SystemError()
            .chain(L"fdatasync()")
            .chain(L"IOPosixDevice::sync()");
    }
}



void IOPosixDevice::setNonBlocking()
{
    int flags = fcntl(fd, F_GETFL, 0);
    if(flags == -1) {
        throw SystemError()
            .chain(L"fcntl(..., F_GETFL, ...)")
            .chain(L"IOPosixDevice::setNonBlocking()");
    }
    if(flags & O_NONBLOCK) return;
    flags |= O_NONBLOCK;
    if(fcntl(fd, F_SETFL, flags)) {
        throw SystemError()
            .chain(L"fcntl(..., F_SETFL, ...)")
            .chain(L"IOPosixDevice::setNonBlocking()");
    }
}



int IOPosixDevice::ioModeFlags(IOMode mode)
{
    switch(mode) {
        case IOClosed:
            throw IOModeError(this, L"open", IONone);

        case IONone:
            return 0;

        case IORead:
            return O_RDONLY;

        case IOWrite:
            return O_WRONLY;

        case IOReadWrite:
            return O_RDWR;
    }
    throw ProgramException();
}



void IOPosixDevice::checkRead() const
{
    IOMode mode = getIOMode();
    if(mode == IORead || mode == IOReadWrite) return;
    throw IOModeError(const_cast<IOPosixDevice*>(this), L"read", IORead);
}



void IOPosixDevice::checkWrite() const
{
    IOMode mode = getIOMode();
    if(mode == IOWrite || mode == IOReadWrite) return;
    throw IOModeError(const_cast<IOPosixDevice*>(this), L"write", IOWrite);
}



void IOPosixDevice::checkOpen(const std::wstring& op) const
{
    IOMode mode = getIOMode();
    if(mode != IOClosed) return;
    throw IOModeError(const_cast<IOPosixDevice*>(this), op, IOWrite);
}



void IOPosixDevice::posixOpen(const char* path, IOMode mode)
{
    fd = TEMP_FAILURE_RETRY(::open(path, ioModeFlags(mode)));
    if(fd == -1) {
        if(errno == ENOENT) throw FileNotFound(path);
        throw SystemError()
            .chain(L"open()")
            .chain(L"IOPosixDevice::posixOpen()");
    }
    try {
        setNonBlocking();
    }
    catch(Exception& e) {
        TEMP_FAILURE_RETRY(::close(fd));
        fd = -1;
        e.chain(L"IOPosixDevice::posixOpen()");
        throw;
    }
    ioMode = mode;
}



void IOPosixDevice::checkErrors()
{
}



}
