/* lw-support/src/lib/IO/PosixDevice.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief POSIX I/O device class.

Many standard devices can be accessed through a POSIX file descriptor.
This is the base class for all such devices. It provides implementations
for reading, writing, etc.; but not for actually opening devices.

\todo Port readBlock() and writeBlock() over to use wait(), and deal
    with any exceptional conditions somehow.

*/
class IOPosixDevice : public IODevice, public Uncopyable {
protected:
    /// The POSIX file descriptor.
    int fd;

    // implemented virtuals
    virtual int getListenFd() const
    { return fd; }

    /// Convenience function: sets the fd into non-blocking mode.
    void setNonBlocking();

    /// Convenience function: converts enumeration value into GNU
    /// \c open() bitfield, throwing an exception if necessary.
    int ioModeFlags(IOMode mode);

    /// Convenience function: POSIX open() wrapper, throws exception.
    void posixOpen(const char* path, IOMode mode);

    /// Convenience function: check that we can read or throw exception.
    void checkRead() const;

    /// Convenience function: check that we can write or throw exception.
    void checkWrite() const;

    /// Convenience function: check device is open or throw exception.
    void checkOpen(const std::wstring& op) const;

public:
    /// Constructor. Does nothing.
    IOPosixDevice()
        : fd(-1)
    { }

    /// Destructor. Closes device.
    virtual ~IOPosixDevice();

    // implemented virtuals
    virtual void close();
    virtual size_t read(char* buf, size_t amt);
    virtual size_t write(const char* buf, size_t amt);
    virtual bool wait(bool read, bool write, int timeout);
    virtual size_t readBlock(char* buf, size_t amt, int timeout = -1,
        IOTimeoutMode timeoutMode = IOTimeoutHardFull);
    virtual size_t writeBlock(const char* buf, size_t amt, int timeout
        = -1, IOTimeoutMode timeoutMode = IOTimeoutHardFull);
    virtual void sync();
    virtual void checkErrors();
};



}
