/* lw-support/src/lib/IO/Util/WriteTask.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Class which writes some data to an I/O device and then closes it.

Instances of this class will write data to an I/O device, using non-blocking I/O, and (on
destruction) will close the device and delete it. It has options for timeouts and deals correctly
with errors. On task completion, the object will delete itself.

A typical usage example would be to accept a TCP connection, instantiate an IOWriteTask (with a
suitable message and timeout, pointed at the TCP connection, connected to a CompletionList object),
register the task with the CompletionList, and then simply process events as usual, forgetting about
the TCP connection and the IOWriteTask instance.

*/
class IOWriteTask : public CompletionTask,
    private EventCallbackIO,
    private EventCallbackTimer
{
private:
    lw::EventManager& eventManager;
    IOInterface* ioDev;
    int timeoutInactive, timeoutOverall;
    char* data, * wr;
    size_t amt;
    bool copy;
    Key timerKey0, timerKey1;

    void tryWrite();

    // from EventCallbackIO
    virtual bool ioReady(uint32_t flags);

    // from EventCallbackTimer
    virtual bool timer(Key key);

public:
    /*! \brief Constructor. Sets up the task.

    \param eventManager The event manager object to use.
    \param ioDev The I/O device in question.
    \param timeoutInactive An inactivity timeout period, in milliseconds. -1 to disable.
    \param timeoutOverall An overall timeout period, in milliseconds. -1 to disable.
    \param completion The object which will be notified about completion. Can be 0.
    \param data Pointer to the block of data to write.
    \param amt Number of bytes of data to write.
    \param copy Pass \c true if you want the data to be copied, or \c false if you guarantee its
        persistence until the task is completed.

    \throws NullPointer if \a ioDev or \a data is 0.
    \throws SystemError if a system call fails.

    Begins the process. Registers the device with the event manager, attempts to write some data
    (these can result in system errors). On completion, the object will delete itself.

    */
    IOWriteTask(lw::EventManager& eventManager, IODevice* ioDev, int timeoutInactive,
                int timeoutOverall, CompletionCallback* completion,
                const char* data, size_t amt, bool copy);



    /// Destructor. Closes and deletes I/O device.
    virtual ~IOWriteTask();



    /// Aborts task, closing and deleting the I/O device and the instance.
    virtual void taskAbort(const lw::Exception& e);
};



}
