/* lw-support/src/lib/Net/Address/IPv4.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief IPv4 network address.

This class is used to represent an IPv4 network address. It has a basic
blocking DNS lookup function.

*/
class NetAddressIPv4 : public NetAddress {
private:
    uint8_t addr[4];

public:
    //BEGIN // Well-known addresses ////////////////////////////////////

    /// 0.0.0.0
    static const NetAddressIPv4 any;

    /// 127.0.0.1
    static const NetAddressIPv4 loopback;



    //BEGIN // Constructors ////////////////////////////////////////////

    /// Default constructor: address 0.0.0.0.
    NetAddressIPv4()
    {
        memset(addr, 0, 4);
    }

    /// Copy constructor.
    NetAddressIPv4(const NetAddressIPv4& other)
        : NetAddress()
    {
        memcpy(addr, other.addr, 4);
    }

    /// Constructor: specify numerically.
    NetAddressIPv4(uint8_t a, uint8_t b, uint8_t c, uint8_t d)
    {
        addr[0] = a;
        addr[1] = b;
        addr[2] = c;
        addr[3] = d;
    }

    /// Constructor: from array.
    NetAddressIPv4(const uint8_t* newAddr)
    {
        memcpy(addr, newAddr, 4);
    }

    /// Constructor: from memory.
    NetAddressIPv4(const char* newAddr)
    {
        memcpy(addr, newAddr, 4);
    }

    // implemented virtual
    virtual NetAddressIPv4* clone() const
    { return new NetAddressIPv4(*this); }



    //BEGIN // Modifying functions /////////////////////////////////////

    /// Assignment operator.
    NetAddressIPv4& operator=(const NetAddressIPv4& other)
    {
        memcpy(addr, other.addr, 4);
        return *this;
    }

    /// Set numerically.
    NetAddressIPv4& set(uint8_t a, uint8_t b, uint8_t c, uint8_t d)
    {
        addr[0] = a;
        addr[1] = b;
        addr[2] = c;
        addr[3] = d;
        return *this;
    }

    /// Set from array.
    NetAddressIPv4& set(const uint8_t* newAddr)
    {
        memcpy(addr, newAddr, 4);
        return *this;
    }

    /// Set from memory.
    NetAddressIPv4& set(const char* newAddr)
    {
        memcpy(addr, newAddr, 4);
        return *this;
    }



    //BEGIN // Query functions /////////////////////////////////////////

    /// Copy to array.
    uint8_t* copy(uint8_t* dest) const
    {
        memcpy(dest, addr, 4);
        return dest;
    }

    /// Copy to memory.
    char* copy(char* dest) const
    {
        memcpy(dest, addr, 4);
        return dest;
    }

    /// For interfacing with the C library.
    struct sockaddr* copy(struct sockaddr_in* s, uint16_t port) const
    {
        s->sin_family = AF_INET;
        s->sin_port = htons(port);
        memcpy(&(s->sin_addr.s_addr), addr, 4);
        return (struct sockaddr*)s;
    }



    //BEGIN // String functions ////////////////////////////////////////

    // implemented virtual
    virtual std::wstring toString() const;



    /*! \brief Construct from string.

    \param str The string containing an IPv4 address to parse.
    \throws ParseError if a parse error occurs.
    \returns A new address object.

    This function parses an IPv4 address in dotted quad format, i.e.
    <tt>a.b.c.d</tt> where a, b, c and d are all 8-bit unsigned values.
    Any trailing or leading whitespace is ignored.

    */
    static NetAddressIPv4 fromString(const std::wstring& str);



    /*! \brief Look up via DNS (blocking).

    \param host The hostname to look up.
    \throws DnsError If an error occurs.
    \returns A new address object.

    This function looks up a specific host via DNS, returning its IPv4
    address. It is a blocking function.

    \todo std::wstring-capable version

    */
    static NetAddressIPv4 fromDNS(const std::string& host);
};



}
