/* lw-support/src/lib/Net/Address/IPv6.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief IPv6 network address.

This class is used to represent an IPv6 network address. It does not
have functions for converting names to IP addresses (DNS); this must be
done with the ... class.

\todo mention DNS class

*/
class NetAddressIPv6 : public NetAddress {
private:
    uint8_t addr[16];

public:
    //BEGIN // Well-known addresses ////////////////////////////////////

    /// The address \c ::
    static const NetAddressIPv6 any;

    /// The address \c ::1
    static const NetAddressIPv6 loopback;



    //BEGIN // Constructors ////////////////////////////////////////////

    /// Default constructor: address \c ::
    NetAddressIPv6()
    {
        memset(addr, 0, 16);
    }

    /// Copy constructor.
    NetAddressIPv6(const NetAddressIPv6& other)
        : NetAddress()
    {
        memcpy(addr, other.addr, 16);
    }

    /// Constructor: from array.
    NetAddressIPv6(const uint8_t* newAddr)
    {
        memcpy(addr, newAddr, 16);
    }

    /// Constructor: from memory.
    NetAddressIPv6(const char* newAddr)
    {
        memcpy(addr, newAddr, 16);
    }

    // implemented virtual
    virtual NetAddressIPv6* clone() const
    { return new NetAddressIPv6(*this); }



    //BEGIN // Modifying functions /////////////////////////////////////

    /// Assignment operator.
    NetAddressIPv6& operator=(const NetAddressIPv6& other)
    {
        memcpy(addr, other.addr, 16);
        return *this;
    }

    /// Set from array.
    NetAddressIPv6& set(const uint8_t* newAddr)
    {
        memcpy(addr, newAddr, 16);
        return *this;
    }

    /// Set from memory.
    NetAddressIPv6& set(const char* newAddr)
    {
        memcpy(addr, newAddr, 16);
        return *this;
    }



    //BEGIN // Query functions /////////////////////////////////////////

    /// Copy to array.
    uint8_t* copy(uint8_t* dest) const
    {
        memcpy(dest, addr, 16);
        return dest;
    }

    /// Copy to memory.
    char* copy(char* dest) const
    {
        memcpy(dest, addr, 16);
        return dest;
    }

    /// For interfacing with the C library.
    struct sockaddr* copy(struct sockaddr_in6* s, uint16_t port) const
    {
        s->sin6_family = AF_INET6;
        s->sin6_port = htons(port);
        memcpy(s->sin6_addr.s6_addr, addr, 16);
        return (struct sockaddr*)s;
    }



    //BEGIN // String functions ////////////////////////////////////////

    // implemented virtual
    virtual std::wstring toString() const;



    /*! \brief Construct from std::string.

    \param str The std::string containing an IPv6 address to parse.
    \throws ParseError if a parse error occurs.
    \returns A new address object.

    This function parses an IPv6 address in colon-delimited format,
    e.g. \c "0123:4567:89ab:cdef:0123:4567:89ab:cdef". It understands
    the compact form, e.g. \c "::1". Any traling or leading whitespace
    is ignored.

    */
    static NetAddressIPv6 fromString(const std::wstring& str);



    /*! \brief Look up via DNS (blocking).

    \param host The hostname to look up.
    \throws DnsError If an error occurs.
    \returns A new address object.

    This function looks up a specific host via DNS, returning its IPv6
    address. It is a blocking function.

    \todo std::wstring-capable version

    */
    static NetAddressIPv6 fromDNS(const std::string& host);
};



}
