/* lw-support/src/lib/Net/Protocols/HTTP.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



//BEGIN // HTTPHeaderParser ////////////////////////////////////////////



HTTPHeaderParser::HTTPHeaderParser()
    : state(StateBegin)
{
}



size_t HTTPHeaderParser::parse(const char* data, size_t amt)
{
    size_t parsed = 0;
    char ch;

begin:
    if(parsed == amt || state == StateDone) return parsed;
    if(state == StateError) throw HTTPMalformed(L"(previous error)");

#define ERROR(p) do { \
    state = StateError; \
    throw HTTPMalformed(p); \
}while(0)

    // extract character (ASCII only)
    ch = data[parsed++];
    if(ch & ~0x7F) {
        state = StateError;
        throw HTTPMalformed(L"Non-ASCII data encountered in HTTP header.");
    }

    switch(state) {
        case StateBegin:
            if(ch == '\r') {
                state = StateNewline2;
                break;
            }
            if(!isgraph(ch) || ch == ':') ERROR(L"invalid field name");
            fieldName = tolower(ch);
            state = StateFieldName;
            break;

        case StateNewline:
            if(ch != '\n') ERROR(L"expecting \\n after \\r");
            state = StateLine;
            break;

        case StateLine:
            if(ch == '\r') {
                state = StateNewline2;
                break;
            }
            if(isblank(ch)) {
                state = StateFieldValue;
                break;
            }
            if(!isgraph(ch) || ch == ':') ERROR(L"invalid field name");
            fieldName = tolower(ch);
            state = StateFieldName;
            break;

        case StateFieldName:
            if(ch == ':') {
                state = StateFieldValue;
                break;
            }
            if(!isgraph(ch)) ERROR(L"invalid field name");
            fieldName += tolower(ch);
            break;

        case StateFieldValue:
            if(ch == '\r') {
                state = StateNewline;
                break;
            }
            if(!isprint(ch) && ch != '\t') ERROR(L"invalid field value");
            fields[fieldName] += ch;
            break;

        case StateNewline2:
            if(ch != '\n') ERROR(L"expecting \\n after \\r");
            state = StateDone;
            break;

        case StateDone:
        case StateError:
            // should never get here!
            throw lw::ProgramException().chain(L"HTTPHeaderParser::parse() [1]");
    }

    goto begin;
#undef ERROR
}



//END // HTTPHeaderParser //////////////////////////////////////////////
//BEGIN // HTTPResponseParser //////////////////////////////////////////



const char* HTTPResponseParser::HTTPSeq = "HTTP/";

HTTPResponseParser::HTTPResponseParser()
    : state(StateHTTPSeq), HTTPSeqPos(HTTPSeq),
      versionMajor(0), versionMinor(0), responseCode(0), fields(hp.fields)
{
}



size_t HTTPResponseParser::parse(const char* data, size_t amt)
{
    size_t parsed = 0;
    char ch;

begin:
    if(parsed == amt || state == StateDone) return parsed;
    if(state == StateError) throw HTTPMalformed(L"(previous error)");
    if(state == StateHeader) {
        try {
            parsed += hp.parse(data + parsed, amt - parsed);
            if(hp.completed()) state = StateDone;
            return parsed;
        }
        catch(...) {
            state = StateError;
            throw;
        }
    }

#define ERROR(p) do { \
    state = StateError; \
    throw HTTPMalformed(p); \
}while(0)

    // extract character (ASCII only)
    ch = data[parsed++];
    if(ch & ~0x7F) {
        state = StateError;
        throw HTTPMalformed(L"Non-ASCII data encountered in HTTP response line.");
    }

    switch(state) {
        case StateHTTPSeq:
            if(ch != *HTTPSeqPos) ERROR(L"expecting 'HTTP/major.minor'");
            if(!*(++HTTPSeqPos)) state = StateHTTPMajor1;
            break;

        case StateHTTPMajor1:
            if(!isdigit(ch)) ERROR(L"expecting 'HTTP/major.minor'");
            versionMajor = ch - '0';
            state = StateHTTPMajor;
            break;

        case StateHTTPMajor:
            if(isdigit(ch)) {
                versionMajor *= 10;
                versionMajor += ch - '0';
            } else if(ch == '.') {
                state = StateHTTPMinor1;
            } else ERROR(L"expecting 'HTTP/major.minor'");
            break;

        case StateHTTPMinor1:
            if(!isdigit(ch)) ERROR(L"expecting 'HTTP/major.minor'");
            versionMinor = ch - '0';
            state = StateHTTPMinor;
            break;

        case StateHTTPMinor:
            if(isdigit(ch)) {
                versionMinor *= 10;
                versionMinor += ch - '0';
            } else if(ch == ' ') {
                state = StateHTTPResp1;
            } else ERROR(L"expecting 'HTTP/major.minor'");
            break;

        case StateHTTPResp1:
            if(!isdigit(ch)) ERROR(L"invalid response code");
            responseCode = 100 * (ch - '0');
            state = StateHTTPResp2;
            break;

        case StateHTTPResp2:
            if(!isdigit(ch)) ERROR(L"invalid response code");
            responseCode += 10 * (ch - '0');
            state = StateHTTPResp3;
            break;

        case StateHTTPResp3:
            if(!isdigit(ch)) ERROR(L"invalid response code");
            responseCode += (ch - '0');
            state = StateHTTPResp4;
            break;

        case StateHTTPResp4:
            if(ch != ' ') ERROR(L"invalid response code");
            state = StateHTTPMsg;
            break;

        case StateHTTPMsg:
            if(ch == '\n') ERROR(L"\\n invalid on its own");
            else if(ch == '\r') state = StateHTTPNewline;
            else status += ch;
            break;

        case StateHTTPNewline:
            if(ch != '\n') ERROR(L"expecting \\r after \\n");
            state = StateHeader;
            break;

        case StateHeader:
        case StateDone:
        case StateError:
            // should never get here!
            throw lw::ProgramException().chain(L"HTTPResponseParser::parse() [1]");
    }

    goto begin;
#undef ERROR
}



//END // HTTPResponseParser ////////////////////////////////////////////
//BEGIN // HTTPRequestParser ///////////////////////////////////////////



const char* HTTPRequestParser::HTTPSeq = "HTTP/";

HTTPRequestParser::HTTPRequestParser()
    : state(StateHTTPMethod1), HTTPSeqPos(HTTPSeq), fields(hp.fields)
{
}



size_t HTTPRequestParser::parse(const char* data, size_t amt)
{
    size_t parsed = 0;
    char ch;

begin:
    if(parsed == amt || state == StateDone) return parsed;
    if(state == StateError) throw HTTPMalformed(L"(previous error)");
    if(state == StateHeader) {
        try {
            parsed += hp.parse(data + parsed, amt - parsed);
            if(hp.completed()) state = StateDone;
            return parsed;
        }
        catch(...) {
            state = StateError;
            throw;
        }
    }

#define ERROR(p) do { \
    state = StateError; \
    throw HTTPMalformed(p); \
}while(0)

    // extract character (ASCII only)
    ch = data[parsed++];
    if(ch & ~0x7F) {
        state = StateError;
        throw HTTPMalformed(L"Non-ASCII data encountered in HTTP request line.");
    }

    switch(state) {
        case StateHTTPMethod1:
            if(!isgraph(ch)) ERROR(L"invalid HTTP method");
            method = ch;
            state = StateHTTPMethod;
            break;

        case StateHTTPMethod:
            if(ch == ' ') {
                state = StateHTTPUrl1;
                break;
            }
            if(!isgraph(ch)) ERROR(L"invalid HTTP method");
            method += ch;
            break;

        case StateHTTPUrl1:
            if(!isgraph(ch)) ERROR(L"invalid HTTP URL");
            url = ch;
            state = StateHTTPUrl;
            break;

        case StateHTTPUrl:
            if(ch == ' ') {
                state = StateHTTPSeq;
                break;
            }
            if(!isgraph(ch)) ERROR(L"invalid HTTP URL");
            url += ch;
            break;

        case StateHTTPSeq:
            if(ch != *HTTPSeqPos) ERROR(L"expecting 'HTTP/major.minor'");
            if(!*(++HTTPSeqPos)) state = StateHTTPMajor1;
            break;

        case StateHTTPMajor1:
            if(!isdigit(ch)) ERROR(L"expecting 'HTTP/major.minor'");
            versionMajor = ch - '0';
            state = StateHTTPMajor;
            break;

        case StateHTTPMajor:
            if(isdigit(ch)) {
                versionMajor *= 10;
                versionMajor += ch - '0';
            } else if(ch == '.') {
                state = StateHTTPMinor1;
            } else ERROR(L"expecting 'HTTP/major.minor'");
            break;

        case StateHTTPMinor1:
            if(!isdigit(ch)) ERROR(L"expecting 'HTTP/major.minor'");
            versionMinor = ch - '0';
            state = StateHTTPMinor;
            break;

        case StateHTTPMinor:
            if(isdigit(ch)) {
                versionMinor *= 10;
                versionMinor += ch - '0';
            } else if(ch == '\r') {
                state = StateHTTPNewline;
            } else ERROR(L"expecting 'HTTP/major.minor'");
            break;

        case StateHTTPNewline:
            if(ch != '\n') ERROR(L"\\r must be followed by \\n");
            state = StateHeader;
            break;

        case StateHeader:
        case StateDone:
        case StateError:
            // should never get here!
            throw lw::ProgramException().chain(L"HTTPRequestParser::parse() [1]");
    }

    goto begin;
#undef ERROR
}



//END // HTTPRequestParser /////////////////////////////////////////////



}
