/* lw-support/src/lib/Net/Protocols/HTTP.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Parser for HTTP header fields.

This class is used to parse the HTTP header fields that occur after the
HTTP request or response line. It is a finite state machine parser that
only consumes as many characters as it needs (i.e. it stops at the final
\c "\r\n\r\n" sequence demanded by the standard). It exposes these
header fields as part of a map.

\warning This isn't 100% RFC-compliant: it won't handle multiple
    instances of the same field name in an RFC-compatible manner.

*/
class HTTPHeaderParser {
private:
    enum State {
        StateBegin,
        StateNewline,
        StateLine,
        StateFieldName,
        StateFieldValue,
        StateNewline2,
        StateDone,
        StateError
    }state;

    std::string fieldName;

public:
    /// Constructor.
    HTTPHeaderParser();

    /// Map of field name (key) to value.
    std::map<std::string, std::string> fields;



    /*! \brief Parse incoming data.

    \param data Pointer to character data.
    \param amt Number of bytes to read.
    \throws lw::HTTPMalformed if the HTTP is malformed in any way.
    \returns Number of bytes parsed. If it returns less than the number
        of bytes given, it has reached the end.

    */
    size_t parse(const char* data, size_t amt);

    /// Returns true if parsing is complete.
    bool completed() const { return state == StateDone; }

    /// Returns true if parsing failed.
    bool failed() const { return state == StateError; }
};



/*! \brief HTTP response parser.

This class parses HTTP responses. It uses an HTTPHeaderParser internally
to deal with the header fields, and has code for dealing with the first
line of the HTTP response.

\warning ASCII only; see also HTTPHeaderParser.

*/
class HTTPResponseParser {
private:
    HTTPHeaderParser hp;

    enum State {
        StateHTTPSeq,
        StateHTTPMajor1,
        StateHTTPMajor,
        StateHTTPMinor1,
        StateHTTPMinor,
        StateHTTPResp1,
        StateHTTPResp2,
        StateHTTPResp3,
        StateHTTPResp4,
        StateHTTPMsg,
        StateHTTPNewline,
        StateHeader,
        StateDone,
        StateError
    }state;

    static const char* HTTPSeq;
    const char* HTTPSeqPos;

public:
    /// Constructor.
    HTTPResponseParser();

    /// HTTP major version number.
    int versionMajor;

    /// HTTP minor version number.
    int versionMinor;

    /// HTTP 3-digit response code.
    int responseCode;

    /// HTTP status message.
    std::string status;

    /// HTTP header fields.
    std::map<std::string, std::string>& fields;

    /*! \brief Parse incoming data.

    \param data Pointer to character data.
    \param amt Number of bytes to read.
    \throws lw::HTTPMalformed if the HTTP is malformed in any way.
    \returns Number of bytes parsed. If it returns less than the number
        of bytes given, it has reached the end.

    */
    size_t parse(const char* data, size_t amt);

    /// Returns true if parsing is complete.
    bool completed() const { return state == StateDone; }

    /// Returns true if parsing failed.
    bool failed() const { return state == StateError; }
};



/*! \brief HTTP request parser.

This class parses HTTP requests. It uses an HTTPHeaderParser internally
to deal with the header fields, and has code for dealing with the first
line of the HTTP request.

\warning ASCII only; see also HTTPHeaderParser.
\warning URIs are simply copied verbatim and are not validated in any
    way whatsoever.

*/
class HTTPRequestParser {
private:
    HTTPHeaderParser hp;

    enum State {
        StateHTTPMethod1,
        StateHTTPMethod,
        StateHTTPUrl1,
        StateHTTPUrl,
        StateHTTPSeq,
        StateHTTPMajor1,
        StateHTTPMajor,
        StateHTTPMinor1,
        StateHTTPMinor,
        StateHTTPNewline,
        StateHeader,
        StateDone,
        StateError
    }state;

    static const char* HTTPSeq;
    const char* HTTPSeqPos;

public:
    /// Constructor.
    HTTPRequestParser();

    /// HTTP method (POST, GET, etc.).
    std::string method;

    /// URL
    std::string url;

    /// HTTP major version number.
    int versionMajor;

    /// HTTP minor version number.
    int versionMinor;

    /// HTTP header fields.
    std::map<std::string, std::string>& fields;

    /*! \brief Parse incoming data.

    \param data Pointer to character data.
    \param amt Number of bytes to read.
    \throws lw::HTTPMalformed if the HTTP is malformed in any way.
    \returns Number of bytes parsed. If it returns less than the number
        of bytes given, it has reached the end.

    */
    size_t parse(const char* data, size_t amt);

    /// Returns true if parsing is complete.
    bool completed() const { return state == StateDone; }

    /// Returns true if parsing failed.
    bool failed() const { return state == StateError; }
};



}
