/* lw-support/src/lib/Net/Protocols/SMTP.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



const char* const SMTPClient::dayName[8] = {
    "(?) ", "Mon ", "Tue ", "Wed ", "Thu ", "Fri ", "Sat ", "Sun "
};

const char* const SMTPClient::monthName[13] = {
    " (?) ", " Jan ", " Feb ", " Mar ", " Apr ", " May ", " Jun ",
    " Jul ", " Aug ", " Sep ", " Oct ", " Nov ", " Dec "
};



size_t SMTPClient::parserFeedData(const char* data, size_t amt)
{
    for(size_t used = 0; used < amt; ++used) {
        char ch = data[used];

        switch(parserState) {
            case ParserStateDone:
                throw std::wstring(L"Program logic error [1]");

            case ParserStateNone:
                if(!isdigit(ch)) throw std::wstring(L"Expecting SMTP reply code.");
                parserCode = 100 * (ch - '0');
                parserState = ParserStateCode1;
                break;

            case ParserStateCode1:
                if(!isdigit(ch)) throw std::wstring(L"Expecting SMTP reply code.");
                parserCode += 10 * (ch - '0');
                parserState = ParserStateCode2;
                break;

            case ParserStateCode2:
                if(!isdigit(ch)) throw std::wstring(L"Expecting SMTP reply code.");
                parserCode += (ch - '0');
                parserState = ParserStateSP;
                break;

            case ParserStateSP:
                if(isspace(ch)) parserState = ParserStateData;
                else if(ch == '-') parserState = ParserStateContData;
                else throw std::wstring(L"Expecting ' ' or '-' after SMTP reply code.");
                break;

            case ParserStateData:
                if(ch == '\n') { // bad server!
                    parserState = ParserStateDone;
                    return used + 1;
                }
                if(ch == '\r') {
                    parserState = ParserStateLF;
                    break;
                }
                parserData.append(&ch, 1);
                break;

            case ParserStateLF:
                if(ch != '\n') throw std::wstring(L"Expecting \\r\\n sequence.");
                parserState = ParserStateDone;
                return used + 1;

            case ParserStateContData:
                if(ch == '\n') parserState = ParserStateContCode0;
                else if(ch == '\r') parserState = ParserStateContLF;
                else parserData.append(&ch, 1);
                break;

            case ParserStateContLF:
                if(ch != '\n') throw std::wstring(L"Expecting \\r\\n sequence.");
                parserState = ParserStateContCode0;
                break;

            case ParserStateContCode0:
                if((parserCode / 100) != (ch - '0'))
                    throw std::wstring(L"Mismatched multi-line SMTP reply code.");
                parserState = ParserStateContCode1;
                break;

            case ParserStateContCode1:
                if(((parserCode % 100) / 10) != (ch - '0'))
                    throw std::wstring(L"Mismatched multi-line SMTP reply code.");
                parserState = ParserStateContCode2;
                break;

            case ParserStateContCode2:
                if((parserCode % 10) != (ch - '0'))
                    throw std::wstring(L"Mismatched multi-line SMTP reply code.");
                parserState = ParserStateSP;
                break;
        }
    }
    return amt;
}



void SMTPClient::parserReset()
{
    parserState = ParserStateNone;
    parserCode = 0;
    parserData.clear();
}



bool SMTPClient::ioReady(uint32_t flags)
{
    if(flags & (IOEventUrgent | IOEventHUP)) {
        state = StateDone;
        logFail(L"Urgent data received / remote hangup");
        return IOListenResultDelete;
    }

    try {
        if(flags & IOEventError) netClient.checkErrors();

        while(tryRead() | tryWrite())
            ;
    }
    catch(Exception& e) {
        logFail(e);
        delete this;
        return false;
    }
    catch(std::wstring& e) {
        logFail(e);
        delete this;
        return false;
    }

    if(state == StateDone) {
        delete this;
        return false;
    }
    return true;
}



bool SMTPClient::tryRead() // returns true if I/O happens
{
    bool ioHappened = false;
    while(true) {
        size_t amt = inBuffer.getRemaining();
        amt = netClient.read(inBuffer.getBuffer(amt), amt);
        inBuffer.addedData(amt);
        if(!amt) return ioHappened;
        ioHappened = true;
        logRead(inBuffer.getData(), inBuffer.getSize());

        while(!inBuffer.isEmpty()) {
            // parse data
            inBuffer.getData(parserFeedData(inBuffer.getData(),
                inBuffer.getSize()));
            if(parserState != ParserStateDone) break;

#define FAIL() do { \
    logFail(parserCode, parserData); \
    state = StateDone; \
    return true; \
}while(0)

#define WRITE(x) do { \
    writePtr = x.c_str(); \
    writeAmt = x.size(); \
}while(0)

            switch(state) {
                case StateDone:
                    throw std::wstring(L"Program logic error [2].");

                case StateNone:
                    if((parserCode / 10) != 22) FAIL();
                    else {
                        state = StateHelo;
                        outBuffer = "helo ";
                        outBuffer.append(hostName);
                        outBuffer.append("\r\n");
                        WRITE(outBuffer);
                    }
                    break;

                case StateHelo:
                    if((parserCode / 10) != 25) FAIL();
                    else {
                        state = StateMailFrom;
                        outBuffer = "mail from: ";
                        outBuffer.append(mailFrom);
                        outBuffer.append("\r\n");
                        WRITE(outBuffer);
                    }
                    break;

                case StateMailFrom:
                    if((parserCode / 10) != 25) FAIL();
                    else {
                        state = StateRcptTo;
                        outBuffer = "rcpt to: ";
                        outBuffer.append(rcptTo);
                        outBuffer.append("\r\n");
                        WRITE(outBuffer);
                    }
                    break;

                case StateRcptTo:
                    if((parserCode / 10) != 25) FAIL();
                    else {
                        state = StateData;
                        outBuffer = "data\r\n";
                        WRITE(outBuffer);
                    }
                    break;

                case StateData:
                    if((parserCode / 10) != 35) FAIL();
                    else {
                        state = StateQuit;
                        WRITE(data);
                    }
                    break;

                case StateQuit:
                    if((parserCode / 10) != 25) FAIL();
                    else {
                        state = StateFinal;
                        outBuffer = "quit\r\n";
                        WRITE(outBuffer);
                    }
                    break;

                case StateFinal:
                    if((parserCode / 10) != 22) FAIL();
                    else {
                        state = StateDone;
                        logDone();
                        return true;
                    }
                    break;
            }

#undef WRITE
#undef FAIL

            // reset parser and go again
            parserReset();
        }
    }
}



bool SMTPClient::tryWrite()
{
    bool ioHappened = false;
    if(!writePtr) return false;

    while(writeAmt) {
        size_t amt = netClient.write(writePtr, writeAmt);
        if(!amt) return ioHappened;
        ioHappened = true;

        logWrite(writePtr, amt);
        writeAmt -= amt;
        writePtr += amt;
    }

    writePtr = 0;
    return true;
}



bool SMTPClient::isWhitespace(char ch)
{
    return (ch >= 0x9 && ch <= 0xD) || ch == 0x20;
}



std::string SMTPClient::wsFieldName(const std::string& fieldName)
{
    // strip whitespace from beginning and end
    std::string::size_type pos, len = fieldName.length();
    for(pos = 0; pos < len; ++pos) if(!isWhitespace(fieldName[pos])) break;
    if(pos == len) return std::string();

    std::wstring::size_type endPos = len - 1;
    while(isWhitespace(fieldName[endPos])) --endPos;

    std::string str = fieldName.substr(pos, endPos - pos + 1);

    // return empty string if there are any illegal chars, and convert
    // to lowercase
    for(pos = 0, len = str.length(); pos < len; ++pos) {
        if(str[pos] <= 0x20 || str[pos] & ~0x7F) return std::string();
        str[pos] = tolower(str[pos]);
    }

    // done
    return str;
}



std::string SMTPClient::wsFieldValue(const std::string& fieldValue)
{
    // strip whitespace from beginning and end
    std::string::size_type pos, len = fieldValue.length();
    for(pos = 0; pos < len; ++pos) if(!isWhitespace(fieldValue[pos])) break;
    if(pos == len) return std::string();

    std::wstring::size_type endPos = len - 1;
    while(isWhitespace(fieldValue[endPos])) --endPos;

    std::string str = fieldValue.substr(pos, endPos - pos + 1);

    // convert remaining whitespace
    for(pos = 0, len = str.length(); pos < len; ++pos)
        if(str[pos] == '\r' || str[pos] == '\n')
            str[pos] = ' ';

    // done
    return str;
}



std::string SMTPClient::buildRFC2822(const std::string& hostName,
    const std::string& mailFrom, const std::string& rcptTo,
    const std::multimap<std::string, std::string>& headerLines,
    const std::string& message)
{
    (void)hostName; // TODO: add message-id
    std::ostringstream o;

    // first, deal with headers (sanitise as we go)
    bool seenDate = false, seenFrom = false, seenTo = false;

    std::multimap<std::string, std::string>::const_iterator
        iter = headerLines.begin(), end = headerLines.end();
    for(; iter != end; ++iter) {
        std::string fieldName = wsFieldName(iter->first);
        std::string fieldValue = wsFieldValue(iter->second);

        if(fieldName.empty()) continue;
        if(fieldName == "date") seenDate = true;
        if(fieldName == "from") seenFrom = true;
        if(fieldName == "to") seenTo = true;
        if(fieldValue.empty()) continue;

        o << fieldName << ": " << fieldValue << "\r\n";
    }

    if(!seenDate) {
        lw::DateTime date = lw::DateTime::now();
        o << std::setfill('0')
          << "Date: "
          << dayName[date.getDate().getWeekDay()]
          << date.getDate().getDay()
          << monthName[date.getDate().getMonth()]
          << date.getDate().getYear()
          << ' '
          << std::setw(2) << date.getTime().h() << ':'
          << std::setw(2) << date.getTime().m() << ':'
          << std::setw(2) << date.getTime().s()
          << " +0000" // FIXME
          << "\r\n"
          << std::setfill(' ');
    }
    if(!seenFrom) o << "From: " << wsFieldValue(mailFrom) << "\r\n";
    if(!seenTo) o << "To: " << wsFieldValue(rcptTo) << "\r\n";

    o << "\r\n";

    // now output the escaped message
    bool wasNewline = true;
    for(std::string::size_type pos = 0, len = message.size(); pos < len; ++pos) {
        char ch = message[pos];
        switch(ch) {
            case '\r':
                ch = ' ';
                break;

            case '\n':
                o << "\r\n";
                wasNewline = true;
                continue;

            case '.':
                if(wasNewline) o.put('.');
                break;
        }
        o.put(ch);
        wasNewline = false;
    }

    o << "\r\n.\r\n";

    // done
    return o.str();
}



SMTPClient::SMTPClient(const NetAddress& netAddress, uint16_t port,
    EventManager& eventManager, const std::string& hostName,
    const std::string& mailFrom, const std::string& rcptTo,
    const std::multimap<std::string, std::string>& headerLines,
    const std::string& message, CompletionNotifier* completionNotifier)
  : completionNotifier(completionNotifier),
    hostName(wsFieldValue(hostName)),
    mailFrom(wsFieldValue(mailFrom)),
    rcptTo(wsFieldValue(rcptTo)),
    data(buildRFC2822(hostName, mailFrom, rcptTo, headerLines, message)),
    state(StateNone), inBuffer(1024, 1024), writePtr(0), writeAmt(0),
    parserState(ParserStateNone), parserCode(0),
    port(port), netAddress(netAddress.clone())
{
    // NOTE: we can't use any of the log* functions in the constructor,
    // because the virtual functions won't be pointing to the derived
    // class yet. We'll have to wait until we're in the I/O callback to
    // log anything.

    // set up client
    netClient.connect(netAddress, port);
    eventManager.registerDevice(&netClient, this, ~0);
}



SMTPClient::~SMTPClient()
{
    if(completionNotifier) completionNotifier->signal();
    delete netAddress;
}



void SMTPClient::sendMail(const NetAddress& netAddress, uint16_t port,
    EventManager& eventManager, const std::string& hostName,
    const std::string& mailFrom, const std::string& rcptTo,
    const std::multimap<std::string, std::string>& headerLines,
    const std::string& message, CompletionNotifier* completionNotifier)
{
    new SMTPClient(netAddress, port, eventManager, hostName,
        mailFrom, rcptTo, headerLines, message, completionNotifier);
}



void SMTPClient::logStart(const std::string&,
    const std::string&, const std::string&,
    const std::string&)
{
}



void SMTPClient::logDone()
{
}



void SMTPClient::logRead(const char*, size_t)
{
}



void SMTPClient::logWrite(const char*, size_t)
{
}



void SMTPClient::logFail(int, const std::string&)
{
}



void SMTPClient::logFail(Exception&)
{
}



void SMTPClient::logFail(const std::wstring&)
{
}



int LoggingSMTPClient::instanceCount = 0;



LoggingSMTPClient::LoggingSMTPClient(const NetAddress& netAddress,
    uint16_t port,
    EventManager& eventManager, const std::string& hostName,
    const std::string& mailFrom, const std::string& rcptTo,
    const std::multimap<std::string, std::string>& headerLines,
    const std::string& message, Log& logger, Log& logError,
    CompletionNotifier* completionNotifier)
  : SMTPClient(netAddress, port, eventManager, hostName, mailFrom, rcptTo,
               headerLines, message, completionNotifier),
    logger(logger), logError(logError)
{
    std::wostringstream o;
    o << L"SMTP <"
      << ++instanceCount
      << L">, "
      << netAddress.toString()
      << L" port "
      << port;
    logID = o.str();
}



LoggingSMTPClient::~LoggingSMTPClient()
{
}



void LoggingSMTPClient::sendMail(const NetAddress& netAddress,
    uint16_t port,
    EventManager& eventManager, const std::string& hostName,
    const std::string& mailFrom, const std::string& rcptTo,
    const std::multimap<std::string, std::string>& headerLines,
    const std::string& message, Log& logger, Log& logError,
    CompletionNotifier* completionNotifier)
{
    new LoggingSMTPClient(netAddress, port, eventManager, hostName,
        mailFrom, rcptTo, headerLines, message,
        logger, logError, completionNotifier);
}



void LoggingSMTPClient::logStart(const std::string& hostName,
    const std::string& mailFrom, const std::string& rcptTo,
    const std::string& data)
{
    logger << Log::start(logID)
           << L"Sending a new email:"
           << L"\n  Host: " << hostName
           << L"\n  From: " << mailFrom
           << L"\n  To  : " << rcptTo
           << L"\n\n"
           << Log::asciiDump(data.c_str(), data.size())
           << Log::end;
}



void LoggingSMTPClient::logDone()
{
    logger << Log::start(logID)
           << L"Email sent successfully."
           << Log::end;
}



void LoggingSMTPClient::logRead(const char* data, size_t amt)
{
    logger << Log::start(logID)
           << L"Incoming data.\n\n"
           << Log::asciiDump(data, amt)
           << Log::end;
}



void LoggingSMTPClient::logWrite(const char* data, size_t amt)
{
    logger << Log::start(logID)
           << L"Wrote data.\n\n"
           << Log::asciiDump(data, amt)
           << Log::end;
}



void LoggingSMTPClient::logFail(int smtpCode, const std::string& errorString)
{
    logError << Log::start(logID)
             << L"SMTP error."
                L"\n  Code  : " << smtpCode
             << L"\n  Reason: " << errorString
             << Log::end;
}



void LoggingSMTPClient::logFail(Exception& e)
{
    logError << Log::start(logID)
             << L"System error.\n"
             << e
             << Log::end;
}



void LoggingSMTPClient::logFail(const std::wstring& s)
{
    logError << Log::start(logID)
             << L"Communication failed: "
             << s
             << Log::end;
}



}
