/* lw-support/src/lib/Net/Protocols/SMTP.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief An SMTP (email) client class.

This class can be used to perform an SMTP transaction. It will send one
email. It is used in an asynchronous manner, and can (optionally) signal
completion through a CompletionNotifier object.

\warning You should only send ASCII characters through this client, but
    it does not perform any checking to enforce this.

\todo Timeout options.
\todo Split SMTP response parser into its own class.

*/
class SMTPClient : private EventCallbackIO {
private:
    // for performing I/O
    CompletionNotifier* completionNotifier;
    NetClientTCP netClient;
    virtual bool ioReady(uint32_t flags);
    bool tryRead();
    bool tryWrite();

    // data
    const std::string hostName, mailFrom, rcptTo, data;

    // builds an RFC2822-compliant message
    static const char* const dayName[8];
    static const char* const monthName[13];
    static std::string buildRFC2822(const std::string& hostName,
        const std::string& mailFrom, const std::string& rcptTo,
        const std::multimap<std::string, std::string>& headerLines,
        const std::string& message);

    static bool isWhitespace(char ch);
    static std::string wsFieldName(const std::string& fieldName);
    static std::string wsFieldValue(const std::string& fieldValue);

    // transaction state
    enum State {
        StateNone,
        StateHelo,
        StateMailFrom,
        StateRcptTo,
        StateData,
        StateQuit,
        StateFinal,
        StateDone
    }state;
    FIFO inBuffer;
    std::string outBuffer;
    const char* writePtr;
    size_t writeAmt;

    // RFC 2821 parser
    size_t parserFeedData(const char* data, size_t amt);
    void parserReset();
    enum ParserState {
        ParserStateNone,
        ParserStateCode1,
        ParserStateCode2,
        ParserStateSP,
        ParserStateData,
        ParserStateLF,
        ParserStateContData,
        ParserStateContLF,
        ParserStateContCode0,
        ParserStateContCode1,
        ParserStateContCode2,
        ParserStateDone
    }parserState;
    int parserCode;
    std::string parserData;

protected:
    /// TCP port to connect to.
    uint16_t port;

    /// Host to connect to.
    NetAddress* netAddress;

    /// Protected constructor for "fire-and-forget".
    SMTPClient(const NetAddress& netAddress, uint16_t port,
        EventManager& eventManager, const std::string& hostName,
        const std::string& mailFrom, const std::string& rcptTo,
        const std::multimap<std::string, std::string>& headerLines,
        const std::string& message, CompletionNotifier* completionNotifier);

    /// Log the start of a new message.
    virtual void logStart(const std::string& hostName,
        const std::string& mailFrom, const std::string& rcptTo,
        const std::string& data);

    /// Log message complete.
    virtual void logDone();

    /// Log a data transfer (input).
    virtual void logRead(const char* data, size_t amt);

    /// Log a data transfer (output).
    virtual void logWrite(const char* data, size_t amt);

    /// Log an SMTP error.
    virtual void logFail(int smtpCode, const std::string& errorString);

    /// Log a connection error.
    virtual void logFail(Exception& e);

    /// Log a communications error.
    virtual void logFail(const std::wstring& s);

public:
    /// Destructor.
    virtual ~SMTPClient();

    /*! \brief Begins sending email.

    \param netAddress The address of the host to connect to.
    \param port The TCP port number of the SMTP service.
    \param eventManager The event manager to use.
    \param hostName The name of the host to present to the SMTP server.
    \param mailFrom Email address of the (SMTP envelope) sender.
    \param rcptTo Email address of the (SMTP envelope) receiver.
    \param headerLines The header lines to use in the email. These will
        be properly escaped, whitespace stripped, etc.
    \param message The actual message to send. It will be properly
        escaped.
    \param completionNotifier The CompletionNotifier object to use to
        signal when the process is complete. May be left as 0 in which
        case no signal will be issued.
    \throws Exception if a system error occurs.

    The constructor will create a properly-formatted email, and begin
    connecting to the SMTP server.

    Several standard header lines will be automatically filled out for
    you. You may override this by providing values for them in the
    \a headerLines parameter. These headers are:
     - \a orig-date (set to current date).
     - \a from (set to the SMTP envelope \a mailFrom).
     - \a to (set to the SMTP envelope \a rcptTo).

    Blank value strings in the \a headerLines array will cause the
    corresponding header line \e not to be generated. Blank key strings
    will be ignored. Whitespace will be stripped from all key strings.
    Key strings are case insensitive.

    Rules for whitespace handling are:
     - all ASCII CR characters are replaced with an ASCII space;
     - all ASCII LF characters in header fields are replaced with an
       ASCII space;
     - any remaining whitespace in field names causes the field to be
       skipped;
     - otherwise, whitespace is left verbatim.

    */
    static void sendMail(const NetAddress& netAddress, uint16_t port,
        EventManager& eventManager, const std::string& hostName,
        const std::string& mailFrom, const std::string& rcptTo,
        const std::multimap<std::string, std::string>& headerLines,
        const std::string& message, CompletionNotifier* completionNotifier);
};



/*! \brief A logging SMTP (email) client class.

This class can be used to perform an SMTP transaction. It will send one
email. It is used in a "fire and forget" manner, and there is no way to
determine the outcome of sending an email (except through the logging
functions).

This class will log its actions to the specified log objects.

\warning You should only send ASCII characters through this client, but
    it does not perform any checking to enforce this.

*/
class LoggingSMTPClient : public SMTPClient {
private:
    static int instanceCount;

protected:
    /// This email's unique ID, for logging.
    std::wstring logID;

    /// Log object.
    Log& logger, & logError;

    /// Protected constructor for "fire-and-forget".
    LoggingSMTPClient(const NetAddress& netAddress, uint16_t port,
        EventManager& eventManager, const std::string& hostName,
        const std::string& mailFrom, const std::string& rcptTo,
        const std::multimap<std::string, std::string>& headerLines,
        const std::string& message, Log& logger, Log& logError,
        CompletionNotifier* completionNotifier);

    /// Log the start of a new message.
    virtual void logStart(const std::string& hostName,
        const std::string& mailFrom, const std::string& rcptTo,
        const std::string& data);

    /// Log message complete.
    virtual void logDone();

    /// Log a data transfer (input).
    virtual void logRead(const char* data, size_t amt);

    /// Log a data transfer (output).
    virtual void logWrite(const char* data, size_t amt);

    /// Log an SMTP error.
    virtual void logFail(int smtpCode, const std::string& errorString);

    /// Log a connection error.
    virtual void logFail(Exception& e);

    /// Log a communications error.
    virtual void logFail(const std::wstring& s);

public:
    /// Destructor.
    virtual ~LoggingSMTPClient();

    /*! \brief Begins sending email, with logging.

    \param netAddress The address of the host to connect to.
    \param port The TCP port number of the SMTP service.
    \param eventManager The event manager to use.
    \param hostName The name of the host to present to the SMTP server.
    \param mailFrom Email address of the (SMTP envelope) sender.
    \param rcptTo Email address of the (SMTP envelope) receiver.
    \param headerLines The header lines to use in the email. These will
        be properly escaped, whitespace stripped, etc.
    \param message The actual message to send. It will be properly
        escaped.
    \param logger The lw::Log object to use.
    \param logError The lw::Log object to use to log errors.
    \param completionNotifier The CompletionNotifier object to use to
        signal when the process is complete. May be left as 0 in which
        case no signal will be issued.
    \throws Exception if a system error occurs.

    The constructor will create a properly-formatted email, and begin
    connecting to the SMTP server.

    Several standard header lines will be automatically filled out for
    you. You may override this by providing values for them in the
    \a headerLines parameter. These headers are:
     - \a orig-date (set to current date).
     - \a from (set to the SMTP envelope \a mailFrom).
     - \a to (set to the SMTP envelope \a rcptTo).

    Blank value strings in the \a headerLines array will cause the
    corresponding header line \e not to be generated. Blank key strings
    will be ignored. Whitespace will be stripped from all key strings.
    Key strings are case insensitive.

    Rules for whitespace handling are:
     - all ASCII CR characters are replaced with an ASCII space;
     - all ASCII LF characters in header fields are replaced with an
       ASCII space;
     - any remaining whitespace in field names causes the field to be
       skipped;
     - otherwise, whitespace is left verbatim.

    */
    static void sendMail(const NetAddress& netAddress, uint16_t port,
        EventManager& eventManager, const std::string& hostName,
        const std::string& mailFrom, const std::string& rcptTo,
        const std::multimap<std::string, std::string>& headerLines,
        const std::string& message, Log& logger, Log& logError,
        CompletionNotifier* completionNotifier);
};



}
