/* lw-support/src/lib/Net/Server.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Base class for all network servers.

This is the base class for any network server. It provides two main
functions: you can call accept() (non-blocking) to accept any waiting
connection; or you can register a callback with an EventManager. Your
callback will get error and new connection events.

*/
class NetServer
  : protected IOPosixDevice,
    private EventCallbackIO
{
private:
    // callback stuff
    EventCallbackNetServer* nsCallback;
    EventManager* eventManager;
	virtual void ioReady(uint32_t flags) throw(); // from EventCallbackIO()

    //BEGIN // Protocol-specific implementation ////////////////////////

public:
    /// Exception used for an unknown protocol.
    class UnknownProtocol : public ProgramException { };

protected:
    /// Enumeration representing understood protocols.
    enum Protocol {
        IPv4,
        IPv6
    }protocol;

    /*! \brief Get an address's protocol type.

    \param addr The address to inquire about.
    \throws UnknownProtocol if the protocol is entirely unknown.
    \returns The type of the address.

    This function can be used to ask what type of protocol an address
    refers to. This isn't a virtual function in NetAddress because all
    the behaviour changing functionality is (for the time being at
    least) going to be in this class.

    */
    static Protocol getProtocol(const NetAddress& addr);

    /*! \brief Block until there's a connection available.

    \param timeout Timeout period, in milliseconds. 0 means immediate;
        negative means forever.
    \param timeoutMode The timeout mode.
    \throws SystemError if a system error occurs.
    \throws IOTimeout if a timeout occurs.
    \returns A newly-allocated I/O device object.

    This function can be called from within acceptBlocking(). Its purpose is to aid the
    implementation of that function. It will wait until there is data available to read on the
    associated fd.

    */
    void readBlock(int timeout = -1, IOTimeoutMode timeoutMode = IOTimeoutHardFull);



public:
    //BEGIN // Constructors etc. ///////////////////////////////////////

    /// Constructor. Does nothing.
    NetServer();

    /// Destructor.
    virtual ~NetServer();



    //BEGIN // Callback system /////////////////////////////////////////

    /*! \brief Called to retrieve a connection (non-blocking).

    \throws SystemError if a system error occurs.
    \returns A newly-allocated I/O device object.
    \retval 0 if there is no connection waiting.

    This function is called whenever you want to accept a new
    connection. It either creates a new I/O device object or returns 0
    if there are no connections waiting.

    */
    virtual lw::NetClient* accept() = 0;



    /*! \brief Called to retrieve a connection (blocking).

    \param timeout Timeout period, in milliseconds. 0 means immediate;
        negative means forever.
    \param timeoutMode The timeout mode.
    \throws SystemError if a system error occurs.
    \throws IOTimeout if a timeout occurs.
    \returns A newly-allocated I/O device object.

    This function is called whenever you want to accept a new connection. It creates a new I/O
    device object.

    */
    virtual lw::NetClient* acceptBlocking(int timeout = -1,
                                          IOTimeoutMode timeoutMode = IOTimeoutHardFull) = 0;



    /*! \brief Registers object with an event manager.

    \param eventManager The event manager object.
    \param callback The NetServerCallback object that will receive
        events for this server.
    \throws SystemError if a system error occurs.

    This function will register your callback object with an event
    manager. Only one callback can be active at once; calling this
    function with a callback already set will simply clear the old
    callback.

    */
    void registerCallback(EventManager& eventManager,
        EventCallbackNetServer& callback);



    /*! \brief Stops callbacks.

    \throws SystemError if a system error occurs.

    If you want to stop generating callback events, call this function.
    Alternatively, you can return \a false in the callback

    */
    void clearCallback();
};



}

/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
*/
