/* lw-support/src/lib/Net/TCP/Server.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



void NetServerTCP::listen(const NetAddress& addr,
    uint16_t port, int queueLen)
{
    try {
        // shut down any previous server
        stop();

        // create a socket
        fd = TEMP_FAILURE_RETRY(socket(PF_INET, SOCK_STREAM, IPPROTO_TCP));
        if(fd == -1)
            throw SystemError().chain(L"socket()");

        // set SO_REUSEADDR, so that we can bind to a socket even if
        // it's in the TCP TIME_WAIT state.
        int reuse = 1;
        if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)))
            throw SystemError().chain(L"setsockopt(..., SOL_SOCKET, SO_REUSEADDR, ...)");

        // bind that socket to the requested address
        switch(protocol = getProtocol(addr)) {
            case IPv4: {
                const NetAddressIPv4& a = dynamic_cast<const NetAddressIPv4&>(addr);
                struct sockaddr_in sa;
                if(TEMP_FAILURE_RETRY(bind(fd, a.copy(&sa, port), sizeof(sa))))
                    throw SystemError().chain(L"bind()");
                break;
            }

            case IPv6: {
                const NetAddressIPv6& a = dynamic_cast<const NetAddressIPv6&>(addr);
                struct sockaddr_in6 sa;
                if(TEMP_FAILURE_RETRY(bind(fd, a.copy(&sa, port), sizeof(sa))))
                    throw SystemError().chain(L"bind()");
                break;
            }
        }

        // start listening
        if(TEMP_FAILURE_RETRY(::listen(fd, queueLen)))
            throw SystemError().chain(L"listen()");

        // set non-blocking I/O mode
        setNonBlocking();
        ioMode = IONone;
    }

    // gracefully deal with any errors
    catch(Exception& e) {
        try {
            stop();
        }
        catch(...) { }

        e.chain(L"NetServerTCP::listen()");
        throw;
    }
}



void NetServerTCP::stop()
{
    try {
        close();
    }
    catch(Exception& e) {
        e.chain(L"NetServerTCP::stop()");
        throw;
    }
}



NetClientTCP* NetServerTCP::accept()
{
    try {
        int sockfd = -1;

        switch(protocol) {
            case IPv4: {
                while(true) {
                    // get connection and address
                    struct sockaddr_in sa, sa2;
                    socklen_t sa_len = sizeof(sa);
                    sockfd = TEMP_FAILURE_RETRY(::accept(
                        fd, (struct sockaddr*)(&sa), &sa_len));

                    // return 0 if there was nothing waiting
                    if(sockfd == -1) {
                        // see accept(2) man page: some errors are propagated
                        // from elsewhere in the TCP stack and we should
                        // handle them by retrying accept().
                        switch(errno) {
                            case EAGAIN:
                                return 0;

                            case ENETDOWN:
                            case EPROTO:
                            case ENOPROTOOPT:
                            case EHOSTDOWN:
                            case ENONET:
                            case EHOSTUNREACH:
                            case EOPNOTSUPP:
                            case ENETUNREACH:
                                continue;

                            default:
                                throw SystemError().chain(L"accept()");
                        }
                    }

                    // get local address
                    sa_len = sizeof(sa2);
                    if(TEMP_FAILURE_RETRY(getsockname(
                        sockfd, (struct sockaddr*)(&sa2), &sa_len)))
                            throw SystemError().chain(L"getsockname()");

                    // build client
                    return new NetClientTCP(sockfd,
                        new NetAddressIPv4((char*)&(sa2.sin_addr.s_addr)), ntohs(sa2.sin_port),
                        new NetAddressIPv4((char*)&(sa.sin_addr.s_addr)), ntohs(sa.sin_port));
                }
            }

            case IPv6: {
                while(true) {
                    // get connection and address
                    struct sockaddr_in6 sa, sa2;
                    socklen_t sa_len = sizeof(sa);
                    sockfd = TEMP_FAILURE_RETRY(::accept(
                        fd, (struct sockaddr*)(&sa), &sa_len));

                    // return 0 if there was nothing waiting
                    if(sockfd == -1) {
                        // see accept(2) man page: some errors are propagated
                        // from elsewhere in the TCP stack and we should
                        // handle them by retrying accept().
                        //
                        // TODO: is this right for IPv6 as well?
                        switch(errno) {
                            case EAGAIN:
                                return 0;

                            case ENETDOWN:
                            case EPROTO:
                            case ENOPROTOOPT:
                            case EHOSTDOWN:
                            case ENONET:
                            case EHOSTUNREACH:
                            case EOPNOTSUPP:
                            case ENETUNREACH:
                                continue;

                            default:
                                throw SystemError().chain(L"accept()");
                        }
                    }

                    // get local address
                    sa_len = sizeof(sa2);
                    if(TEMP_FAILURE_RETRY(getsockname(
                        sockfd, (struct sockaddr*)(&sa2), &sa_len)))
                            throw SystemError().chain(L"getsockname()");

                    // build client
                    return new NetClientTCP(sockfd,
                        new NetAddressIPv6(sa2.sin6_addr.s6_addr), ntohs(sa2.sin6_port),
                        new NetAddressIPv6(sa.sin6_addr.s6_addr), ntohs(sa.sin6_port));
                }
            }
        }

        throw ProgramException().chain(L"NetServerTCP::connect()");
    }
    catch(Exception& e) {
        e.chain(L"NetServerTCP::accept()");
        throw;
    }
}



NetClientTCP* NetServerTCP::acceptBlocking(int timeout, IOTimeoutMode timeoutMode)
{
    readBlock(timeout, timeoutMode);
    NetClientTCP* client = accept();

    // TODO: better error handling here would be appropriate... restart timeout somehow
    if(!client) throw ProgramException().chain(L"NetServerTCP::acceptBlocking()");
    return client;
}



}
