/* lw-support/src/lib/String/basics.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



std::wstring stripWhitespace(const std::wstring& str)
{
    std::wstring::size_type pos, len = str.length();
    for(pos = 0; pos < len; ++pos) if(!isWhitespace(str[pos])) break;
    if(pos == len) return std::wstring();

    std::wstring::size_type endPos = len - 1;
    while(isWhitespace(str[endPos])) --endPos;

    return str.substr(pos, endPos - pos + 1);
}



std::string ucs4ToUtf8(const std::wstring& ucs4, bool force)
{
    // compute size of string, skip empty strings
    std::wstring::size_type len = ucs4.length(), rd_pos, start;
    if(!len) return std::string();

    // skip BOM
    start = (ucs4[0] == 0xFEFF) ? 1 : 0;

    // compute required size and check characters
    int dlen = 0;
    for(rd_pos = start; rd_pos < len; ++rd_pos) {
        wchar_t ch = ucs4[rd_pos];
        if(!force && (ch & ~0x7FFFFFFF)) {
            throw BadSourceChar((const char*)&ch, (const char*)(&ch + 4),
                rd_pos * 4, L"UCS-4", L"Character out of range.");
        }

        if(ch < 0x80) ++dlen;
        else if(ch < 0x800) dlen += 2;
        else if(ch < 0x10000) dlen += 3;
        else if(ch < 0x200000) dlen += 4;
        else if(ch < 0x4000000) dlen += 5;
        else dlen += 6;
    }

    // allocate memory
    std::string dest(dlen, 0);

    // convert string
    std::string::size_type wr_pos = 0;
    for(rd_pos = start; rd_pos < len; ++rd_pos) {
        uint32_t ch = ucs4[rd_pos];

        if(ch & ~0x7FFFFFFF) continue;

        if(ch < 0x80) {
            dest[wr_pos++] = ch;

        } else if(ch < 0x800) {
            dest[wr_pos++] = 0xC0 | ((ch >> 6) & 0x1F);
            dest[wr_pos++] = 0x80 | (ch & 0x3F);

        } else if(ch < 0x10000) {
            dest[wr_pos++] = 0xE0 | ((ch >> 12) & 0xF);
            dest[wr_pos++] = 0x80 | ((ch >> 6) & 0x3F);
            dest[wr_pos++] = 0x80 | (ch & 0x3F);

        } else if(ch < 0x200000) {
            dest[wr_pos++] = 0xF0 | ((ch >> 18) & 0x7);
            dest[wr_pos++] = 0x80 | ((ch >> 12) & 0x3F);
            dest[wr_pos++] = 0x80 | ((ch >> 6) & 0x3F);
            dest[wr_pos++] = 0x80 | (ch & 0x3F);

        } else if(ch < 0x4000000) {
            dest[wr_pos++] = 0xF8 | ((ch >> 24) & 0x3);
            dest[wr_pos++] = 0x80 | ((ch >> 18) & 0x3F);
            dest[wr_pos++] = 0x80 | ((ch >> 12) & 0x3F);
            dest[wr_pos++] = 0x80 | ((ch >> 6) & 0x3F);
            dest[wr_pos++] = 0x80 | (ch & 0x3F);

        } else {
            dest[wr_pos++] = 0xFC | ((ch >> 30) & 0x1);
            dest[wr_pos++] = 0x80 | ((ch >> 24) & 0x3F);
            dest[wr_pos++] = 0x80 | ((ch >> 18) & 0x3F);
            dest[wr_pos++] = 0x80 | ((ch >> 12) & 0x3F);
            dest[wr_pos++] = 0x80 | ((ch >> 6) & 0x3F);
            dest[wr_pos++] = 0x80 | (ch & 0x3F);

        }
    }

    return dest;
}



std::string ucs4ToAscii(const std::wstring& ucs4, bool force)
{
    // compute size of string, skip empty strings
    std::wstring::size_type len = ucs4.length(), rd_pos, start;
    if(!len) return std::string();

    // skip BOM
    start = (ucs4[0] == 0xFEFF) ? 1 : 0;

    // allocate destination string
    int skipped = 0;
    std::string ascii(len - start, 0);

    // convert each char, checking values as we go
    for(rd_pos = start; rd_pos < len; ++rd_pos) {
        if(ucs4[rd_pos] & ~0x7F) {
            ++skipped;
            if(force) continue;
            throw CannotConvertChar(ucs4[rd_pos], rd_pos, L"UCS-4", L"ASCII");
        }

        ascii[rd_pos] = ucs4[rd_pos];
    }

    // strip off any skipped characters
    if(skipped) ascii.resize(len - skipped);

    // done
    return ascii;
}



namespace {

void utf8AuxConsume(wchar_t& ch, const std::string& utf8,
    std::string::size_type& rd_pos, std::string::size_type ch_start)
{
    uint8_t p = utf8[rd_pos++];
    if((p & 0xC0) != 0x80) {
        throw BadSourceChar(utf8.c_str() + ch_start,
            utf8.c_str() + rd_pos, ch_start, L"UTF-8",
            L"Invalid character sequence (expecting byte 10xxxxxx).");
    }

    ch <<= 6;
    ch |= p & 0x3F;
}

std::wstring utf8ToUcs4Forced(const std::string& str)
{
    // compute size of std::string, skip empty std::strings
    std::string::size_type len = str.length(), rd_pos = 0;
    if(!len) return std::wstring();

    // skip possible BOM
    if(len >= 3 && str.substr(0, 3) == "\xEF\xBB\xBF") rd_pos = 3;

    // for FSM
    enum CharType {
        CTypeAscii,
        CTypeContinue,
        CType2,
        CType3,
        CType4,
        CType5,
        CType6
    }ctype;
    int cremain = 0;

    // destination buffer
    std::wstring d;

    // begin conversion using FSM
    uint8_t ch;
    wchar_t dch;

    while(rd_pos < len) {
        ch = str[rd_pos++];

        // classify
        if(ch & 0x80) {
            if((ch & 0xC0) == 0x80) ctype = CTypeContinue;
            else if((ch & 0xE0) == 0xC0) ctype = CType2;
            else if((ch & 0xF0) == 0xE0) ctype = CType3;
            else if((ch & 0xF8) == 0xF0) ctype = CType4;
            else if((ch & 0xFC) == 0xF8) ctype = CType5;
            else if((ch & 0xFE) == 0xFC) ctype = CType6;
            else {
                continue; // must be 0xFE or 0xFF, both invalid
            }
        } else {
            ctype = CTypeAscii;
        }

        // decode multi-byte
        if(cremain) {
            if(ctype != CTypeContinue) {
                --rd_pos;
                cremain = 0;  // was expecting multi-byte; abort conversion
                continue;     // of this char and restart
            }
            dch <<= 6;
            dch |= ch & 0x3F;
            if(!--cremain) d.append(&dch, 1);

        // decode start char
        } else {
            switch(ctype) {
                case CTypeContinue:
                    // not valid here
                    continue;

                case CTypeAscii:
                    dch = ch;
                    d.append(&dch, 1);
                    continue;

                case CType2:
                    cremain = 1;
                    dch = ch & 0x1F;
                    continue;

                case CType3:
                    cremain = 2;
                    dch = ch & 0x0F;
                    continue;

                case CType4:
                    cremain = 3;
                    dch = ch & 0x07;
                    continue;

                case CType5:
                    cremain = 4;
                    dch = ch & 0x03;
                    continue;

                case CType6:
                    cremain = 5;
                    dch = ch & 0x01;
                    continue;
            }
        }
    }

    // done
    return d;
}

}



std::wstring utf8ToUcs4(const std::string& utf8, bool force)
{
    // when conversion is forced, we need to use a more hardy algorithm
    if(force) return utf8ToUcs4Forced(utf8);

    // compute size of std::string, skip empty std::strings
    std::string::size_type len = utf8.length(), rd_pos, ch_start = 0;
    if(!len) return std::wstring();

    // skip BOM
    std::string::size_type start = 0;
    if(len >= 3 && utf8.substr(0, 3) == "\xEF\xBB\xBF") start = 3;

    // compute required size
    std::wstring::size_type dlen = 0;
    rd_pos = start;
    uint8_t ch;
    while(rd_pos < len) {
        ch_start = rd_pos;
        ch = utf8[rd_pos];
        if((ch & 0x80) == 0) ++rd_pos;
        else if((ch & 0xE0) == 0xC0) rd_pos += 2;
        else if((ch & 0xF0) == 0xE0) rd_pos += 3;
        else if((ch & 0xF8) == 0xF0) rd_pos += 4;
        else if((ch & 0xFC) == 0xF8) rd_pos += 5;
        else if((ch & 0xFE) == 0xFC) rd_pos += 6;
        else {
            throw BadSourceChar(utf8.c_str() + ch_start,
                utf8.c_str() + ch_start + 1, ch_start, L"UTF-8",
                L"Invalid character in std::string.");
        }
        ++dlen;
    }
    if(rd_pos != len) {
        throw BadSourceChar(utf8.c_str() + ch_start,
            utf8.c_str() + rd_pos, ch_start, L"UTF-8",
            L"Last character appears to be truncated.");
    }

    // allocate memory
    std::wstring dest(dlen, 0);

    // helper macros
#define CONSUME() utf8AuxConsume(ch, utf8, rd_pos, ch_start)
#define CONSUME_FIRST(mask) ch = (p & mask)
#define CHECK(min) do \
    if(ch < min) { \
        throw BadSourceChar(utf8.c_str() + ch_start, \
            utf8.c_str() + rd_pos, ch_start, L"UTF-8", \
            L"Character encoded in an overlong sequence."); \
    } \
while(0)

    // loop through each character in turn
    rd_pos = start;
    for(std::wstring::size_type wr_pos = 0; wr_pos < dlen; ++wr_pos) {
        wchar_t ch = 0;
        ch_start = rd_pos;
        uint8_t p = utf8[rd_pos++];

        if(!(p & 0x80)) {
            ch = p;

        } else if((p & 0xFE) == 0xFC) {
            CONSUME_FIRST(0x01);
            CONSUME();
            CONSUME();
            CONSUME();
            CONSUME();
            CONSUME();
            CHECK(0x4000000);

        } else if((p & 0xFC) == 0xF8) {
            CONSUME_FIRST(0x03);
            CONSUME();
            CONSUME();
            CONSUME();
            CONSUME();
            CHECK(0x200000);

        } else if((p & 0xF8) == 0xF0) {
            CONSUME_FIRST(0x07);
            CONSUME();
            CONSUME();
            CONSUME();
            CHECK(0x10000);

        } else if((p & 0xF0) == 0xE0) {
            CONSUME_FIRST(0x0F);
            CONSUME();
            CONSUME();
            CHECK(0x800);

        } else if((p & 0xE0) == 0xC0) {
            CONSUME_FIRST(0x1F);
            CONSUME();
            CHECK(0x80);

        }
        dest[wr_pos] = ch;
    }

#undef CONSUME
#undef CONSUME_FIRST
#undef CHECK

    return dest;
}



std::wstring asciiToUcs4(const std::string& ascii, bool force)
{
    int skipped = 0;
    std::string::size_type size = ascii.size(), pos;
    std::wstring d(size, 0);

    for(pos = 0; pos < size; ++pos) {
        if(d[pos] & ~0x7F) {
            ++skipped;
            if(!force) throw BadSourceChar(ascii.c_str() + pos,
                ascii.c_str() + pos + 1, pos, L"ASCII",
                L"Invalid character (out of range).");
            continue;
        }
        d[pos] = ascii[pos];
    }

    d.resize(size - skipped);

    return d;
}



std::wstring strToUcs4(const std::string& str)
{
    std::string::size_type size = str.size(), pos;
    std::wstring d(size, 0);
    for(pos = 0; pos < size; ++pos) d[pos] = str[pos];
    return d;
}



}
