/* lw-support/src/lib/String/basics.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Return true if character is whitespace.

\param ch The character to classify.
\retval true if character represents whitespace.
\retval false if character does not represent whitespace.

This function is identical to the C library's isspace() function, except
that it handles Unicode code points.

*/
inline bool isWhitespace(wchar_t ch) __attribute__((const));
bool isWhitespace(wchar_t ch)
{
    return ((ch >= 0x9 && ch <= 0xD) ||
        ch == 0x20 ||
        ch == 0x85 ||
        ch == 0xA0 ||
        ch == 0x1680 ||
        ch == 0x180E ||
        (ch >= 0x2000 && ch <= 0x200A) ||
        ch == 0x2028 ||
        ch == 0x2029 ||
        ch == 0x202F ||
        ch == 0x205F ||
        ch == 0x3000);
}



/*! \brief Strip leading and trailing whitespace.

\param str The UCS-4 encoded string to strip whitespace from.
\returns A copy of the string without trailing/leading whitespace.

*/
std::wstring stripWhitespace(const std::wstring& str);



/*! \brief Convert a UCS-4 string to UTF-8.

\param ucs4 The (UCS-4 encoded) string to convert.
\param force Set this to \a true if you want to force the conversion to
    happen even in the face of invalid characters.
\returns A UTF-8 representation of \a ucs4.
\throws BadSourceChar if a character in the string is outside the
    Unicode range (i.e. 0-0x7FFFFFFF).

This function can be used to convert a UCS-4 string into a UTF-8 string.
It will throw an exception if any error is encountered, unless \a force
is \a true, in which case invalid characters will be skipped. It will
strip a BOM (byte order mark) from the beginning of a string.

*/
std::string ucs4ToUtf8(const std::wstring& ucs4, bool force = false);



/*! \brief Convert a UCS-4 string to ASCII.

\param ucs4 The (UCS-4 encoded) string to convert.
\param force Set this to \a true if you want to force the conversion to
    happen even in the face of invalid characters.
\returns A UTF-8 representation of \a ucs4.
\throws BadSourceChar if a character in the string is outside the
    ASCII range (i.e. 0-0x7F).

This function can be used to convert a UCS-4 string into an ASCII
string. It will throw an exception if any error is encountered, unless \a force
is \a true, in which case invalid characters will be skipped. It will
strip a BOM (byte order mark) from the beginning of a string.

*/
std::string ucs4ToAscii(const std::wstring& ucs4, bool force = false);



/*! \brief Convert a UTF-8 string to UCS-4.

\param utf8 The (UTF-8 encoded) string to convert.
\param force Set this to \a true if you want to force the conversion to
    happen even in the face of invalid characters.
\returns A UCS-4 representation of \a utf8.
\throws BadSourceChar if an invalid byte sequence is encountered in the
    UTF-8 string.

This function can be used to convert a UTF-8 string into a UCS-4 string.
It will throw an exception if any error is encountered, unless \a force
is \a true, in which case invalid characters will be skipped. It will
strip a BOM (byte order mark) from the beginning of a string.

*/
std::wstring utf8ToUcs4(const std::string& utf8, bool force = false);



/*! \brief Convert an ASCII string to UCS-4.

\param ascii The ASCII string to convert.
\param force Set this to \a true if you want to force the conversion to
    happen even in the face of invalid characters.
\returns A UCS-4 representation of \a ascii.
\throws BadSourceChar if an invalid byte (outside the range 0-x7F) is
    encountered in the ASCII string.

This function can be used to convert an ASCII string into a UCS-4
string. It will throw an exception if any error is encountered, unless
\a force is \a true, in which case invalid characters will be skipped.

*/
std::wstring asciiToUcs4(const std::string& ascii, bool force = false);



/*! \brief Convert an 8-bit string to UCS-4.

\param str The 8-bit string to convert.
\returns A UCS-4 representation of \a str.

This function is used to perform a direct conversion of an 8-bit string
into UCS-4. It doesn't take any Unicode issues into account. It always
succeeds. This function can be used if the resultant string will be
checked for validity elsewhere anyway.

*/
std::wstring strToUcs4(const std::string& str);



}
