/* lw-support/src/lib/String/strToInt.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



//BEGIN // Auxiliary conversion functions //////////////////////////////
namespace {

uint64_t strToIntDec(const std::wstring& str,
    std::wstring::size_type pos, std::wstring::size_type strLen)
{
    uint64_t value = 0;
    for(; pos < strLen; ++pos) {
        int ch = str[pos];
        if(ch < '0' || ch > '9')
            throw ParseError(L"Not a decimal number.", str, pos);
        if(value > 1844674407370955161LL) throw NumericOverflow();
        value *= 10;
        if(~value < (uint64_t)(ch - '0')) throw NumericOverflow();
        value += ch - '0';
    }
    return value;
}



uint64_t strToIntHex(const std::wstring& str,
    std::wstring::size_type pos, std::wstring::size_type strLen)
{
    uint64_t value = 0;
    int maxDigRemaining = 16;
    for(; pos < strLen; ++pos) {
        if(!maxDigRemaining--) throw NumericOverflow();
        value <<= 4;

        int ch = str[pos];
        switch(ch) {
            case '0' ... '9':
                value |= ch - '0';
                break;

            case 'a' ... 'f':
                value |= ch - 'a' + 10;
                break;

            case 'A' ... 'F':
                value |= ch - 'A' + 10;
                break;

            default:
                throw ParseError(L"Not a decimal number.", str, pos);
        }
    }
    return value;
}

}
//END // Auxiliary conversion functions ////////////////////////////////



int64_t strToInt64(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    bool negative = false;
    std::wstring::size_type pos = 0;
    uint64_t result;

    // check for sign
    if(str[0] == '-') {
        negative = true;
        ++pos;
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        result = strToIntHex(str, pos, strLen);
    } else {
        if(pos == strLen) return 0;
        result = strToIntDec(str, pos, strLen);
    }

    // check for overflow
    if(negative && result == 0x8000000000000000LL) return -0x8000000000000000LL;
    if(result & ~0x7FFFFFFFFFFFFFFFLL) throw NumericOverflow();

    int64_t p = (int64_t)result;
    if(negative) p *= -1;
    return p;
}



int32_t strToInt32(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    bool negative = false;
    std::wstring::size_type pos = 0;
    uint64_t result;

    // check for sign
    if(str[0] == '-') {
        negative = true;
        ++pos;
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        result = strToIntHex(str, pos, strLen);
    } else {
        if(pos == strLen) return 0;
        result = strToIntDec(str, pos, strLen);
    }

    // check for overflow
    if(negative && result == 0x80000000) return -0x80000000;
    if(result & ~0x7FFFFFFF) throw NumericOverflow();

    int32_t p = (int32_t)result;
    if(negative) p *= -1;
    return p;
}



int16_t strToInt16(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    bool negative = false;
    std::wstring::size_type pos = 0;
    uint64_t result;

    // check for sign
    if(str[0] == '-') {
        negative = true;
        ++pos;
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        result = strToIntHex(str, pos, strLen);
    } else {
        if(pos == strLen) return 0;
        result = strToIntDec(str, pos, strLen);
    }

    // check for overflow
    if(negative && result == 0x8000) return -0x8000;
    if(result & ~0x7FFF) throw NumericOverflow();

    int16_t p = (int16_t)result;
    if(negative) p *= -1;
    return p;
}



int8_t strToInt8(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    bool negative = false;
    std::wstring::size_type pos = 0;
    uint64_t result;

    // check for sign
    if(str[0] == '-') {
        negative = true;
        ++pos;
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        result = strToIntHex(str, pos, strLen);
    } else {
        if(pos == strLen) return 0;
        result = strToIntDec(str, pos, strLen);
    }

    // check for overflow
    if(negative && result == 0x80) return -0x80;
    if(result & ~0x7F) throw NumericOverflow();

    int8_t p = (int8_t)result;
    if(negative) p *= -1;
    return p;
}



uint64_t strToUInt64(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    std::wstring::size_type pos = 0;

    // check for sign
    if(str[0] == '-') {
        throw ParseError(L"Negative number found.", str, 0);
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        return strToIntHex(str, pos, strLen);
    }

    if(pos == strLen) return 0;
    return strToIntDec(str, pos, strLen);
}



uint32_t strToUInt32(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    std::wstring::size_type pos = 0;
    uint64_t result;

    // check for sign
    if(str[0] == '-') {
        throw ParseError(L"Negative number found.", str, 0);
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        result = strToIntHex(str, pos, strLen);
    } else {
        if(pos == strLen) return 0;
        result = strToIntDec(str, pos, strLen);
    }

    // check for overflow
    if(result & ~0xFFFFFFFF) throw NumericOverflow();

    return uint32_t(result);
}



uint16_t strToUInt16(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    std::wstring::size_type pos = 0;
    uint64_t result;

    // check for sign
    if(str[0] == '-') {
        throw ParseError(L"Negative number found.", str, 0);
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        result = strToIntHex(str, pos, strLen);
    } else {
        if(pos == strLen) return 0;
        result = strToIntDec(str, pos, strLen);
    }

    // check for overflow
    if(result & ~0xFFFF) throw NumericOverflow();

    return uint16_t(result);
}



uint8_t strToUInt8(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    if(!strLen) throw ParseError(L"Empty string.", str, 0);

    std::wstring::size_type pos = 0;
    uint64_t result;

    // check for sign
    if(str[0] == '-') {
        throw ParseError(L"Negative number found.", str, 0);
    } else if(str[0] == '+') {
        ++pos;
    }

    if(pos == strLen) {
        throw ParseError(L"String only contained sign character.",
            str, 1);
    }

    // check for hexadecimal
    if(str[pos] == '0' && ++pos != strLen &&
        (str[pos] == 'x' || str[pos] == 'X'))
    {
        if(++pos == strLen)
            throw ParseError(L"String only contained hex sequence.", str, pos);
        result = strToIntHex(str, pos, strLen);
    } else {
        if(pos == strLen) return 0;
        result = strToIntDec(str, pos, strLen);
    }

    // check for overflow
    if(result & ~0xFF) throw NumericOverflow();

    return uint8_t(result);
}



bool strToBool(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);

    // get string length
    std::wstring::size_type strLen = str.length();
    switch(strLen) {
        case 0:
            throw ParseError(L"Empty string.", str, 0);

        case 1: // "0" or "1"
            // check for numeric values
            if(str[0] == '0') return false;
            if(str[0] == '1') return true;
            throw ParseError(L"Unrecognised boolean value.", str, 0);

        case 4: // "true"
            if(tolower(str[0]) != 't' ||
                tolower(str[1]) != 'r' ||
                tolower(str[2]) != 'u' ||
                tolower(str[3]) != 'e')
                    throw ParseError(L"Unrecognised boolean value.", str, 0);
            return true;

        case 5: // "false"
            if(tolower(str[0]) != 'f' ||
                tolower(str[1]) != 'a' ||
                tolower(str[2]) != 'l' ||
                tolower(str[3]) != 's' ||
                tolower(str[4]) != 'e')
                    throw ParseError(L"Unrecognised boolean value.", str, 0);
            return false;
    }

    throw ParseError(L"Unrecognised boolean value.", str, 0);
}



long double strToLongDouble(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);
    if(str.empty()) throw ParseError(L"Empty string.", str, 0);

    // perform conversion
    errno = 0;
    wchar_t* tailPtr = 0;
    long double d = wcstold(str.c_str(), &tailPtr);

    // check for errors
    if(*tailPtr) {
        throw ParseError(L"Not a valid floating point number.", str,
            tailPtr - str.c_str());
    }
    if(errno) throw NumericOverflow();

    return d;
}



double strToDouble(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);
    if(str.empty()) throw ParseError(L"Empty string.", str, 0);

    // perform conversion
    errno = 0;
    wchar_t* tailPtr = 0;
    double d = wcstod(str.c_str(), &tailPtr);

    // check for errors
    if(*tailPtr) {
        throw ParseError(L"Not a valid floating point number.", str,
            tailPtr - str.c_str());
    }
    if(errno) throw NumericOverflow();

    return d;
}



float strToFloat(const std::wstring& strIn)
{
    std::wstring str = stripWhitespace(strIn);
    if(str.empty()) throw ParseError(L"Empty string.", str, 0);

    // perform conversion
    errno = 0;
    wchar_t* tailPtr = 0;
    float d = wcstof(str.c_str(), &tailPtr);

    // check for errors
    if(*tailPtr) {
        throw ParseError(L"Not a valid floating point number.", str,
            tailPtr - str.c_str());
    }
    if(errno) throw NumericOverflow();

    return d;
}



}
