/* lw-support/src/lib/String/strToInt.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. -ddd, ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   -0xhhh, 0xhhh or +0xhhh), case insensitive.

*/
int64_t strToInt64(const std::wstring& str);



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. -ddd, ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   -0xhhh, 0xhhh or +0xhhh), case insensitive.

*/
int32_t strToInt32(const std::wstring& str);



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. -ddd, ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   -0xhhh, 0xhhh or +0xhhh), case insensitive.

*/
int16_t strToInt16(const std::wstring& str);



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. -ddd, ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   -0xhhh, 0xhhh or +0xhhh), case insensitive.

*/
int8_t strToInt8(const std::wstring& str);



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   0xhhh or +0xhhh), case insensitive.

*/
uint64_t strToUInt64(const std::wstring& str);



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   0xhhh or +0xhhh), case insensitive.

*/
uint32_t strToUInt32(const std::wstring& str);



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer, or
    if the integer is negative.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   0xhhh or +0xhhh), case insensitive.

*/
uint16_t strToUInt16(const std::wstring& str);



/*! \brief Parse a string, converting it to an integer.

\param str The string to parse.
\throws ParseError if the string does not represent a valid integer, or
    if the integer is negative.
\throws NumericOverflow if the value represented in the string does not
    fit into an integer.
\returns The value represented in the string.

This function parses a string and returns it as an integer. Any trailing
or leading whitespace is ignored. It understands two types of number:
 - a decimal, with optional leading sign (i.e. ddd or +ddd).
 - a hexadecimal, with optional leading sign, in C notation (i.e.
   0xhhh or +0xhhh), case insensitive.

*/
uint8_t strToUInt8(const std::wstring& str);



/*! \brief Parse a string, converting it to a boolean.

\param str The string to parse.
\throws ParseError if the string does not represent a valid boolean.
\returns The value represented in the string.

This function parses a string and returns it as a boolean. Any trailing
or leading whitespace is ignored. It understands two notations:
 - the strings "0" or "false" (case insensitive), for a false value.
 - the strings "1" or "true" (case insensitive), for a true value.

*/
bool strToBool(const std::wstring& str);



/*! \brief Parse a string, converting it to a number.

\param str The string to parse
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into a double-precision floating point number.
\returns The value represented in the string.

This function parses a string and returns it as a quad-precision
floating point number. Any trailing or leading whitespace is ignored.

*/
long double strToLongDouble(const std::wstring& str);



/*! \brief Parse a string, converting it to a number.

\param str The string to parse
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into a double-precision floating point number.
\returns The value represented in the string.

This function parses a string and returns it as a double-precision
floating point number. Any trailing or leading whitespace is ignored.

*/
double strToDouble(const std::wstring& str);



/*! \brief Parse a string, converting it to a number.

\param str The string to parse
\throws ParseError if the string does not represent a valid integer.
\throws NumericOverflow if the value represented in the string does not
    fit into a double-precision floating point number.
\returns The value represented in the string.

This function parses a string and returns it as a single-precision
floating point number. Any trailing or leading whitespace is ignored.

*/
float strToFloat(const std::wstring& str);



}
