/* lw-support/src/lib/Thread/CompletionNotifier.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Completion object.

A completion object is a synchronisation device used in the following
manner:
 - interested parties register for a notification,
 - the task is begun,
 - on completion, all registered parties are notified.

*/
class CompletionNotifier {
private:
    pthread_mutex_t mut;
    pthread_cond_t cond;
    bool done;

public:
    /// Constructor.
    CompletionNotifier()
        : done(false)
    {
        pthread_mutex_init(&mut, 0);
        pthread_cond_init(&cond, 0);
    }

    /// Destructor.
    ~CompletionNotifier();



    /*! \brief Signal completion.

    This will wake all threads currently waiting on the completion
    condition, and will record the object as completed. Any future calls
    to the wait() function will automatically succeed.

    */
    inline void signal()
    {
        pthread_mutex_lock(&mut);
        done = true;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mut);
    }



    /*! \brief Wait for completion.

    Threads calling this function will block until the object is marked
    as complete.

    */
    inline void wait()
    {
        if(done) return; // avoid lock contention if possible

        pthread_mutex_lock(&mut);
        if(done) { // in case variable was changed before lock acquired
            pthread_mutex_unlock(&mut);
            return;
        }

        pthread_cond_wait(&cond, &mut);
        pthread_mutex_unlock(&mut);
        return;
    }



    /// Thrown if a timeout occurs.
    class Timeout : public ProgramException { };

    /*! \brief Wait for completion, with a timeout.

    \param maxTime The maximum amount of time to wait, in milliseconds.
    \throws CompletionNotifier::Timeout if a timeout occurs.

    Threads calling this function will block until the object is marked
    as complete. If you pass an negative value for \a maxTime, then it
    will wait indefinitely. A zero value will cause it to return
    immediately.

    */
    void wait(int maxTime);
};



}
