/* lw-support/src/lib/Thread/MutexPadlock.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Mutex utility class for safe locking/unlocking.

A Mutex \e padlock will unlock its associated Mutex when it is destroyed.
This means that you can use it to lock a Mutex and then safely unlock it
regardless of how the function exit occurs. A padlock is constructed
unlocked, and you may call the lock and unlock functions at will.

\warning Behaviour is undefined if you attempt to unlock an
         already-unlocked Mutex. This type of padlock is not recursive;
         calling lock twice in succession will probably result in
         deadlock. But see MutexPadlockRecursive.

\sa MutexHolder

*/
class MutexPadlock {
private:
    Mutex& mut;
    bool locked;

public:
    /*! \brief Constructor.

    \param m The Mutex to operate on.

    Constructs the padlock object, but does not lock the Mutex.

    */
    MutexPadlock(Mutex& m)
        : mut(m), locked(false)
    { }



    /*! \brief Destructor (unlocks padlock, if locked).

    If the padlock is locked, the destructor will unlock it. However, if
    it is not locked, then nothing will happen.

    */
    ~MutexPadlock()
    {
        if(locked) mut.unlock();
    }



    /// Lock the Mutex.
    inline void lock()
    {
        mut.lock();
        locked = true;
    }



    /// Unlock the Mutex.
    inline void unlock()
    {
        mut.unlock();
        locked = false;
    }
};



/*! \brief Mutex utility class for safe, recursive locking/unlocking.

A Mutex \e padlock will unlock its associated Mutex when it is destroyed.
This means that you can use it to lock a Mutex and then safely unlock it
regardless of how the function exit occurs. A padlock is constructed
unlocked, and you may call the lock and unlock functions at will.

Note that this padlock allows recursive calls. It even works with fast
Mutexes, because it keeps count internally itself.

\warning Behaviour is undefined if you attempt to unlock an
         already-unlocked Mutex.

\sa MutexHolder

*/
class MutexPadlockRecursive {
private:
    Mutex& mut;
    int count;

public:
    /*! \brief Constructor.

    \param m The Mutex to operate on.

    Constructs the padlock object, but does not lock the Mutex.

    */
    MutexPadlockRecursive(Mutex& m)
        : mut(m), count(0)
    { }



    /*! \brief Destructor (unlocks padlock, if locked).

    If the padlock is locked, the destructor will unlock it. However, if
    it is not locked, then nothing will happen.

    */
    ~MutexPadlockRecursive()
    {
        if(count) mut.unlock();
    }



    /// Lock the Mutex.
    inline void lock()
    {
        if(!count++) mut.lock();
    }



    /// Unlock the Mutex.
    inline void unlock()
    {
        if(!--count) mut.unlock();
    }
};



/*! \brief Mutex utility class for safe locking/unlocking.

On construction, this object locks its associated Mutex. On destruction,
it releases the lock. In essence, it is a less flexible (but less error
prone) Mutex padlock.

\sa MutexPadlock

*/
class MutexHolder {
private:
    Mutex& mut;

public:
    /*! \brief Constructor.

    \param m The Mutex to lock.

    Locks the Mutex.

    */
    MutexHolder(Mutex& m)
        : mut(m)
    {
        mut.lock();
    }



    /*! \brief Destructor.

    Unlocks the Mutex.

    */
    ~MutexHolder()
    {
        mut.unlock();
    }
};



}
