/* lw-support/src/lib/Thread/Semaphore.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Semaphore object.

Supports the usual signal and wait operations, as well as manipulation
of the value. In addition, there is an error handling mechanism which
can be used to signal every thread waiting on the semaphore that it has
become invalid.

*/
class Semaphore {
private:
    pthread_mutex_t mut;
    pthread_cond_t cond;
    int val, waiting;
    bool err;

public:
    /*! \brief Constructor.

    \param init The initial value of the semaphore.

    Creates the semaphore object, and sets its initial value to
    \a init.

    */
    Semaphore(unsigned int init = 0);



    /*! \brief Destructor.

    Destroys the semaphore. Any threads currently waiting on the
    semaphore receive an error signal, so it is safe to delete a
    semaphore object while there are threads waiting on it.

    \warning If a semaphore is deleted in the time between a thread
             deciding to call the wait() function and the time the
             actual system call is executed, behaviour is undefined.

    */
    ~Semaphore();



    /*! \brief Semaphore error signal.

    Thrown if a thread is waiting on the semaphore when an error
    condition is signalled. This also occurs when the semaphore is
    destroyed but some threads are waiting on it.

    */
    class Error : public ProgramException { };



    /// Thrown if a semaphore wait operation times out.
    class Timeout : public ProgramException { };



    /// Increment / post / signal the semaphore.
    inline void signal()
    {
        pthread_mutex_lock(&mut);
        ++val;
        pthread_cond_signal(&cond);
        pthread_mutex_unlock(&mut);
    }



    /*! \brief Decrement / wait on the semaphore.

    \throws Error if an error is signalled or the semaphore destroyed.

    Decrements the value of the semaphore. If the value is zero when
    this function is called, the calling thread is instead put to sleep,
    to be awoken when an error is signalled or the semaphore is posted.

    */
    inline void wait()
    {
        pthread_mutex_lock(&mut);
        ++waiting;
        while(!val && !err) pthread_cond_wait(&cond, &mut);
        if(err) {
            --waiting;
            pthread_mutex_unlock(&mut);
            throw Error();
        }
        --val;
        --waiting;
        pthread_mutex_unlock(&mut);
    }



    /*! \brief Decrement / wait on the semaphore, with a timeout.

    \param maxTime The maximum length of time to wait before timing
                    out, in milliseconds.
    \throws Error if an error is signalled or the semaphore destroyed.
    \throws Timeout if the timeout is reached before the thread can
                    decrement the semaphore.

    Decrements the value of the semaphore. If the value is zero when
    this function is called, the calling thread is instead put to sleep,
    to be awoken when an error is signalled or the semaphore is posted.
    This version includes a timeout. If you pass a negative \a maxTime,
    it will wait indefinitely; a zero value will cause the semaphore to
    be decremented immediately or an error will be thrown.

    */
    void wait(int maxTime);



    /// Signal error.
    inline void signalError()
    {
        pthread_mutex_lock(&mut);
        err = true;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mut);
    }



    /// Retrieve the semaphore's current value.
    inline int value()
    {
        if(err) throw Error();
        return val;
    }
};



}
