/* lw-support/src/lib/Thread/Thread.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Thread object.

This class provides multithreading facilities. You must override its
run() function. The thread is begun after the first call to the start()
function. The thread's execution finishes once run() returns.

The Thread object also contains some static members, yield() and
sleep(), that affect the \em calling thread.

*/
class Thread : public Uncopyable {
private:
    pthread_t thr;
    mutable Semaphore sem;
    bool done, exception; // to store the system state

    Semaphore semStart;
    static void* taskWrapper(void*);



protected:
    /*! \brief Thread function (override this).

    This method must be overridden to provide the thread's behaviour. It
    may either exit normally or by throwing an exception; this can be
    determined through the isDone() and isException() functions.

    */
    virtual void run() = 0;



public:
    /*! \brief Constructor.

    \param stackSize The size of the stack, in bytes. Must be at least
        lw::Thread::minStackSize, or 0 for the system default.
    \throws SystemError if a system error occurs.

    Constructing a new thread object starts up a new thread. The thread
    blocks on a semaphore until you call start() for the first time.
    This is to avoid the situation where the thread calls a method or
    refers to a variable in the derived class which may not have been
    fully instantiated yet.

    */
    Thread(size_t stackSize = 0);



    /*! \brief Destructor.

    When deleted, the thread object's destructor will free up any thread
    resources. It will wait for the thread to quit first.

    \warning The derived class's destructor will be called first, even
        while the thread is still running. You should probably call
        wait() in your derived class's destructor.

    */
    virtual ~Thread();



    /// Start thread running.
    void start();



    /// Minimum stack size, in bytes.
    static const size_t minStackSize;



    /// Reports whether the thread has finished executing.
    bool isDone() const
    {
        return done;
    }

    /// Reports whether the thread finished because of an exception.
    bool isException() const
    {
        return exception;
    }



    /*! \brief Wait for the thread to finish, with a timeout.

    \param timeout The maximum length of time to wait before timing
        out, in milliseconds (negative means forever).
    \throws Semaphore::Timeout if the timeout is reached before the
        thread has finished.

    */
    void wait(int timeout) const;



    /*! \brief Signal that the thread needs to quit.

    This function, by default, does nothing. However, if you have some
    specific way of signalling to your thread function that it should
    quit, then you can use that method here.

    */
    virtual void signalQuit();



    /*! \brief Yields CPU to another thread.

    Briefly yields the CPU to another thread, but does not put the
    process to sleep.

    */
    static inline void yield()
    {
        pthread_yield();
    }



    /// Causes the current thread to sleep for a length of time.
    static void sleep(const ElapsedTime& time);



    /*! \brief Prints a stack trace.

    This is a debugging function that prints a stack trace onto stderr. It assumes that fd 2 is
    stderr. It will also be confused by compiler optimisations. See the `backtrace()' libc function
    for details.

    */
    static void dbg_backtrace();
};



}
