/* lw-support/src/lib/Time/Date.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {




void Date::fixupDate()
{
    while(ord < 1) {
        --year;
    ord += isLeapYear(year) ? 366 : 365;
    }

while(ord > (isLeapYear(year) ? 366 : 365)) {
ord -= isLeapYear(year) ? 366 : 365;
    ++year;
}
}



/* This comes from the algorithm given on Wikipedia at
http://en.wikipedia.org/wiki/Calculating_the_day_of_the_week */
namespace {

inline int weekDayJan0(int year)
{
    div_t cent = div(year, 100);
    div_t leap = div(cent.quot, 4);

    int centOffset = (3 - leap.rem) * 2;
    int yrOffset = cent.rem + ((cent.rem + 3) >> 2) - 1;

    return (centOffset + yrOffset) % 7;
}

}



Date::Date(int year, int month, int day)
    : year(year)
{
    if(month < 1 || month > 12 || day < 1 || day > 31 ||
       !(ord = (isLeapYear(year) ? leapMDtoOrd : normalMDtoOrd)[month - 1][day - 1]) )
    {
        throw BadDate();
    }
}



Date::Date(int year, int ordinalDay)
    : year(year), ord(ordinalDay)
{
    if(ord < 1 || ord > 366 || (ord == 366 && !isLeapYear(year))) {
        throw BadDate();
    }
}



Date Date::now()
{
    time_t t = time(0);
    struct tm tt;
    localtime_r(&t, &tt);
    return Date(tt.tm_year + 1900, tt.tm_mon + 1, tt.tm_mday);
}



Date Date::fromWeek(int year, int week, int weekDay)
{
    // sanity checks
    if(weekDay < 1 || weekDay > 7 || week < 1 || week > 53)
        throw BadDate();
    if(week == 53 && (weekDayJan0(year) != 3 &&
                      (!isLeapYear(year) || weekDayJan0(year) != 2)))
        throw BadDate();

    // find date of Monday, week 1
    Date d(year, 1);
    switch(weekDayJan0(year)) {
    case 0: // Sunday
        break;

    case 1: // Monday
        d -= 1;
        break;

    case 2:
        d -= 2;
        break;

    case 3:
        d -= 3;
        break;

    case 4:
        d += 3;
        break;

    case 5:
        d += 2;
        break;

    case 6:
        d += 1;
        break;
    }

    // now move to correct week and day
    d += (week - 1) * 7 + (weekDay - 1);

    return d;
}



int Date::getMonth() const
{
return (isLeapYear(year) ? leapOrdtoMD : normalOrdtoMD)[ord][0];
}



int Date::getDay() const
{
return (isLeapYear(year) ? leapOrdtoMD : normalOrdtoMD)[ord][1];
}



int Date::getWeekDay() const
{
    int wd = (weekDayJan0(year) + ord) % 7;
return wd ?: 7;
}



int Date::getWeek() const
{
    if(isLeapYear(year)) return leapOrdtoWeek[weekDayJan0(year)][ord];
    else return normalOrdtoWeek[weekDayJan0(year)][ord];
}



Date& Date::operator+=(int numDays)
{
    ord += numDays;
    fixupDate();
    return *this;
}



Date& Date::operator-=(int numDays)
{
    ord -= numDays;
    fixupDate();
    return *this;
}



Date& Date::operator++()
{
    ++ord;
    fixupDate();
    return *this;
}



Date Date::operator++(int)
{
    Date copy(*this);
    ++ord;
    fixupDate();
    return copy;
}



Date& Date::operator--()
{
    --ord;
    fixupDate();
    return *this;
}



Date Date::operator--(int)
{
    Date copy(*this);
    --ord;
    fixupDate();
    return copy;
}



int Date::operator-(const Date& sub) const
{
    int diff = ord - sub.ord;

    for(int xYear = year; xYear < sub.year; ++xYear) {
    diff -= isLeapYear(xYear) ? 366 : 365;
    }
    for(int yYear = year; yYear > sub.year; --yYear) {
    diff += isLeapYear(yYear - 1) ? 366 : 365;
    }

    return diff;
}



/* A good ISO8601 description is at Wikipedia:
http://en.wikipedia.org/wiki/ISO_8601 */

std::wstring Date::toString(bool extended, ISO8601Format format,
                            ISO8601Precision precision) const
{
    // sanity check arguments
    if(format == ISO8601FormatDetect) throw BadFormat();
    switch(precision) {
    case ISO8601PrecisionFull:
    case ISO8601PrecisionDay:
        break;

    case ISO8601PrecisionMonth:
    case ISO8601PrecisionYear:
        if(format != ISO8601FormatCalendar) throw BadFormat();
        break;

    case ISO8601PrecisionWeek:
        if(format != ISO8601FormatWeekDate) throw BadFormat();
        break;

    default:
        throw BadFormat();
    }

    // set up output stream
    std::wostringstream o;
    o << std::setfill(L'0');

    // deal with the weekdate format
    if(format == ISO8601FormatWeekDate) {
        int wk = getWeek();
        if(!wk) {
            o << std::setw(4) << (year + 1);
            if(extended) o << '-';
            o << "W01";

        } else if(wk < 0) {
            o << std::setw(4) << (year - 1);
            if(extended) o << '-';
            o << 'W' << std::setw(2) << (-wk);

        } else {
            o << std::setw(4) << year;
            if(extended) o << '-';
            o << 'W' << std::setw(2) << wk;

        }

        if(precision == ISO8601PrecisionWeek) return o.str();
        if(extended) o << '-';
        o << getWeekDay();
        return o.str();
    }

    // output the year
    o << std::setw(4) << year;

    // deal with the ordinal format
    if(format == ISO8601FormatOrdinal) {
        if(extended) o << L'-';
        o << std::setw(3) << getOrdinal();
        return o.str();
    }

    // output the month and (possibly) day
    switch(precision) {
    case ISO8601PrecisionFull:
    case ISO8601PrecisionDay:
        if(extended) o << L'-';
        o << std::setw(2) << getMonth();
        if(extended) o << L'-';
        o << std::setw(2) << getDay();
        break;

    case ISO8601PrecisionMonth:
        if(extended) o << L'-';
        o << std::setw(2) << getMonth();
        break;

    default:
        break;
    }

    return o.str();
}



namespace {

/* The ISO8601 date parsing subsystem uses a bunch of parallel state
machines. Each active state machine is fed a character; if it can't cope
with that character, the machine becomes inactive. On completion, each
machine is asked for a probability. The machine with highest probability
wins. */
class ISO8601DateParser {
public:
    // Processes a char, returning false if the char doesn't fit the
    // parser.
    virtual bool input(int ch) = 0;

    // Probability system used to determine which state machine should
    // 'win' if several complete
    enum Probability {
        Impossible,
            Unlikely,
            Likely,
            Definite
    };
    virtual Probability getProbability() = 0;

    // Return a date object.
    virtual Date getDate(const std::wstring& str) = 0;

    virtual ~ISO8601DateParser() { }
};

class ISO8601DateParser_CalendarBasic : public ISO8601DateParser {
private:
    int numDigits;

public:
    ISO8601DateParser_CalendarBasic()
        : numDigits(0)
    { }

    virtual bool input(int ch)
    {
        switch(ch) {
        case '0' ... '9':
            ++numDigits;
            return true;

        case '-':
            if(!numDigits) return true;
        }
        return false;
    }

    virtual Probability getProbability()
    {
    return (numDigits < 8) ? Impossible : Likely;
    }

    virtual Date getDate(const std::wstring& str)
    {
        int strLen = str.length();
        return Date(strToInt32(str.substr(0, strLen - 4)),
                    strToInt32(str.substr(strLen - 4, 2)),
                    strToInt32(str.substr(strLen - 2, 2)));
    }

    virtual ~ISO8601DateParser_CalendarBasic() { }
};

class ISO8601DateParser_CalendarExtended : public ISO8601DateParser {
private:
    int numDigits;
    enum {
        stateNothing,
            stateYear,
            stateMonth,
            stateDay
    }state;
    int year, month, day;
    bool yearNegative;

public:
    ISO8601DateParser_CalendarExtended()
        : numDigits(0), state(stateNothing),
        year(0), month(0), day(0), yearNegative(false)
    { }

    virtual bool input(int ch)
    {
        switch(state) {
        case stateNothing:
            state = stateYear;
            switch(ch) {
            case '-':
                yearNegative = true;
                return true;

            case '0' ... '9':
                break;

            default:
                return false;
            }
                // fall through

        case stateYear:
            if(ch == '-' && numDigits >= 4) {
                state = stateMonth;
                numDigits = 0;
                return true;
            }
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            year *= 10;
            year += ch - '0';
            return true;

        case stateMonth:
            if(numDigits == 2) {
                state = stateDay;
                numDigits = 0;
                return ch == '-';
            }
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            month *= 10;
            month += ch - '0';
            return true;

        case stateDay:
            if(numDigits == 2) return false;
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            day *= 10;
            day += ch - '0';
            return true;
        }
        return false;
    }

    virtual Probability getProbability()
    {
        return (state == stateDay && numDigits == 2) ?
            Definite : Impossible;
    }

    virtual Date getDate(const std::wstring&)
    {
        return Date(year, month, day);
    }

    virtual ~ISO8601DateParser_CalendarExtended() { }
};

class ISO8601DateParser_OrdinalBasic : public ISO8601DateParser {
private:
    int numDigits;

public:
    ISO8601DateParser_OrdinalBasic()
        : numDigits(0)
    { }

    virtual bool input(int ch)
    {
        switch(ch) {
        case '0' ... '9':
            ++numDigits;
            return true;

        case '-':
            if(!numDigits) return true;
        }
        return false;
    }

    virtual Probability getProbability()
    {
        if(numDigits < 7) return Impossible;
    return (numDigits == 7) ? Likely : Unlikely;
    }

    virtual Date getDate(const std::wstring& str)
    {
        int strLen = str.length();
        return Date(strToInt32(str.substr(0, strLen - 3)),
                    strToInt32(str.substr(strLen - 3, 3)));
    }

    virtual ~ISO8601DateParser_OrdinalBasic() { }
};

class ISO8601DateParser_OrdinalExtended : public ISO8601DateParser {
private:
    int numDigits;
    enum {
        stateNothing,
            stateYear,
            stateDay
    }state;
    int year, day;
    bool yearNegative;

public:
    ISO8601DateParser_OrdinalExtended()
        : numDigits(0), state(stateNothing),
        year(0), day(0), yearNegative(false)
    { }

    virtual bool input(int ch)
    {
        switch(state) {
        case stateNothing:
            state = stateYear;
            switch(ch) {
            case '-':
                yearNegative = true;
                return true;

            case '0' ... '9':
                break;

            default:
                return false;
            }
                // fall through

        case stateYear:
            if(ch == '-' && numDigits >= 4) {
                state = stateDay;
                numDigits = 0;
                return true;
            }
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            year *= 10;
            year += ch - '0';
            return true;

        case stateDay:
            if(numDigits == 3) return false;
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            day *= 10;
            day += ch - '0';
            return true;
        }
        return false;
    }

    virtual Probability getProbability()
    {
        return (state == stateDay && numDigits == 3) ?
            Definite : Impossible;
    }

    virtual Date getDate(const std::wstring&)
    {
        return Date(year, day);
    }

    virtual ~ISO8601DateParser_OrdinalExtended() { }
};

class ISO8601DateParser_WeekDateBasic : public ISO8601DateParser {
private:
    int numDigits;
    enum {
        stateNothing,
            stateYear,
            stateWeeks
    }state;

public:
    ISO8601DateParser_WeekDateBasic()
        : numDigits(0), state(stateNothing)
    { }

    virtual bool input(int ch)
    {
        switch(state) {
        case stateNothing:
            state = stateYear;
            switch(ch) {
            case '-':
                return true;

            case '0' ... '9':
                break;

            default:
                return false;
            }
                // fall through

        case stateYear:
            if(ch == 'W' && numDigits >= 4) {
                state = stateWeeks;
                numDigits = 0;
                return true;
            }
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            return true;

        case stateWeeks:
            if(numDigits++ == 3) return false;
            if(ch < '0' || ch > '9') return false;
            return true;
        }
        return false;
    }

    virtual Probability getProbability()
    {
        return (state == stateWeeks && numDigits == 3) ?
            Definite : Impossible;
    }

    virtual Date getDate(const std::wstring& str)
    {
        std::wstring::size_type strLen = str.length();
        return Date::fromWeek(strToInt32(str.substr(0, strLen - 4)),
                              strToInt32(str.substr(strLen - 3, 2)),
                              str[strLen - 1] - '0');
    }

    virtual ~ISO8601DateParser_WeekDateBasic() { }
};

class ISO8601DateParser_WeekDateExtended : public ISO8601DateParser {
private:
    int numDigits;
    enum {
        stateNothing,
            stateYear,
            stateWeekDash,
            stateWeek,
            stateDay
    }state;

public:
    ISO8601DateParser_WeekDateExtended()
        : numDigits(0), state(stateNothing)
    { }

    virtual bool input(int ch)
    {
        switch(state) {
        case stateNothing:
            state = stateYear;
            switch(ch) {
            case '-':
                return true;

            case '0' ... '9':
                break;

            default:
                return false;
            }
                // fall through

        case stateYear:
            if(ch == '-' && numDigits >= 4) {
                state = stateWeekDash;
                numDigits = 0;
                return true;
            }
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            return true;

        case stateWeekDash:
            state = stateWeek;
            return ch == 'W';

        case stateWeek:
            if(numDigits == 2) {
                state = stateDay;
                numDigits = 0;
                return ch == '-';
            }
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            return true;

        case stateDay:
            if(numDigits == 1) return false;
            if(ch < '0' || ch > '9') return false;
            ++numDigits;
            return true;
        }
        return false;
    }

    virtual Probability getProbability()
    {
        return (state == stateDay && numDigits == 1) ?
            Definite : Impossible;
    }

    virtual Date getDate(const std::wstring& str)
    {
        std::wstring::size_type strLen = str.length();
        return Date::fromWeek(strToInt32(str.substr(0, strLen - 6)),
                              strToInt32(str.substr(strLen - 4, 2)),
                              str[strLen - 1] - '0');
    }

    virtual ~ISO8601DateParser_WeekDateExtended() { }
};

}

Date Date::fromString(const std::wstring& str, ISO8601Format format)
{
    // set up a std::list of participating state machines
    ISO8601DateParser_CalendarBasic machineCalendarBasic;
    ISO8601DateParser_CalendarExtended machineCalendarExtended;
    ISO8601DateParser_OrdinalBasic machineOrdinalBasic;
    ISO8601DateParser_OrdinalExtended machineOrdinalExtended;
    ISO8601DateParser_WeekDateBasic machineWeekDateBasic;
    ISO8601DateParser_WeekDateExtended machineWeekDateExtended;

    std::vector<ISO8601DateParser*> stateMachines;
    switch(format) {
    case ISO8601FormatDetect:
        stateMachines.push_back(&machineCalendarBasic);
        stateMachines.push_back(&machineCalendarExtended);
        stateMachines.push_back(&machineOrdinalBasic);
        stateMachines.push_back(&machineOrdinalExtended);
        stateMachines.push_back(&machineWeekDateBasic);
        stateMachines.push_back(&machineWeekDateExtended);
        break;

    case ISO8601FormatCalendar:
        stateMachines.push_back(&machineCalendarBasic);
        stateMachines.push_back(&machineCalendarExtended);
        break;

    case ISO8601FormatOrdinal:
        stateMachines.push_back(&machineOrdinalBasic);
        stateMachines.push_back(&machineOrdinalExtended);
        break;

    case ISO8601FormatWeekDate:
        stateMachines.push_back(&machineWeekDateBasic);
        stateMachines.push_back(&machineWeekDateExtended);
        break;
    }

    // pass the std::string through all the state machines
    int numMachines = stateMachines.size();
    std::wstring::size_type pos = 0, len = str.length();

    for(pos = 0; pos < len; ++pos) {
        int ch = str[pos];

        for(int iter = 0; iter < numMachines; ++iter) {
            if(!stateMachines[iter]->input(ch)) {
                // this machine can't process the input any more
                stateMachines.erase(stateMachines.begin() + iter);
                --iter;
                --numMachines;
            }
        }
    }

    // choose the machine with the highest likelihood
    ISO8601DateParser* bestMachine = 0;
    ISO8601DateParser::Probability bestProbability = ISO8601DateParser::Impossible;
    for(int iter = 0; iter < numMachines; ++iter) {
        ISO8601DateParser::Probability p =
            stateMachines[iter]->getProbability();
        if(p > bestProbability) {
            bestProbability = p;
            bestMachine = stateMachines[iter];
        }
    }

    // return the date object or fail
    if(bestMachine) return bestMachine->getDate(str);

    throw ParseError(L"Cannot interpet as ISO8601 date", str, 0);
}



std::wostream& operator<<(std::wostream& ostr, const Date& date)
{
    ostr << date.toString();
    return ostr;
}



}

/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
*/
