/* lw-support/src/lib/Time/Date.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief A Gregorian date.

This class holds a single date in the Gregorian calendar. It makes no
provision for dates in other formats (i.e. it uses the Proleptic
Gregorian Calendar). It also provides for conversion to/from ISO8601
formats. It uses astronomical years.

*/
class Date {
private:
    int year, ord;
    void fixupDate(); // fix up after arithmetic

public:
    //BEGIN // Constructors etc. ///////////////////////////////////////

    /*! \brief Constructor (calendar date).

    \param year The astronimical year.
    \param month Month of the year (1-12).
    \param day Day of the month (1-31).
    \throws BadDate if \a day or \a month are invalid.

    */
    Date(int year, int month, int day);

    /*! \brief Constructor (ordinal date).

    \param year The astronimical year.
    \param ordinalDay Day of the year (1-366).
    \throws BadDate if \a ordinalDay is invalid.

    */
    Date(int year, int ordinalDay);

    /*! \brief Make object from ISO8601 weekday form.

    \param year The week's year.
    \param week The week number (1-53).
    \param weekDay The week day (1-7).
    \throws BadDate if \a week or \a weekDay are invalid.

    */
    static Date fromWeek(int year, int week, int weekDay);

    /// Copy constructor.
    Date(const Date& date)
        : year(date.year), ord(date.ord)
    { }

    /// Assignment operator.
    Date& operator=(const Date& date)
    {
        year = date.year;
        ord = date.ord;
        return *this;
    }

    /// Create object from current calendar day.
    static Date now();



    //BEGIN // String functions ////////////////////////////////////////

    /// Thrown when format specifiers don't make sense, etc.
    class BadFormat : public ProgramException { };



    /*! \brief Convert to std::string (ISO8601) representation.

    \param format Which ISO8601 format to use.
    \param precision Which precision to use.
    \param extended Whether to use basic or extended notation.
    \throws BadFormat if \a format and/or \a precision don't make sense
        for this type.
    \returns The ISO8601 representation of the date.

    This function converts the date into its ISO8601 notation, with a
    variety of possible formatting options.

    */
    std::wstring toString(bool extended = false,
        ISO8601Format format = ISO8601FormatCalendar,
        ISO8601Precision precision = ISO8601PrecisionFull) const;



    /*! \brief Parse an ISO8601 date.

    \param str The ISO8601 std::string to parse.
    \param format Which ISO8601 format to use.
    \throws ParseError if the std::string cannot be parsed.
    \throws BadDate if the date is parsed correctly but is not valid.
    \returns A newly-created date object.

    \note There is some ambiguity between the ISO8601 basic forms, since
        an 8-digit sequence could either be a basic calendar form or a
        basic ordinal form with a 5-digit year. For this reason, all
        sequences of 8 or more digits are interpreted as basic calendar
        form if \a format is \a ISO8601FormatDetect.

    \note The year part is assumed to always be at least four digits.

    */
    static Date fromString(const std::wstring& str,
        ISO8601Format format = ISO8601FormatDetect);



    //BEGIN // Access functions ////////////////////////////////////////

    /// Returns the year.
    inline int getYear() const
    { return year; }

    /// Returns the month.
    int getMonth() const;

    /// Returns the day.
    int getDay() const;

    /// Returns the ordinal day.
    inline int getOrdinal() const
    {
        return ord;
    }

    /// Returns the week day (1 = Monday, 7 = Sunday).
    int getWeekDay() const;

    /*! \brief Returns the week number.

    \returns The ISO8601 week number.
    \retval 0 if this is week 1 of next year.
    \retval -n if this is week \a n of last year.

    Returns the day's week number within the year. There are a couple of
    special return values; a zero means that this is in fact week 1 of
    next year, and a negative number \a -n means that this is week \a n
    of last year.

    */
    int getWeek() const;



    //BEGIN // Modify functions / operators ////////////////////////////

    /*! \brief Advance a number of days.

    \param numDays The number of days to advance by. Can be negative.
    \returns A reference to the modified object.

    Moves the date forward a number of days, correctly accounting for
    month and year boundaries.

    */
    Date& operator+=(int numDays);



    /*! \brief Roll back a number of days.

    \param numDays The number of days to reverse by. Can be negative.
    \returns A reference to the modified object.

    Moves the date backward a number of days, correctly accounting for
    month and year boundaries.

    */
    Date& operator-=(int numDays);



    /// Advance a day.
    Date& operator++();

    /// Advance a day.
    Date operator++(int);

    /// Go back a day.
    Date& operator--();

    /// Go back a day.
    Date operator--(int);



    //BEGIN // Comparison operators ////////////////////////////////////

    /*! \brief Get difference in days.

    \param date The date to subtract from this one.
    \returns The difference, in days, between two dates.

    */
    int operator-(const Date& date) const;



    /// Returns true if two dates are equal.
    inline bool operator==(const Date& date) const
    {
        return (year == date.year) && (ord == date.ord);
    }

    /// Returns true if two dates are not equal.
    inline bool operator!=(const Date& date) const
    {
        return (year != date.year) || (ord != date.ord);
    }

    /// Returns true if \a date is less than \a this.
    inline bool operator<(const Date& date) const
    {
        return (year < date.year) || (year == date.year &&
            (ord < date.ord));
    }

    /// Returns true if \a date is less than or equal to \a this.
    inline bool operator<=(const Date& date) const
    {
        return (year < date.year) || (year == date.year &&
            (ord <= date.ord));
    }

    /// Returns true if \a date is greater than \a this.
    inline bool operator>(const Date& date) const
    {
        return (year > date.year) || (year == date.year &&
            (ord > date.ord));
    }

    /// Returns true if \a date is greater than or equal to \a this.
    inline bool operator>=(const Date& date) const
    {
        return (year > date.year) || (year == date.year &&
            (ord >= date.ord));
    }



    //BEGIN // Utility functions ///////////////////////////////////////

    /// Returns true if \a year is a leap year.
    static inline bool isLeapYear(int year)
    {
        if(year < 0) year *= -1;
        return !(year % 400) || ((year % 100) && !(year & 3));
    }

    /// Returns number of days in the given \a month and \a year.
    static inline int daysInMonth(int year, int month)
    {
        switch(month) {
            case 4:
            case 6:
            case 9:
            case 11:
                return 30;

            case 2:
                return (isLeapYear(year) ? 29 : 28);
        }
        return 31;
    }
};



/// Stream insertion operator.
std::wostream& operator<<(std::wostream&, const Date&);



}
