/* lw-support/src/lib/Time/DateTime.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Stores a date and a time.

This class uses (internally) a DayTime and a Date class. These allow it
to represent pretty much arbitrary timescales (well, about +/- 2e9 years
with the precision of a nanosecond).

Altough you can individually operate on the internal Date and DayTime
instances, this isn't always a good idea. For instance, adding 24 hours
onto the DayTime object won't alter the Date. Instead, use the
addHours() function and equivalents.

*/
class DateTime {
protected:
    /// The day.
    Date date;

    /// The time.
    DayTime time;

public:
    //BEGIN // Constructors, etc. //////////////////////////////////////

    /// Construct from date and time.
    DateTime(const Date& date, const DayTime& time)
        : date(date), time(time)
    { }

    /// Construct (at midnight) from date.
    DateTime(const Date& date)
        : date(date), time(0, 0, 0)
    { }

    /// Copy constructor.
    DateTime(const DateTime& dateTime)
        : date(dateTime.date), time(dateTime.time)
    { }

    /// Assignment operator.
    DateTime& operator=(const DateTime& dateTime)
    {
        date = dateTime.date;
        time = dateTime.time;
        return *this;
    }

    /// Construct from calendar time.
    static DateTime now()
    { return DateTime(Date::now(), DayTime::now()); }



    //BEGIN // String routines /////////////////////////////////////////

    /*! \brief Convert to ISO8601 std::string.

    \param extended \a true if you want to use the extended form,
        \a false if not.
    \param decimalTime \a true if you want to use the decimal time
        notation.
    \param precision How precise you want the converted std::string to be.
    \param dateFormat Which format to use for the date.

    Converts the entire date/time into an ISO8601-formatted std::string. See
    documentation for Date and DayTime for more information about the
    parameters and their results.

    \sa lw::Date::toString(), lw::DayTime::toString()

    */
    std::wstring toString(bool extended = false, bool decimalTime = false,
        ISO8601Format dateFormat = ISO8601FormatCalendar,
        ISO8601Precision precision = ISO8601PrecisionFull) const;



    /*! \brief Create from ISO8601 std::string.

    \param str The ISO8601 std::string to parse.
    \param dateFormat Which ISO8601 format to use.
    \throws ParseError if the std::string cannot be parsed.
    \throws BadDate if the date is parsed correctly but is not valid.
    \returns A newly-created date object.

    Attempts to parse \a str as an ISO8601 date/time std::string. If the time
    part is not present, the time will be set to midnight.

    \note There is some ambiguity between the ISO8601 basic date forms,
        since an 8-digit sequence could either be a basic calendar form
        or a basic ordinal form with a 5-digit year. For this reason,
        all sequences of 8 or more digits are interpreted as basic
        calendar form if \a dateFormat is \a ISO8601FormatDetect.

    \note The year part is assumed to always be at least four digits.

    */
    static DateTime fromString(const std::wstring& str,
        ISO8601Format dateFormat = ISO8601FormatDetect);



    //BEGIN // Access functions ////////////////////////////////////////

    /// Get date component.
    const Date& getDate() const
    { return date; }

    /// Get date component.
    Date& getDate()
    { return date; }

    /// Get time component.
    const DayTime& getTime() const
    { return time; }

    /// Get time component.
    DayTime& getTime()
    { return time; }



    //BEGIN // Modify functions ////////////////////////////////////////

    /// Set date component.
    DateTime& setDate(const Date& date)
    {
        this->date = date;
        return *this;
    }

    /// Set time component.
    DateTime& setTime(const DayTime& time)
    {
        this->time = time;
        return *this;
    }

    /// Set date and time.
    DateTime& set(const Date& date, const DayTime& time)
    {
        this->date = date;
        this->time = time;
        return *this;
    }



    /// Add a number of nanoseconds.
    DateTime& addNanoSeconds(int64_t ns);

    /// Add a number of seconds.
    DateTime& addSeconds(int s);

    /// Add a number of minutes.
    DateTime& addMinutes(int m);

    /// Add a number of hours.
    DateTime& addHours(int h);

    /// Add a number of days.
    DateTime& addDays(int d);




    //BEGIN // Operators ///////////////////////////////////////////////

    /// Returns true if the dates are equal.
    bool operator==(const DateTime& dateTime) const
    { return date == dateTime.date && time == dateTime.time; }

    /// Returns true if the dates are not equal.
    bool operator!=(const DateTime& dateTime) const
    { return date != dateTime.date || time != dateTime.time; }

    /// Returns true if less than \a dateTime.
    bool operator<(const DateTime& dateTime) const
    {
        return date < dateTime.date ||
            (date == dateTime.date && time < dateTime.time);
    }

    /// Returns true if less than or equal to \a dateTime.
    bool operator<=(const DateTime& dateTime) const
    {
        return date < dateTime.date ||
            (date == dateTime.date && time <= dateTime.time);
    }

    /// Returns true if greater than \a dateTime.
    bool operator>(const DateTime& dateTime) const
    {
        return date > dateTime.date ||
            (date == dateTime.date && time > dateTime.time);
    }

    /// Returns true if greater than or equal to \a dateTime.
    bool operator>=(const DateTime& dateTime) const
    {
        return date > dateTime.date ||
            (date == dateTime.date && time >= dateTime.time);
    }
};



/// Stream insertion operator.
std::wostream& operator<<(std::wostream&, const DateTime&);



}
