/* lw-support/src/lib/Time/DayTime.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



DayTime& DayTime::wrap(int& dayDelta)
{
    int d = 0;
    dayDelta = 0;
    if(_ns < 0) {
        _ns += tenExp9;
        _s--;
    }
    if(_ns >= tenExp9) {
        _ns -= tenExp9;
        _s++;
    }
    if(_s < 0) {
        d = 1 + (_s / -60);
        _s += d * 60;
        _m -= d;
    }
    if(_s >= 60) {
        d = _s / 60;
        _s -= d * 60;
        _m += d;
    }
    if(_m < 0) {
        d = 1 + (_m / -60);
        _m += d * 60;
        _h -= d;
    }
    if(_m >= 60) {
        d = _m / 60;
        _m -= d * 60;
        _h += d;
    }
    if(_h < 0) {
        d = 1 + (_h / -24);
        _h += d * 24;
        dayDelta = -d;
    }
    if(_h >= 24) {
        d = _h / 24;
        _h -= d * 24;
        dayDelta = d;
    }
    return *this;
}



std::wstring DayTime::toString(bool extended, ISO8601Precision precision,
    bool decimal) const
{
    std::wostringstream o, o2;
    o << std::setfill(L'0');
    o2 << std::fixed;

    switch(precision) {
        case ISO8601PrecisionHour:
            o << std::setw(2) << _h;
            if(decimal) {
                double f = (_m / 60.0 + _s / 3600.0) + _ns / 3600e9;
                o2 << f;
                o << (o2.str().substr(1, std::wstring::npos));
            }
            break;

        case ISO8601PrecisionMinute:
            o << std::setw(2) << _h;
            if(extended) o << ':';
            o << std::setw(2) << _m;
            if(decimal) {
                double f = (_s / 60.0) + _ns / 60e9;
                o2 << f;
                o << (o2.str().substr(1, std::wstring::npos));
            }
            break;

        case ISO8601PrecisionSecond:
            o << std::setw(2) << _h;
            if(extended) o << ':';
            o << std::setw(2) << _m;
            if(extended) o << ':';
            o << std::setw(2) << _s;
            if(decimal) {
                double f = _ns / 1e9;
                o2 << f;
                o << (o2.str().substr(1, std::wstring::npos));
            }
            break;

        case ISO8601PrecisionFull:
            o << std::setw(2) << _h;
            if(extended) o << ':';
            o << std::setw(2) << _m;
            if(extended) o << ':';
            o << std::setw(2) << _s;
            if(_ns) {
                double f = _ns / 1e9;
                o2 << f;
                o << (o2.str().substr(1, std::wstring::npos));
            }
            break;

        default:
            throw BadFormat();
    }
    return o.str();
}



namespace {

/* This function parses basic and extended format ISO8601 times (but not
the decimal notations, which are handled separately in fromString). It
returns the number of elements parsed (1, 2 or 3). This can be used to
determine which element a decimal fraction applies to.
*/
int ISO8601TimeParser(const std::wstring& str, int& h, int& m, int& s)
{
    // The remaining ISO8601 formats can all be disambiguated on length.
    switch(str.length()) {
        case 8: // hh:mm:ss
            if(str[2] != ':' || str[5] != ':') {
                throw ParseError(L"Not a recognised ISO8601 format.",
                    str, str[2] == ':' ? 5 : 2);
            }
            h = strToInt32(str.substr(0, 2));
            m = strToInt32(str.substr(3, 2));
            s = strToInt32(str.substr(6, 2));
            return 3;

        case 6: // hhmmss
            h = strToInt32(str.substr(0, 2));
            m = strToInt32(str.substr(2, 2));
            s = strToInt32(str.substr(4, 2));
            return 3;

        case 5: // hh:mm
            if(str[2] != ':') {
                throw ParseError(L"Not a recognised ISO8601 format.",
                    str, 2);
            }
            h = strToInt32(str.substr(0, 2));
            m = strToInt32(str.substr(3, 2));
            return 2;

        case 4: // hhmm
            h = strToInt32(str.substr(0, 2));
            m = strToInt32(str.substr(2, 2));
            return 2;

        case 2: // hh
            h = strToInt32(str);
            return 1;
    }

    throw ParseError(L"Not a recognised ISO8601 format.", str, 0);
}

}



DayTime DayTime::fromString(const std::wstring& str)
{
    int h = 0, m = 0, s = 0, ns = 0;
    std::wstring::size_type decimal = std::wstring::npos;

    try {
        // Check for decimal time.
        if( (decimal = str.find('.')) != std::wstring::npos ||
            (decimal = str.find(',')) != std::wstring::npos)
        {
            // get fractional part
            double fraction = strToDouble(
                str.substr(decimal, std::wstring::npos));

            switch(ISO8601TimeParser(str.substr(0, decimal), h, m, s)) {
                case 1:
                    m = (int)(fraction * 60);
                    s = (int)fmod(fraction * 3600, 60);
                    ns = (int)fmod(fraction * 3600 * tenExp9, tenExp9);
                    break;

                case 2:
                    s = (int)(fraction * 60);
                    ns = (int)fmod(fraction * 60 * tenExp9, tenExp9);
                    break;

                case 3:
                    ns = (int)(fraction * tenExp9);
                    break;
            }

        } else {
            // Interpret as non-decimal time
            ISO8601TimeParser(str, h, m, s);
        }

        return DayTime(h, m, s, ns);
    }
    catch(Exception& ex) {
        // build a suitable error message
        std::wostringstream chainStr;
        chainStr << "DayTime::fromString(\"" << str << "\")";

        ex.chain(chainStr.str());
        throw;
    }
}



DayTime DayTime::now()
{
    struct timeval tv;
    gettimeofday(&tv, 0);
    time_t t = tv.tv_sec;
    struct tm tt;
    localtime_r(&t, &tt);
    return DayTime(tt.tm_hour, tt.tm_min, tt.tm_sec, tv.tv_usec * 1000);
}



DayTime& DayTime::addNanoSeconds(int64_t ns)
{
    lldiv_t dv = lldiv(ns, tenExp9);
    _s += dv.quot;
    _ns += dv.rem;
    int dummy;
    return wrap(dummy);
}



DayTime& DayTime::addNanoSeconds(int64_t ns, int& dayDelta)
{
    lldiv_t dv = lldiv(ns, tenExp9);
    _s += dv.quot;
    _ns += dv.rem;
    return wrap(dayDelta);
}



std::wostream& operator<<(std::wostream& ostr, const DayTime& d)
{
    ostr << d.toString();
    return ostr;
}



}
