/* lw-support/src/lib/Time/DayTime.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Represents a 24-hour day time.

This class is used to represent a day time (i.e. between 00:00:00 and
23:59:59). It will "wrap around" if you modify it so that it goes
outside this range. The function forms will allow you to retrieve the
number of days difference in such cases. There is no support for
leap seconds. The class also has a fractional second member (in
nanoseconds).

This class is intended for representing a specific time, not an elapsed
time. Use the ElapsedTime class for that.

*/
class DayTime {
private:
    int _h, _m, _s, _ns;

    DayTime& wrap(int& dayDelta);

public:
    //BEGIN // Constructors ////////////////////////////////////////////

    /*! \brief Construct from time.

    \param h The hour (0-23).
    \param m The minute (0-59).
    \param s The second (0-59).
    \param ns The nanosecond (a positive value &lt; 1s).
    \throws BadTimeValue if any value is out of range.

    */
    DayTime(int h, int m, int s = 0, int ns = 0)
        : _h(h), _m(m), _s(s), _ns(ns)
    {
        if(_h < 0 || _h > 23 || _m < 0 || _m > 59 || _s < 0 || _s > 59
            || _ns < 0 || _ns >= tenExp9)
                throw BadTimeValue();
    }

    /*! \brief Construct from elapsed time.

    \param time The ElapsedTime object to construct from.
    \throws BadTimeValue if \a time represents a period >= 86400s.

    */
    DayTime(const ElapsedTime& time)
    {
        uint64_t p = time.to_s();
        if(p >= 86400) throw BadTimeValue();
        _h = p / 3600;
        _m = (p / 60) % 60;
        _s = p % 60;
        _ns = time.to_ns() % tenExp9;
    }

    /// Copy constructor.
    DayTime(const DayTime& other)
        : _h(other._h), _m(other._m), _s(other._s), _ns(other._ns)
    { }

    /// Construct from calendar time.
    static DayTime now();



    //BEGIN // Setting functions ///////////////////////////////////////

    /*! \brief Set time.

    \param h The hour (0-23).
    \param m The minute (0-59).
    \param s The second (0-59).
    \param ns The nanosecond (a positive value &lt; 1s).
    \throws BadTimeValue if any value is out of range.
    \returns A reference to itself.

    */
    DayTime& set(int h, int m, int s = 0, int ns = 0)
    {
        if(h < 0 || h > 23 || m < 0 || m > 59 || s < 0 || s > 59
            || ns < 0 || ns >= tenExp9)
                throw BadTimeValue();
        _h = h;
        _m = m;
        _s = s;
        _ns = ns;
        return *this;
    }

    /// Set time.
    DayTime& operator=(const DayTime& other)
    {
        _h = other._h;
        _m = other._m;
        _s = other._s;
        _ns = other._ns;
        return *this;
    }

    /*! \brief Set time.

    \param time The ElapsedTime object to construct from.
    \throws BadTimeValue if \a time represents a period >= 86400s.
    \returns A reference to itself.

    */
    DayTime& operator=(const ElapsedTime& time)
    {
        uint64_t p = time.to_s();
        if(p >= 86400) throw BadTimeValue();
        _h = p / 3600;
        _m = (p / 60) % 60;
        _s = p % 60;
        _ns = time.to_ns() % tenExp9;
        return *this;
    }



    //BEGIN // Query functions /////////////////////////////////////////

    /// Return hour (0-23).
    int h() const
    { return _h; }

    /// Return minute (0-59).
    int m() const
    { return _m; }

    /// Return second (0-59).
    int s() const
    { return _s; }

    /// Return nanosecond (&lt;1s)
    int ns() const
    { return _ns; }

    /// Convert to an ElapsedTime object.
    ElapsedTime toElapsedTime() const
    {
        if(_ns) return ElapsedTime(_h * 3600 + _m * 60 + _s,
            _ns / 1000, _ns % 1000);
        else return ElapsedTime(_h * 3600 + _m * 60 + _s);
    }

    /// Convert into seconds since midnight.
    int numSeconds() const
    { return _h * 3600 + _m * 60 + _s; }

    /// Convert into nanoseconds since midnight.
    int64_t numNanoSeconds() const
    { return (_h * 3600 + _m * 60 + _s) * tenExp9 + _ns; }



    //BEGIN // Arithmetic functions ////////////////////////////////////

    /// Advance time.
    DayTime& addNanoSeconds(int64_t ns);

    /// Advance time.
    DayTime& addNanoSeconds(int64_t ns, int& dayDelta);

    /// Advance time.
    DayTime& addSeconds(int s)
    {
        int dummy;
        _s += s;
        return wrap(dummy);
    }

    /// Advance time.
    DayTime& addSeconds(int s, int& dayDelta)
    {
        _s += s;
        return wrap(dayDelta);
    }

    /// Advance time.
    DayTime& addMinutes(int m)
    {
        int dummy;
        _m += m;
        return wrap(dummy);
    }

    /// Advance time.
    DayTime& addMinutes(int m, int& dayDelta)
    {
        _m += m;
        return wrap(dayDelta);
    }

    /// Advance time.
    DayTime& addHours(int h)
    {
        int dummy;
        _h += h;
        return wrap(dummy);
    }

    /// Advance time.
    DayTime& addHours(int h, int& dayDelta)
    {
        _h += h;
        return wrap(dayDelta);
    }



    //BEGIN // String routines /////////////////////////////////////////

    /// Thrown when format specifiers don't make sense, etc.
    class BadFormat : public ProgramException { };



    /*! \brief Parse a std::string object to get a day time.

    \param str The std::string to parse.
    \throws ParseError if a parsing error occurs.
    \returns A new DayTime object constructed from the std::string data.

    Parses an ISO8601 std::string, accepting either basic or extended format
    notations, at any precision. Handles decimal notation as well.

    */
    static DayTime fromString(const std::wstring& str);



    /*! \brief Convert to std::string.

    \param extended True if you want the extended ISO8601
        representation; false if not.
    \param precision The precision to which you want to go. This takes
        a different meaning if \a decimal is true.
    \param decimal True if you want to use decimal time notation; false
        if not.
    \returns A UCS-4 std::string representation of the day time.

    Creates a std::string representation of the time in ISO8601 form. If you
    choose to use decimal time, then the full precision will be printed,
    but the \a precision argument specifies where to start the fraction.
    If you use lw::ISO8601PrecisionFull, and \a decimal is set to false,
    the decimal part will still be printed if and only if the number of
    nanoseconds is non-zero.

    */
    std::wstring toString(bool extended = true,
        ISO8601Precision precision = ISO8601PrecisionSecond,
        bool decimal = false) const;



    //BEGIN // Operators (comparison) //////////////////////////////////

    /// Returns true if \a this is equal to \a other.
    bool operator==(const DayTime& other) const
    { return _h == other._h && _m == other._m && _s == other._s; }

    /// Returns true if \a this is not equal to \a other.
    bool operator!=(const DayTime& other) const
    { return _h != other._h || _m != other._m || _s != other._s; }

    /// Returns true if \a this is earlier than \a other.
    bool operator<(const DayTime& other) const
    { return numNanoSeconds() < other.numNanoSeconds(); }

    /// Returns true if \a this is earlier than or equal to \a other.
    bool operator<=(const DayTime& other) const
    { return numNanoSeconds() <= other.numNanoSeconds(); }

    /// Returns true if \a this is later than \a other.
    bool operator>(const DayTime& other) const
    { return numNanoSeconds() > other.numNanoSeconds(); }

    /// Returns true if \a this is later than or equal to \a other.
    bool operator>=(const DayTime& other) const
    { return numNanoSeconds() >= other.numNanoSeconds(); }



    //BEGIN // Operators (arithmetic) //////////////////////////////////

    /// Advance time.
    DayTime& operator+=(int s)
    { return addSeconds(s); }

    /// Retard time.
    DayTime& operator-=(int s)
    { return addSeconds(-s); }

    /// Advanced time.
    DayTime operator+(int s) const
    {
        DayTime t(*this);
        return t.addSeconds(s);
    }

    /// Retarded time.
    DayTime operator-(int s) const
    {
        DayTime t(*this);
        return t.addSeconds(-s);
    }

    /// Returns time difference in nanoseconds.
    int64_t operator-(const DayTime& other) const
    { return numNanoSeconds() - other.numNanoSeconds(); }

};



/// Stream insertion operator.
std::wostream& operator<<(std::wostream&, const DayTime& d);



}
