/* lw-support/src/lib/Time/ElapsedTime.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/// Integer constant: 1e9.
const int64_t tenExp9 = 1000000000;

/// Integer constant: 1e6.
const int64_t tenExp6 = 1000000;

/// Integer constant: 1e3.
const int64_t tenExp3 = 1000;



/*! \brief An elapsed time period.

This class is used to represent an elapsed time period. It has a
resolution of nanoseconds, and uses a 64-bit integer (thus providing
a maximum representable period of 1.84e10 seconds, or about 584 years).
It is intended as a direct replacement for the C library's <tt>struct
timeval</tt> and <tt>struct timespec</tt>. It cannot represent negative
time periods.

\sa lw::StopWatch, lw::DayTime

*/
class ElapsedTime {
private:
    uint64_t time;

    // constructor with dummy argument to avoid collisions
    explicit ElapsedTime(uint64_t ns, bool)
        : time(ns)
    { }

public:
    //BEGIN // Constructors ////////////////////////////////////////////

    /*! \brief Default constructor.

    \param s The number of seconds (must be &gt;= 0)
    \param ms The number of milliseconds (must be &gt;= 0 and &lt; 1000)
    \param us The number of microseconds (must be &gt;= 0 and &lt; 1000)
    \param ns The number of nanoseconds (must be &gt;= 0 and &lt; 1000)
    \throws BadTimeValue if any one of the arguments is out of range.

    Constructs an elapsed time period. This constructor is designed for
    intuitive use -- the more arguments you specify, the more accurate
    the result. Seconds are the most significant argument.

    */
    ElapsedTime(int s = 0, int ms = 0, int us = 0, int ns = 0)
        : time(s * tenExp9 + ms * tenExp6 + us * tenExp3 + ns)
    {
        if(s < 0 || ms < 0 || ms >= tenExp3 || us < 0 || us >= tenExp3
            || ns < 0 || ns >= tenExp3)
        {
            throw BadTimeValue();
        }
    }

    /*! \brief Construct from floating point value in seconds.

    \param s The fractional number of seconds.
    \throws BadTimeValue if \a s is negative.

    Rounds the fractional time to the nearest nanosecond.

    */
    explicit ElapsedTime(long double s)
        : time(uint64_t(s * tenExp9))
    {
        if(s < 0) throw BadTimeValue();
    }

    /*! \brief Construct from floating point value in seconds.

    \param s The fractional number of seconds.
    \throws BadTimeValue if \a s is negative.

    Rounds the fractional time to the nearest nanosecond.

    */
    explicit ElapsedTime(double s)
        : time(uint64_t(s * tenExp9))
    {
        if(s < 0) throw BadTimeValue();
    }

    /*! \brief Construct from floating point value in seconds.

    \param s The fractional number of seconds.
    \throws BadTimeValue if \a s is negative.

    Rounds the fractional time to the nearest nanosecond.

    */
    explicit ElapsedTime(float s)
        : time(uint64_t(s * tenExp9))
    {
        if(s < 0) throw BadTimeValue();
    }



    /// Copy constructor.
    ElapsedTime(const ElapsedTime& other)
        : time(other.time)
    { }



    /*! \brief Construct from a struct timeval.

    \param tv The timeval to construct from.
    \throws BadTimeValue if \a tv's members are out of range.

    */
    ElapsedTime(struct timeval* tv);

    /*! \brief Construct from a struct spec.

    \param ts The timespec to construct from.
    \throws BadTimeValue if \a ts's members are out of range.

    */
    ElapsedTime(struct timespec* ts);



    /// Construct from nanoseconds.
    static ElapsedTime from_ns(uint64_t ns)
    { return ElapsedTime(ns, true); }

    /// Explicitly construct from microseconds.
    static ElapsedTime from_us(uint64_t us)
    { return ElapsedTime(us * tenExp3, true); }

    /// Explicitly construct from milliseconds.
    static ElapsedTime from_ms(uint64_t ms)
    { return ElapsedTime(ms * tenExp6, true); }



    //BEGIN // Query functions /////////////////////////////////////////

    /// Returns true if the time period is zero.
    inline bool isZero() const { return !time; }

    /// Returns the time in nanoseconds.
    uint64_t to_ns() const { return time; }

    /// Returns the time in microseconds.
    uint64_t to_us() const { return time / tenExp3; }

    /// Returns the time in milliseconds.
    uint64_t to_ms() const { return time / tenExp6; }

    /// Returns the time in seconds.
    uint64_t to_s() const { return time / tenExp9; }

    /// Returns fractional time.
    long double toLongDouble() const { return time / 1e9; }

    /// Returns fractional time.
    double toDouble() const { return time / 1e9; }

    /// Returns fractional time.
    float toFloat() const { return time / 1e9; }



    //BEGIN // Set functions ///////////////////////////////////////////

    /// Assignment operator.
    ElapsedTime& operator=(const ElapsedTime& other)
    {
        time = other.time;
        return *this;
    }

    /// Clears the elapsed time.
    ElapsedTime clear()
    {
        time = 0;
        return *this;
    }

    /// Sets the elapsed time (nanoseconds).
    ElapsedTime set_ns(uint64_t ns)
    {
        time = ns;
        return *this;
    }

    /// Sets the elapsed time (microseconds).
    ElapsedTime set_us(uint64_t us)
    {
        time = us * tenExp3;
        return *this;
    }

    /// Sets the elapsed time (milliseconds).
    ElapsedTime set_ms(uint64_t ms)
    {
        time = ms * tenExp6;
        return *this;
    }

    /// Sets the elapsed time (seconds).
    ElapsedTime set_s(uint64_t s)
    {
        time = s * tenExp9;
        return *this;
    }



    /*! \brief Sets the elapsed time from a fraction.

    \param s The fractional number of seconds.
    \throws BadTimeValue if \a s is negative.

    Rounds the fractional time to the nearest nanosecond.

    */
    ElapsedTime set(long double s)
    {
        if(s < 0) throw BadTimeValue();
        time = (uint64_t)s;
        return *this;
    }

    /*! \brief Sets the elapsed time from a fraction.

    \param s The fractional number of seconds.
    \throws BadTimeValue if \a s is negative.

    Rounds the fractional time to the nearest nanosecond.

    */
    ElapsedTime set(double s)
    {
        if(s < 0) throw BadTimeValue();
        time = (uint64_t)s;
        return *this;
    }

    /*! \brief Sets the elapsed time from a fraction.

    \param s The fractional number of seconds.
    \throws BadTimeValue if \a s is negative.

    Rounds the fractional time to the nearest nanosecond.

    */
    ElapsedTime set(float s)
    {
        if(s < 0) throw BadTimeValue();
        time = (uint64_t)s;
        return *this;
    }



    //BEGIN // String Routines /////////////////////////////////////////

    /*! \brief Returns a std::string representation of the elapsed time.

    \param decPlaces The number of decimal places.
    \param exp True if you want exponential (scientific) notation.

    This function will return a std::string representation of an elapsed time
    period. The formatting can be specified. Units (<tt>s, ms, us,
    ns</tt>) will be printed after the number. In scientific notation,
    the unit is always seconds.

    */
    std::wstring toString(int decPlaces = 1, bool exp = false) const;



    /*! \brief Constructs a time period from a std::string representation.

    \param str The std::string to interpret.
    \throws ParseError if a parsing error occurs.

    This function will try to interpret a std::string as a time period. It
    accepts an integral or floating point number (possibly in C's
    exponential notation), possibly followed by a unit. Acceptable
    units are <tt>s, ms, us, ns</tt>. If omitted, seconds are assumed.
    Any whitespace in the std::string will be ignored.

    */
    static ElapsedTime fromString(const std::wstring& str);



    //BEGIN // Operators (comparison) //////////////////////////////////

    /// Returns true if the time period is zero.
    inline bool operator!() const { return !time; }

    /// Returns true if the time period is non-zero.
    inline operator bool() const { return time; }

    /// Returns true if two time periods are equal.
    inline bool operator==(const ElapsedTime& other) const
    { return time == other.time; }

    /// Returns true if two time periods are not equal.
    inline bool operator!=(const ElapsedTime& other) const
    { return time != other.time; }

    /// Less than comparison.
    inline bool operator<(const ElapsedTime& other) const
    { return time < other.time; }

    /// Less than or equal to comparison.
    inline bool operator<=(const ElapsedTime& other) const
    { return time <= other.time; }

    /// Greater than comparison.
    inline bool operator>(const ElapsedTime& other) const
    { return time > other.time; }

    /// Greater than or equal to comparison.
    inline bool operator>=(const ElapsedTime& other) const
    { return time >= other.time; }



    //BEGIN // Operators (arithmetic) //////////////////////////////////

    /// Addition operation.
    inline ElapsedTime& operator+=(const ElapsedTime& other)
    {
        time += other.time;
        return *this;
    }

    /*! \brief Subtraction operation.

    \param other The time period to subtract from this one.
    \throws BadTimeOverflow if \a other is a greater time period than
        \a this
    \returns A reference to the newly-modified object.

    */
    inline ElapsedTime& operator-=(const ElapsedTime& other)
    {
        if(other.time > time)
            throw BadTimeOverflow();
        time -= other.time;
        return *this;
    }

    /// Addition operation.
    inline ElapsedTime operator+(const ElapsedTime& other) const
    {
        return ElapsedTime(time + other.time, true);
    }

    /*! \brief Subtraction operation.

    \param other The time period to subtract from this one.
    \throws BadTimeOverflow if \a other is a greater time period than
        \a this
    \returns The time period difference.

    */
    inline ElapsedTime operator-(const ElapsedTime& other) const
    {
        if(other.time > time)
            throw BadTimeOverflow();
        return ElapsedTime(time - other.time, true);
    }
};



/// Stream insertion operator.
std::wostream& operator<<(std::wostream&, const ElapsedTime&);



}
