/* lw-support/src/lib/Util/Completion.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/// Task completion callback data base class.
class CompletionData {
public:
    /// Destructor (does nothing).
    virtual ~CompletionData()
    { }
};



/*! \brief Completion callback interface.

Certain tasks use a callback to signal when they have been completed. This interface supports this
by providing a common callback function. To implement this, you derive your signal's target class
from lw::CompletionCallback and override the two functions.

The task class must take a pointer to a CompletionCallback instance as a parameter. When its task is
done, it will call the completed() function; if the task must be stopped early, it will call the
aborted() function. This function also provides an indication of what caused the task to abort. The
task class may be derived from CompletionTask, but this is not a requirement.

It is possible to associate arbitrary data with the task. This must be generated in or passed to the
task object. It will then be passed as the \a data parameter to the callback function. The data
%class must be derived from the lw::CompletionData class.

*/
class CompletionCallback {
public:
    /// Destructor (does nothing).
    virtual ~CompletionCallback()
    { }



    /*! \brief Task complete signal.

    \param data Arbitrary data associated with the task.

    This function is called to signal that the task has been completed, presumably successfully.
    If any data was associated with the task, it will be passed in through the \a data pointer,
    although it would be valid for this to be zero.

    */
    virtual void completed(CompletionData* data) = 0;



    /*! \brief Task aborted signal.

    \param data Arbitrary data associated with the task.
    \param e The exception that caused the task to abort.

    This function is called to signal that the task has been aborted without completing.
    If any data was associated with the task, it will be passed in through the \a data pointer,
    although it would be valid for this to be zero.

    */
    virtual void aborted(CompletionData* data, const lw::Exception& e) = 0;
};



/// Task already completed.
class TaskAlreadyCompleted : public lw::ProgramException { };



/*! \brief Completion task suggested interface.

This class is the suggested (but not required) interface for a task that uses the completion
notifier interface provided by CompletionCallback. It is a safe and effective way of ensuring that
all required completion events are generated and only one event is ever generated. It is also the
required interface for the CompletionList interface.

Instances of this class will pass a pointer to themselves as the \a data parameter to the callback
functions. If the callback class throws an exception, this will be propagated back to the function
that generated the signal (or discarded if it was the destructor).

*/
class CompletionTask : public CompletionData {
private:
    CompletionCallback* callback;

protected:
    /*! \brief Report task completed.

    \throws lw::TaskAlreadyCompleted if the task is already done.

    This function will generate the signal to the callback class. The data passed will be a pointer
    to the instance itself. Once signalled, no further signal can be sent. Any attempt to send more
    than one signal will result in an lw::TaskAlreadyCompleted exception.

    \sa lw::CompletionCallback::completed()

    */
    void taskCompleted();



    /*! \brief Report task aborted.

    \param e Exception giving reason for task abort.
    \throws lw::TaskAlreadyCompleted if the task is already done.

    This function will generate the signal to the callback class. The data passed will be a pointer
    to the instance itself. Once signalled, no further signal can be sent. Any attempt to send more
    than one signal will result in an lw::TaskAlreadyCompleted exception.

    \sa lw::CompletionCallback::completed()

     */
    void taskAborted(const lw::Exception& e);



public:
    /*! \brief Constructor.

    \param callback The callback object.
    \throws lw::NullPointer if \a callback is zero.

    Stores the callback object for later signalling.

    */
    CompletionTask(CompletionCallback* callback);



    /*! \brief Destructor. Aborts task if not completed.

    The destructor will check to see if the task has been completed or not. If it has, then it will
    take no action; if it has not, the task will be aborted with a TaskAborted exception.

    */
    virtual ~CompletionTask();



    /*! \brief Abort the task.

    \param e Exception giving reason for task abort.
    \throws lw::TaskAlreadyCompleted if the task is already done.

    This function is called by the destructor if the task is still in progress when the object is
    destroyed. It may also be made available to other functions in your class and potentially users
    of your class (lw::CompletionList for example).

    On calling this function, the task must be aborted and the taskAborted() signal must be sent
    before returning from the function.

    */
    virtual void taskAbort(const lw::Exception& e) = 0;
};



/*! \brief List of tasks to be completed.

This class is a useful utility which maintains a list of tasks in progress. As each task is
completed or aborted, it is removed from the list. When the object is destroyed, or when abortAll()
is called, it will call the lw::CompletionTask::taskAbort() function for every task still in the
list.

This class is not thread-safe. A set of three debug functions are provided to cater for
unexpected conditions. dbg_taskNotAborted() is called if lw::CompletionTask::abortTask() does not
result in an immediate call to completed() or aborted(). dbg_dataNotTask() is called if completed()
or aborted() are passed a CompletionData pointer that is not a CompletionTask pointer.
dbg_taskNotActive() is called if completed() or aborted() are passed a CompletionTask pointer that
is not in the active task list. Their default behaviour is to print an error message through
lw::stderr, but they can be overridden.

\todo We use an unsorted list here; sort it.

*/
class CompletionList : public CompletionCallback,
    private Uncopyable
{
private:
    std::list<CompletionTask*> tasks;
    CompletionTask* abortingTask;
    typedef std::list<CompletionTask*>::iterator taskIter;

protected:
    /// Debug function.
    virtual void dbg_taskNotAborted(CompletionTask* task);
    /// Debug function.
    virtual void dbg_dataNotTask(CompletionData* data);
    /// Debug function.
    virtual void dbg_taskNotActive(CompletionTask* task);

public:
    /// Constructor.
    CompletionList();



    /*! \brief Destructor. Aborts all active tasks.

    The destructor will abort all active tasks by calling abortAll() if necessary.

    */
    virtual ~CompletionList();



    /*! \brief Register a new task.

    \param task The task to be completed.
    \throws lw::NullPointer if \a task is zero.

    Registers a new task for completion. The task must have been instantiated to use this
    list instance as its callback (or at least the callback signals must reach this object). If the
    task is still active when the list instance is destroyed, or abortAll() is called, the task's
    own abort function will be called.

    \note There is no guarding against adding the same task twice. Doing so will break things.

    */
    void registerTask(CompletionTask* task);



    /*! \brief Abort all active tasks.

    Attempts to abort all active tasks by calling lw::CompletionTask::abortTask() for each such
    task. Any exception thrown from this function will be ignored; however, if there are any tasks
    left unregistered at the end of the function, dbg_taskNotAborted() will be called for each
    remaining task.

    */
    void abortAll();



    /// Returns true if there are no active tasks.
    bool empty();



    // implemented virtuals from lw::CompletionCallback
    virtual void completed(CompletionData* data);
    virtual void aborted(CompletionData* data, const lw::Exception& e);
};



}
