/* lw-support/src/lib/Util/FIFO.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief FIFO memory buffer.

This class is an implementation of a FIFO memory buffer. It has an
absolute maximum upper limit on how much memory has been consumed, and
you can specify how much of that limit is initially allocated. The
functions are all inline for speed.

*/
class FIFO {
private:
    // These refer to the block of memory that is allocated.
    char* buffer;
    size_t bufferSize, maxSize;

    // These refer to the data in the buffer.
    char* data; // set to zero if dataSize == 0
    size_t dataSize;

    // Reallocate buffer ('required' is the new total buffer size that
    // is needed).
    void reallocBuffer(size_t required);

    // Fast realloc check, also moves data if that would save an
    // allocation.
    inline void ensureBuffer(size_t required)
    {
        if(required > bufferSize) {
            reallocBuffer(required);
            return;
        }

        if(data && (data + required) > (buffer + bufferSize)) {
            memmove(buffer, data, dataSize);
            data = buffer;
            return;
        }
    }

public:
    //BEGIN // Constructors etc. ///////////////////////////////////////

    /*! \brief Constructor.

    \param hint This hints at how much memory should be allocated.
    \param maxSize The absolute maximum number of bytes to allocate.

    Sets up the FIFO buffer. The absolute maximum size must be specified
    here (it can be changed later). A hint as to how much memory to
    initially allocate must also be provided.

    */
    FIFO(size_t hint, size_t maxSize)
        : buffer((hint && maxSize) ? new char[std::min(hint, maxSize)] : 0),
          bufferSize(hint), maxSize(maxSize),
          data(0), dataSize(0)
    { }

    /// Destructor (frees allocated memory).
    ~FIFO()
    { delete [] buffer; }



    /*! \brief Change absolute maximum size.

    \param newMaxSize The buffer's new absolute maximum size in bytes.
    \throws FIFOOverflow if the buffer currently contains more than
        \a newMaxSize bytes.

    This function will change the upper size limit of the FIFO memory
    buffer. It can be used to make the buffer larger at any time. It can
    also be used to make the buffer smaller (possibly freeing up memory
    in the process); but you cannot make the buffer so small that its
    contents no longer fit.

    */
    inline void setMaxSize(size_t newMaxSize)
    {
        if(newMaxSize < dataSize) throw FIFOOverflow();
        maxSize = newMaxSize;

        // might have to reallocate the buffer if it is too big
        if(bufferSize > newMaxSize) reallocBuffer(newMaxSize);
    }



    /// Query the absolute maximum size.
    inline size_t getMaxSize() const
    {
        return maxSize;
    }



    //BEGIN // Buffering functions /////////////////////////////////////

    /*! \brief Gets ready to add some data.

    \param amt The maximum amount of data the buffer will be required to
        accept.
    \returns A pointer to the memory location at which the data should
        be written.
    \throws FIFOOverflow if the buffer is too small.

    This function ensures that the buffer has enough space to accept at
    least \a amt bytes of additional data. This might require allocating
    more memory or moving the current buffer data around. If you call
    this twice in a row, then the first call is totally forgotten. It
    doesn't change the object's internal state; it just makes sure that
    you can add \a amt bytes if you want to, and tells you where.

    \sa addedData()

    */
    inline char* getBuffer(size_t amt)
    {
        ensureBuffer(amt + dataSize);
        return data ? data + dataSize : buffer;
    }



    /*! \brief Records how much data was added to the buffer.

    \param amt The amount of data that was actually added to the buffer.

    Once you have prepared the buffer with getBuffer(), you must add the
    data. This function records how much data was added (so that new
    data will be added after it).

    \warning No checking is done on \a amt, so make sure you put the
        right amount. Also make sure that getBuffer() and addedData()
        calls occur in pairs.

    \sa getBuffer(), addData()

    */
    inline void addedData(size_t amt)
    {
        if(!data) data = buffer;
        dataSize += amt;
    }



    /*! \brief Adds some data to the buffer.

    \param amt The number of bytes to add.
    \param newData The data to add.
    \throws FIFOOverflow if the buffer is too small.

    Adds a specific block of data to the buffer. This can only be used
    if the data is coming from somewhere else in memory (e.g. an
    mmap()ed file), and you know in advance how much there is.

    \sa addedData()

    */
    inline void addData(const char* newData, size_t amt)
    {
        ensureBuffer(amt + dataSize);
        if(data) {
            memcpy(data + dataSize, newData, amt);
            dataSize += amt;
        } else {
            memcpy(data = buffer, newData, amt);
            dataSize = amt;
        }
    }



    /*! \brief Adds a byte of data to the buffer.

    \param byte The byte of data to add.
    \throws FIFOOverflow if the buffer is too small.

    Adds a single byte of data to the buffer. This is the most efficient
    method for byte-oriented buffering.

    */
    inline void addByte(char byte)
    {
        ensureBuffer(dataSize + 1);
        if(data) {
            data[dataSize++] = byte;
        } else {
            *(data = buffer) = byte;
            dataSize = 1;
        }
    }



    //BEGIN // Reading functions ///////////////////////////////////////

    /// Returns how many bytes of data there are in the buffer.
    inline size_t getSize() const
    {
        return dataSize;
    }

    /// Returns how many bytes of data there are in the buffer.
    inline size_t size() const
    {
        return dataSize;
    }



    /// Returns how many bytes are available to the buffer.
    inline size_t getRemaining() const
    {
        return maxSize - dataSize;
    }



    /// Returns how many (preallocated) bytes are available to the buffer.
    inline size_t getRemainingPreallocated() const
    {
        return bufferSize - dataSize;
    }



    /// Returns pointer to start of buffer (0 if empty).
    inline const char* getData() const
    {
        return data;
    }



    /*! \brief Retrieves some data from the buffer.

    \param amt The number of bytes to retrieve.
    \throws FIFOUnderflow if the buffer contains less than \a amt bytes.
    \returns A pointer to the start of the data (0 if \a amt = 0).

    This function allows you to get data out of the buffer. It will
    record how many bytes were used, so that space will be free the next
    time anyone wants to use it.

    \warning The pointer returned may become invalid as soon as you
        perform any other operation on the FIFO, so you must finish
        using the data it points to before calling any other functions.

    \sa getSize()

    */
    inline char* getData(size_t amt)
    {
        // sanity checking
        if(!amt) return 0;
        if(amt > dataSize) throw FIFOUnderflow();

        // eat data
        char* result = data;
        if(!(dataSize -= amt)) data = 0;
        else data += amt;
        return result;
    }



    /*! \brief Gets a single byte of data from the buffer.

    \returns The byte of data removed from the buffer.
    \throws FIFOUnderflow if the buffer is empty.
    \returns A single byte of data.

    Retrieves a single byte of data from the buffer. This is the most
    efficient method for byte-oriented buffering.

    */
    inline char getByte()
    {
        if(!dataSize) throw FIFOUnderflow();
        char byte = *data;
        if(!--dataSize) data = 0;
        else ++data;
        return byte;
    }



    //BEGIN // Convenience functions ///////////////////////////////////

    /// Returns true if the buffer is empty.
    inline bool isEmpty() const
    {
        return !dataSize;
    }

    /// Returns true if the buffer is empty.
    inline bool empty() const
    {
        return !dataSize;
    }

    /// Returns true if the buffer is full.
    inline bool isFull() const
    {
        return dataSize == maxSize;
    }

    /// Empties the buffer.
    inline void clear()
    {
        data = 0;
        dataSize = 0;
    }
};



}
