/* lw-support/src/lib/Util/Key.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \brief Key object.

The key object is used to store some form of identifying key. Typically this will be used to store a
unique key number.

\note We use a 64-bit integer to store the key. It is somewhat conceivable that this could be
    wrapped around at some point. If you're really generating keys that frequently, don't use this
    class and instead write a 128-bit or wider class.

*/
class Key {
public:
    /// Constructor. Generates an invalid key.
    Key()
        : key(0)
    { }

    /// Copy constructor.
    Key(const Key& other)
        : key(other.key)
    { }

    /// Assignment operator.
    Key& operator=(Key other)
    { key = other.key; return *this; }

    /// Compares two keys for equality.
    bool operator==(Key other) const
    { return key == other.key; }

    /// Compares two keys for inequality.
    bool operator!=(Key other) const
    { return key != other.key; }

    /// Allows ordering by key.
    bool operator<(Key other) const
    { return key < other.key; }

    /// Returns status of key.
    bool operator!() const
    { return !key; }

    /// Returns status of key.
    operator void*() const
    { return key ? (void*)this : 0; }

    /// Returns status of key.
    bool valid() const
    { return key; }

    /// Makes the key invalid.
    void clear()
    { key = 0; }



    /*! \brief Represent the key as a string.

    \returns A string representing the value of the key.

    This function allows the key to be represented as a string. The string will always be started
    with the sequence \c {KEY{ and it will always end with \c }} but its contents are implementation
    defined. The string generated will consist only of ASCII characters and all characters will be
    printable.

    */
    std::wstring toString() const;

    /// Construct a key from a string.
    static lw::Key fromString(const std::wstring& str);

private:
    typedef uint_fast64_t keyType;
    keyType key;

    // Explicitly construct a key from a value.
    Key(keyType key) : key(key) { }

    static keyType globalKey;

public:
    /// Generate a program-wide unique key.
    static Key get()
    { ++globalKey; return Key(globalKey); }
};



}
