/* lw-support/src/lib/Util/Log.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



Log stderr(std::wcerr, true);



Log::Log(std::wostream& outStream, bool enabled)
  : out(outStream), wstr(0), enabled(enabled), streamWidth(78),
    dateExtended(true), dateDecimal(false), dateFormat(lw::ISO8601FormatCalendar),
    datePrecision(lw::ISO8601PrecisionFull)
{
}



Log::Log(const std::string& fileName, bool enabled)
  : out(file), wstr(0), enabled(enabled), streamWidth(78),
    dateExtended(true), dateDecimal(false), dateFormat(lw::ISO8601FormatCalendar),
    datePrecision(lw::ISO8601PrecisionFull)
{
    file.open(fileName.c_str());
    // FIXME: better exception required
    if(!file) throw lw::FileNotFound(fileName.c_str())
        .chain(L"Log::Log()");
}



Log::~Log()
{
    delete wstr;
}



void Log::checkBox()
{
    if(!wstr) throw BoxException();
}



Log& Log::setStreamWidth(size_t streamWidth)
{
    this->streamWidth = streamWidth;
    return *this;
}



Log& Log::setDateFormat(bool dateExtended, bool dateDecimal,
    lw::ISO8601Format dateFormat, lw::ISO8601Precision datePrecision)
{
    this->dateExtended = dateExtended;
    this->dateDecimal = dateDecimal;
    this->dateFormat = dateFormat;
    this->datePrecision = datePrecision;
    return *this;
}



Log& Log::startBox(const std::wstring& subject)
{
    if(wstr) throw BoxException();
    wstr = new std::wostringstream();

    wstr->clear();
    *wstr << L"==== "
          << DateTime::now().toString(dateExtended, dateDecimal,
                 dateFormat, datePrecision)
          << L" ====";

    int remaining = streamWidth - (wstr->str().length() + subject.length() + 6);
    if(remaining > 0) *wstr << std::wstring(remaining, '=');
    *wstr << L' '
          << subject
          << L" ====\n\n";

    return *this;
}



Log& Log::enable(bool enabled)
{
    this->enabled = enabled;
    return *this;
}



Log& Log::disable()
{
    enabled = false;
    return *this;
}



Log& Log::endBox()
{
    checkBox();
    if(enabled) out << wstr->str() << L"\n\n" << std::endl;
    delete wstr;
    wstr = 0;
    return *this;
}



Log& Log::abortBox()
{
    delete wstr;
    wstr = 0;
    return *this;
}



Log& Log::hexDumpData(const char* data, size_t amt)
{
    checkBox();

    const size_t blocksPerLine = std::max(streamWidth / 34, size_t(1)),
        charsPerLine = blocksPerLine * 8,
        upperBlockSize = charsPerLine * ((amt + charsPerLine - 1) / charsPerLine);
    size_t blocksThisLine = 0, lineStartPos = 0;
    wchar_t ascii[charsPerLine + 1];
    ascii[charsPerLine] = 0;

    *wstr << L"Hex dump, "
          << amt
          << " bytes:\n"
          << std::hex
          << std::setfill(L'0');

    for(size_t pos = 0; pos < amt; ++pos) {
        uint8_t ch = data[pos];
        if(!(pos & 7)) {
            *wstr << L' ';
            if(blocksThisLine++ == blocksPerLine) {
                blocksThisLine = 1;
                *wstr << ascii << L"\n ";
                lineStartPos = pos;
            }
        }
        *wstr << std::setw(2) << uint(ch) << L' ';
        ascii[pos - lineStartPos] = wchar_t(isprint(ch) ? ch : '.');
    }

    for(size_t pos = amt; pos < upperBlockSize; ++pos) {
        if(!(pos & 7)) {
            *wstr << L' ';
            if(blocksThisLine++ == blocksPerLine) break;
        }
        *wstr << L"   ";
        ascii[pos - lineStartPos] = ' ';
    }

    *wstr << L' ' << ascii << L'\n';
    return *this;
}



Log& Log::asciiDumpData(const char* data, size_t amt)
{
    checkBox();

    *wstr << L"- ASCII dump, "
          << amt
          << " bytes: -----------------\n"
          << std::hex
          << std::setfill(L'0');

    for(size_t pos = 0; pos < amt; ++pos) {
        switch(data[pos]) {
            case '\r':
                *wstr << L"\\r";
                continue;

            case '\n':
                *wstr << L"\\n\n";
                continue;

            case '\t':
                *wstr << L"\\t";
                continue;
        }

        if(isprint(data[pos])) {
            *wstr << wchar_t(data[pos]);
        } else {
            *wstr << L'<'
                  << std::setw(2)
                  << uint(uint8_t(data[pos]))
                  << L'>';
        }
    }

    *wstr << L"\n- End of dump ---------------------\n";
    return *this;
}



Log::LogManipArgStart Log::start(const std::wstring& subject)
{
    return LogManipArgStart(subject);
}



Log::LogManipArgHex Log::hexDump(const char* data, size_t amt)
{
    return LogManipArgHex(data, amt);
}



Log::LogManipArgAscii Log::asciiDump(const char* data, size_t amt)
{
    return LogManipArgAscii(data, amt);
}



template<>
    Log& operator<<(Log& log, const Log::LogManip& l)
{
    log.checkBox();

    switch(l) {
    case Log::end:
        log.endBox();
        break;

    case Log::abort:
        log.abortBox();
        break;
    }

    return log;
}



template<>
    Log& operator<<(Log& log, const lw::Log::LogManipArgStart& l)
{
    log.startBox(l.subject);
    return log;
}




template<>
    Log& operator<<(Log& log, const lw::Log::LogManipArgHex& l)
{
    log.hexDumpData(l.data, l.amt);
    return log;
}



template<>
    Log& operator<<(Log& log, const lw::Log::LogManipArgAscii& l)
{
    log.asciiDumpData(l.data, l.amt);
    return log;
}



template<>
    Log& operator<<(Log& log, const char* const& str)
{
    if(!str) {
        *(log.wstr) << L"{{{null}}}";
        return log;
    }

    try {
        // attempt UTF-8 conversion
        *(log.wstr) << lw::utf8ToUcs4(str);
    }
    catch(...) {
        // `safe' conversion
        *(log.wstr) << L"{{{unknown encoding}}}";
        for(const char* q = str; *q; ++q) {
            switch(*q) {
            case '\n':
            case '\t':
            case ' ' ... '~': // 32 ... 126
                log.wstr->put(*q);
                break;

            default:
                log.wstr->put('?');
            }
        }
    }

    return log;
}



template<>
    Log& operator<<(Log& log, const std::string& str)
{
    return (log << str.c_str());
}



lw::Log& operator<<(lw::Log& log, const std::exception& e)
{
    return (log << e.what());
}



lw::Log& operator<<(lw::Log& log, const lw::Exception& e)
{
    std::cerr << L"[DEBUG] in lw::Exception-specific insertion operator\n";
    return (log << e.toString());
}



std::wostream& operator<<(std::wostream& ostr, const std::exception& e)
{
    ostr << lw::strToUcs4(e.what());
    return ostr;
}



}


/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
*/
