/* lw-support/src/lib/Util/Log.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lw {



/*! \defgroup Log Logging functions.

The logging functions provide a convenient way to log arbitrary data to an arbitrary output stream.
Log entries are broken into boxes and each entry is timestamped. The class that allows this is
lw::Log; an overloaded stream insertion operator allows you to treat it as an STL stream (and also
allows additional lw data types to be output to the stream).

*/
/*!@{*/



/*! \brief Overloaded stream insertion operator required for convenience.

This operator will insert a std::exception into a std::ostream. Since it is safely ensconsed in the
lw namespace, it won't affect code outside that namespace. It is required because the template
function operator<< will call it when outputting a std::exception-derived class.

*/
std::wostream& operator<<(std::wostream&, const std::exception&);



// forward declare of output functions
template<class T>
    Log& operator<<(Log& log, const T& t);



/*! \brief Log object.

This object provides a simple and easy-to-use interface for logging to a
standard I/O stream. It uses the concept of a "message box". Each box
consists of a header (date, time and subject). Boxes are separated by
three blank lines.

The object allows output of built-in types and any of the LW Support
objects that have a toString() method. It converts any <tt>const char*</tt>
or <tt>std::string</tt> objects into UCS-4 first -- it does this by first assuming that they are
UTF-8 and, if the conversion process fails, falls back to ASCII (substituting non-printable
characters with a question mark).

Logs can be tied to any standard output stream. By default, logs are
turned off (the output is not produced); use enable() to turn on the
output, and disable() to turn it off again. You can have as many logs as
you like pointing to a single stream; messages are only printed once
they have been completely assembled.

You must start a new box with startBox() (or the start manipulator). If
you are already in a box, a BoxException will be thrown. If you try to
log some output when not in a box, a BoxException will also be thrown.

\note <tt>std::endl</tt> and <tt>std::flush</tt> are not supported.

*/
class Log : public lw::Uncopyable {
private:
    // output streams
    std::wofstream file;
    std::wostream& out;
    std::wostringstream* wstr;
    bool enabled;
    size_t streamWidth;

    // for checking if we are in a box
    void checkBox();

    // for formatting dates
    bool dateExtended, dateDecimal;
    lw::ISO8601Format dateFormat;
    lw::ISO8601Precision datePrecision;

    // for STL-like output
    template<class T> friend
        Log& operator<<(Log& log, const T& t);

public:
    //BEGIN // Constructors etc. ///////////////////////////////////////



    /*! \brief Constructor (outputs to existing stream).

    \param outStream The output stream to send logs to.
    \param enabled Whether or not the object is enabled as soon as it is
        constructed; default is \a false.

    Sets up the logging object so that it will output to an existing
    output stream. The default stream width is set to 78 characters.

    */
    Log(std::wostream& outStream, bool enabled = false);



    /*! \brief Constructor (opens a file).

    \param fileName The file to write to.
    \param enabled Whether or not the object is enabled as soon as it is
        constructed; default is \a false.

    \throws lw::Exception if the file cannot be opened.

    Sets up the logging object so that it will output to an existing
    output stream.

    */
    Log(const std::string& fileName, bool enabled = false);



    /// Destructor.
    ~Log();



    //END // Constructors etc. /////////////////////////////////////////
    //BEGIN // Control functions ///////////////////////////////////////



    /// Thrown when a boxing error occurs.
    class BoxException : public ProgramException { };



    /// For funky STL-like manipulators (without arguments).
    enum LogManip {
        end,
        abort
    };



    /// Set stream width (e.g. for screen).
    Log& setStreamWidth(size_t streamWidth);

    /*! Set date conversion options.

    \param extended \a true (the default) selects ISO8601 extended
        format; \a false selects ISO8601 basic.
    \param decimalTime \a true if you want to print the time as a
        decimal fraction; \a false (the default) if you don't.
        Specifying full precision will get you decimal time anyway.
    \param dateFormat The calendar format to use. See lw-support.
    \param precision The precision to use. See lw-support.

    Sets options for formatting the date and time at the top of each
    box. See lw::DateTime::toString() for full details.

    */
    Log& setDateFormat(bool extended = true, bool decimalTime = false,
        lw::ISO8601Format dateFormat = lw::ISO8601FormatCalendar,
        lw::ISO8601Precision precision = lw::ISO8601PrecisionFull);



    /// Enable output to stream.
    Log& enable(bool enabled = true);

    /// Disable output to stream.
    Log& disable();

    /// Start a new box.
    Log& startBox(const std::wstring& subject);

    /// End current box.
    Log& endBox();

    /// Abort current box.
    Log& abortBox();



    //END // Control functions /////////////////////////////////////////
    //BEGIN // Convenience functions ///////////////////////////////////



    /*! \brief Perform a hex dump.

    \param data Pointer to the data to write.
    \param amt Number of bytes to write.

    Presents the data in a convenient, easy-to-read hex format. Prints
    a newline after the data, but not before. Prints a header with the
    block size in bytes.

    */
    Log& hexDumpData(const char* data, size_t amt);



    /*! \brief Perform an ASCII dump.

    \param data Pointer to the data to write.
    \param amt Number of bytes to write.

    Prints ASCII data. It will show newlines and tabs with C-style \\r,
    \\n and \\t sequences. Hex sequences are shown in angle brackets for
    other non-printable characters. Prints
    a newline after the data, but not before. Prints a header with the
    block size in bytes.

    */
    Log& asciiDumpData(const char* data, size_t amt);



    //END // Convenience functions /////////////////////////////////////
    //BEGIN // Complex manipulators ////////////////////////////////////



    /// Manipulator object for starting a new box.
    class LogManipArgStart {
        const std::wstring& subject;
        LogManipArgStart(const std::wstring& subject)
            : subject(subject)
        { }

        friend class Log;
        template<class T> friend
            Log& operator<<(Log& log, const T& t);
    };

    /// Start a new box (manipulator).
    static LogManipArgStart start(const std::wstring& subject);



    /// Manipulator object for performing a hex dump.
    class LogManipArgHex {
        const char* data;
        size_t amt;
        LogManipArgHex(const char* data, size_t amt)
            : data(data), amt(amt)
        { }

        friend class Log;
        template<class T> friend
            Log& operator<<(Log& log, const T& t);
    };

    /*! \brief Perform a hex dump (manipulator).

    \param data Pointer to the data to write.
    \param amt Number of bytes to write.

    Presents the data in a convenient, easy-to-read hex format. Prints
    a newline after the data, but not before. Prints a header with the
    block size in bytes.

    */
    static LogManipArgHex hexDump(const char* data, size_t amt);



    /// Manipulator object for performing an ASCII dump.
    class LogManipArgAscii {
        const char* data;
        size_t amt;
        LogManipArgAscii(const char* data, size_t amt)
            : data(data), amt(amt)
        { }

        friend class Log;
        template<class T> friend
            Log& operator<<(Log& log, const T& t);
    };

    /*! \brief Perform an ASCII dump (manipulator).

    \param data Pointer to the data to write.
    \param amt Number of bytes to write.

    Prints ASCII data. It will show newlines and tabs with C-style \\r,
    \\n and \\t sequences. Hex sequences are shown in angle brackets for
    other non-printable characters. Prints
    a newline after the data, but not before. Prints a header with the
    block size in bytes.

    */
    static LogManipArgAscii asciiDump(const char* data, size_t amt);



    //END // Complex manipulators //////////////////////////////////////
};



/*! \brief Standard error log.

Some of the library functions may wish to log error conditions on stdout. They will do so through
this log object. This log object will point at std::wcerr and is on by default. It should only be
disabled if std::wcerr is closed.

*/
extern lw::Log stderr;



/*! \brief STL-like output for lw::Log.

This overloaded output operator provides STL-like output (stream insertion) for the lw::Log object.
Several specialised functions are provided to allow output of common lw objects.

\param log The log object to output onto.
\param t The object to output.

*/
template<class T>
    lw::Log& operator<<(lw::Log& log, const T& t)
{
    log.checkBox();
    *(log.wstr) << t;
    return log;
}



/// Stream insertion operator for lw::Log::LogManip.
template <>
    lw::Log& operator<<(lw::Log&, const lw::Log::LogManip&);



/// Stream insertion operator for lw::Log::start().
template<>
    lw::Log& operator<<(lw::Log&, const lw::Log::LogManipArgStart&);



/// Stream insertion operator for lw::Log::hexDump().
template<>
    lw::Log& operator<<(lw::Log&, const lw::Log::LogManipArgHex&);



/// Stream insertion operator for lw::Log::asciiDump().
template<>
    lw::Log& operator<<(lw::Log&, const lw::Log::LogManipArgAscii&);



/// Stream insertion operator for an 8-bit string.
template<>
    lw::Log& operator<<(lw::Log&, const char* const&);



/// Stream insertion operator for an 8-bit string.
template<>
    lw::Log& operator<<(lw::Log&, const std::string&);



/*!@}*/
}
