/* lw-support/src/tests/Event_Timer.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

#include "lw-support"

#include <iostream>
#include <iomanip>

// some log objects, for luck
lw::Log Log(std::wcout, true), Err(std::wcerr, true);



// display callback object
class Display : private lw::EventCallbackTimer {
private:
    class Counter : public lw::EventCallbackTimer {
    private:
        std::map<lw::Key, int> counts;
        lw::StopWatch stopWatch;

        std::wstring toHz(int counts, uint64_t ns_elapsed)
        {
            std::wostringstream str;
            double freq = counts;
            freq /= (ns_elapsed * 1e-9);
            str << std::setprecision(3);
            if(freq < 0.5e-3) str << (freq * 1e6) << "uHz";
            else if(freq < 0.5) str << (freq * 1e3) << "mHz";
            else if(freq < 0.5e3) str << (freq) << "Hz";
            else if(freq < 0.5e6) str << (freq / 1e3) << "kHz";
            else if(freq < 0.5e9) str << (freq / 1e6) << "MHz";
            else str << freq << "Hz";
            return str.str();
        }

    public:
        virtual ~Counter()
        {
        }

        void dump()
        {
            Log << lw::Log::start(L"Counter")
                << L"Counters:\n";

            for(std::map<lw::Key, int>::const_iterator iter = counts.begin(),
                end = counts.end(); iter != end; ++iter)
            {
                Log << iter->first.toString()
                    << L", counts = "
                    << iter->second
                    << L", estimated frequency = "
                    << toHz(iter->second, stopWatch.elapsed().to_ns())
                    << L'\n';
            }

            Log << lw::Log::end;
        }

        virtual void timer(lw::Key key) throw()
        {
            if(!key) std::wcout << L"[DEBUG] Invalid key!\n";
            ++counts[key];
            // TODO de-register if we have over a million hits
        }
    };

    Counter c;

    virtual void timer(lw::Key key) throw()
    {
        if(!key) std::wcout << L"[DEBUG] Invalid key!\n";
        c.dump();
    }

public:
    Display(lw::EventManager& eventManager)
    {
        eventManager.registerTimer(this, lw::ElapsedTime(0), lw::ElapsedTime(1));

        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e+1));
        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e+0));
        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e-1));
        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e-2));
        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e-3));
        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e-4));
        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e-5));
        eventManager.registerTimer(&c, lw::ElapsedTime(0), lw::ElapsedTime(1e-6));
    }

    virtual ~Display()
    {
    }

};



// thread object to shut down event manager
class ShutdownThread : private lw::Thread {
private:
    lw::EventManager& eventManager;
    lw::Log Log;

    virtual void run()
    {
        Log << lw::Log::start(L"Thread") << L"Running.\n" << lw::Log::end;
        lw::Thread::sleep(lw::ElapsedTime(100));
        Log << lw::Log::start(L"Thread") << L"Completed.\n" << lw::Log::end;
        //eventManager.shutdown(); TODO
    }

public:
    ShutdownThread(lw::EventManager& eventManager)
        : eventManager(eventManager), Log(std::wcout, true)
    {
        start();
    }

    virtual ~ShutdownThread()
    {
        wait(-1);
    }
};



int main(int argc, char* argv[])
{
    if(argc == 2 && !strcmp(argv[1], "--print-summary")) {
        std::wcout << L"A test of the timer event subsystem.\n";
        return 0;
    }

    if(argc != 1) {
        std::wcerr << L"Not expecting any commandline arguments.\n";
        return 1;
    }

    try {
        // an initial delay, see ticket:6 for reason (basically we want to catch issues with
        // registering events after the event manager thread has started)
        //Log << lw::Log::start(L"main()") << L"Sleeping 5s..." << lw::Log::end;
        //lw::Thread::sleep(lw::ElapsedTime::from_ms(5000));

        // instantiate event manager
        lw::EventManager eventManager(0, 10);

        // set up timer objects
        Display display(eventManager);
        ShutdownThread shutdownThread(eventManager);

        Log << lw::Log::start(L"main()") << L"Test started." << lw::Log::end;

        // event loop
        try {
            while(true) eventManager.processEvents(-1);
        }
        catch(lw::EventManager::Shutdown&) { }

        Log << lw::Log::start(L"main()") << L"Test completed." << lw::Log::end;
    }
    catch(lw::Exception& e) {
        Err << lw::Log::start(L"main()") << e << lw::Log::end;
        throw;
    }
    catch(...) {
        Err << lw::Log::start(L"main()")
            << L"Unknown exception caught.\n"
            << lw::Log::end;
        throw;
    }

    return 0;
}

/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
*/
